/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "config.h"

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/sema.h>
#include <sys/mutex.h>
#include <sys/interrupt.h>
#include <sys/endian.h>
#include <sys/mbuf.h>
#include <sys/kthread.h>
#include <sys/socket.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <sys/selinfo.h>
#include <sys/poll.h>
#include <sys/ioccom.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>
#include <net/if_types.h>

#include <net80211/ieee80211_var.h>
#include <net80211/ieee80211_proto.h>
#include <net80211/ieee80211_ioctl.h>
#ifdef WITH_BPF
#include <net80211/ieee80211_radiotap.h>
#endif

#include <p54_ioctl.h>

#include "isl38xx_host.h"
#include "isl38xx_common.h"
#include "transport.h"
#include "input_layer.h"
#include "devnode.h"
#include "pmalloc.h"
#ifdef WITH_SOFTMAC
#include "softmac.h"
#endif
#ifdef WITH_FREEMAC
#include "freemac.h"
#endif
#include "netif.h"
#include "pda.h"
#include "p54u.h"

static void p54u_devnode_initq(struct p54u_devnode_queue *queue, int size)
{
	queue->q = malloc(size, M_P54U, M_NOWAIT);
	queue->p = queue->c = 0;
	queue->s = size;
}

static void p54u_devnode_cleanupq(struct p54u_devnode_queue *queue)
{
	free(queue->q, M_P54U);
}

static int p54u_devnode_inq(struct p54u_devnode_queue *queue)
{
	uint32_t inq;
	inq = queue->p - queue->c;
	return inq;
}

static uint8_t p54u_devnode_qconsume(struct p54u_devnode_queue *queue)
{
	uint8_t ret;

	ret = queue->q[queue->c % queue->s];
	(queue->c)++;
	return ret;
}

static void p54u_devnode_qproduce(struct p54u_devnode_queue *queue, uint8_t val)
{
	queue->q[queue->p % queue->s] = val;
	(queue->p)++;
}

#define P54U_UNIT(s) (minor(s)&0x1f)

static int p54u_devnode_open(struct cdev *dev, int flag, int fmt, struct thread *p)
{
	struct p54u_softc *sc;

	sc = p54u_get_sc(P54U_UNIT(dev));
	if(!sc)
		return EIO;
	if(sc->devnode_softc.is_open)
		return EBUSY; /* We don't want to talk to two processes at the same time */
	sc->devnode_softc.is_open = 1;

	return 0;
}

static int p54u_devnode_close(struct cdev *dev, int flag, int fmt, struct thread *p)
{
	struct p54u_softc *sc;

	sc = p54u_get_sc(P54U_UNIT(dev));
	if(!sc)
		return EIO;
	sc->devnode_softc.is_open = 0;

	return 0;
}

static int p54u_devnode_poll(struct cdev *dev, int events, struct thread *p)
{
	struct p54u_softc *sc;
	int revents;

	if(suser(curthread) != 0)
		return EPERM;

	sc = p54u_get_sc(P54U_UNIT(dev));
	if(!sc)
		return EIO;

	if((events&(POLLIN|POLLOUT|POLLRDNORM|POLLWRNORM)) == 0)
		return POLLNVAL;
	revents = 0;

	if(((events & POLLIN) != 0) && (p54u_devnode_inq(&sc->devnode_softc.rxq) != 0))
		revents |= POLLIN;
	if(((events & POLLRDNORM) != 0) && (p54u_devnode_inq(&sc->devnode_softc.rxq) != 0))
		revents |= POLLRDNORM;
	if(((events & POLLOUT) != 0) && (p54u_devnode_inq(&sc->devnode_softc.txq) != P54U_DEVNODE_TXQSIZE))
		revents |= POLLOUT;
	if(((events & POLLWRNORM) != 0) && (p54u_devnode_inq(&sc->devnode_softc.rxq) != 0))
		revents |= POLLWRNORM;

	if(revents == 0)
		selrecord(p, &sc->devnode_softc.selp);

	return revents;
}

static int p54u_devnode_read(struct cdev *dev, struct uio *uio, int flag)
{
	struct p54u_softc *sc;
	int rv;
	uint8_t data;

	if(suser(curthread) != 0)
		return EPERM;

	sc = p54u_get_sc(P54U_UNIT(dev));
	if(!sc)
		return EIO;

	rv = 0;
	while(uio->uio_resid > 0)
	{
		if(p54u_devnode_inq(&sc->devnode_softc.rxq) > 0) {
			data = p54u_devnode_qconsume(&sc->devnode_softc.rxq);
			rv = uiomove(&data, 1, uio);
			if(rv != 0)
				break;
			p54u_devnode_onread(sc);
		} else {
			rv = tsleep(&sc->devnode_softc.rxq, PZERO|PCATCH, "p54tr", 0);
			if(rv != 0)
				break;
			if(p54u_devnode_inq(&sc->devnode_softc.rxq) == 0) {
				rv = EIO;
				break;
			}
		}
	}

	return rv;
}

static int p54u_devnode_write(struct cdev *dev, struct uio *uio, int flag)
{
	struct p54u_softc *sc;
	int rv;
	uint8_t data;

	if(suser(curthread) != 0)
		return EPERM;

	sc = p54u_get_sc(P54U_UNIT(dev));
	if(!sc)
		return EIO;

	rv = 0;
	while(uio->uio_resid > 0)
	{
		if(p54u_devnode_inq(&sc->devnode_softc.txq) >= P54U_DEVNODE_TXQSIZE) {
			rv = tsleep(&sc->devnode_softc.txq, PZERO|PCATCH, "p54tw", 0);
			if(rv != 0)
				break;
			if(p54u_devnode_inq(&sc->devnode_softc.txq) >= P54U_DEVNODE_TXQSIZE) {
				rv = EIO;
				break;
			}
		}
		rv = uiomove(&data, 1, uio);
		if(rv != 0)
			break;
		p54u_devnode_qproduce(&sc->devnode_softc.txq, data);
		p54u_devnode_onwrite(sc);
	}

	return rv;
}

static int p54u_ioctl_reset(struct p54u_softc *sc)
{
	if(suser(curthread) != 0)
		return EPERM;
	if(!p54u_reset(sc))
		return EIO;
	return 0;
}

static int p54u_ioctl_firmware(struct p54u_softc *sc, struct p54_firmware *firmware)
{
	char *fw;
	int r;

	if(suser(curthread) != 0)
		return EPERM;
	fw = malloc(firmware->size, M_P54U, M_WAITOK);
	if((r = copyin(firmware->data, fw, firmware->size)) != 0) {
		free(fw, M_P54U);
		return r;
	}
	if(!p54u_boot(sc, fw, firmware->size)) {
		free(fw, M_P54U);
		return EIO;
	}
	free(fw, M_P54U);
	return 0;
}


static int p54u_ioctl_p54hwinfo(struct p54u_softc *sc, struct p54_hwinfo *hwinfo)
{
	hwinfo->eeprom_available = sc->eeprom_available;
	hwinfo->device_type = sc->device_type;
	if(sc->eeprom_available) {
		memcpy(&hwinfo->manufacturing_part_number[0], &sc->manufacturing_part_number[0], P54_MAX_ID_SIZE);
		memcpy(&hwinfo->nic_serial_number[0], &sc->nic_serial_number[0], P54_MAX_ID_SIZE);
		hwinfo->platform_id = sc->platform_id;
		hwinfo->antenna1_gain = sc->antenna1_gain;
		hwinfo->antenna2_gain = sc->antenna2_gain;
	}
	return 0;
}

static int p54u_ioctl_p54fwinfo(struct p54u_softc *sc, struct p54_fwinfo *fwinfo)
{
	if(!sc->firmware_running)
		return EAGAIN;
	memcpy(&fwinfo->type[0], &sc->firmware_type_string[0], P54_FWTYPE_SIZE);
#ifdef WITH_SOFTMAC

	if(P54U_IS_SOFTMAC(sc))
	{
		memcpy(&fwinfo->version[0], &sc->firmware_version_string[0], P54_MAX_VERSION_STRING_SIZE);
		memcpy(&fwinfo->build[0], &sc->firmware_build_string[0], P54_MAX_BUILD_STRING_SIZE);

		fwinfo->variant = sc->firmware_variant;
		fwinfo->bottom_compat = sc->firmware_bottom_compat;
		fwinfo->top_compat = sc->firmware_top_compat;
		fwinfo->physical_variant = sc->firmware_physical_variant;
		fwinfo->physical_bottom_compat = sc->firmware_physical_bottom_compat;
		fwinfo->physical_top_compat = sc->firmware_physical_top_compat;
	}
#endif
#ifdef WITH_FREEMAC
	if(P54U_IS_FREEMAC(sc))
	{
		strcpy(&fwinfo->version[0], "[FIXME] FreeMAC");
		strcpy(&fwinfo->build[0], "[FIXME] FreeMAC");

		fwinfo->variant = 0;
		fwinfo->bottom_compat = 0;
		fwinfo->top_compat = 0;
	}
#endif
	return 0;
}

static int p54u_ioctl_p54eepromdump(struct p54u_softc *sc, char **data)
{
	if(!sc->eeprom_available)
		return EAGAIN;
	return copyout(sc->eeprom_data, *data, P54U_EEPROM_SIZE);
}

static int p54u_ioctl_p54stats(struct p54u_softc *sc, struct p54_stats *stats)
{
	struct mbuf *m;

	if(!sc->firmware_running)
		return EAGAIN;
#ifdef WITH_SOFTMAC

	if(P54U_IS_SOFTMAC(sc))
		m = p54u_lmac_stats_readreq(sc, sc->control_address);
#endif
#ifdef WITH_FREEMAC

	if(P54U_IS_FREEMAC(sc))
		m = p54u_fmac_stats_readreq(sc, sc->control_address);
#endif

	if(m == NULL)
		return ENOMEM;
	p54u_enqueue_frame(sc, m, 0, sc->control_address);
	if(sema_timedwait(&sc->stats_read_sema, P54U_TRANSFER_TIMEOUT_HZ) != 0)
		return ETIMEDOUT;
	stats->rx_packets = sc->stats_rx_packets;
	stats->time = sc->stats_time;
	stats->unk1 = sc->stats_unk1;
	stats->unk2 = sc->stats_unk2;
	stats->unk3 = sc->stats_unk3;
	stats->unk4 = sc->stats_unk4;
	stats->unk5 = sc->stats_unk5;
	stats->unk6 = sc->stats_unk6;
	stats->unk7 = sc->stats_unk7;
	return 0;
}

static int p54u_ioctl_p54leds(struct p54u_softc *sc, struct p54_leds *leds)
{
	struct mbuf *m;

	if(suser(curthread) != 0)
		return EPERM;
	if(!sc->firmware_running)
		return EAGAIN;
	if(leds->drv_auto)
	{
		if(sc->leds_auto)
			return 0;
		sc->leds_auto = 1;
		if(!p54u_update_leds(sc, sc->sc_ic.ic_state))
			return ENOMEM;
	} else
	{
		sc->leds_auto = 0;
#ifdef WITH_SOFTMAC

		if(P54U_IS_SOFTMAC(sc))
			m = p54u_lmac_leds(sc, sc->control_address, leds->mode, leds->perm_setting, leds->temp_setting, leds->duration);
#endif
#ifdef WITH_FREEMAC

		if(P54U_IS_FREEMAC(sc))
			m = p54u_fmac_leds(sc, leds->mode, leds->perm_setting, leds->temp_setting, leds->duration);
#endif

		if(!p54u_enqueue_frame(sc, m, 0, sc->control_address))
			return ENOMEM;
	}
	return 0;
}

static int p54u_ioctl_p54pmalloc(struct p54u_softc *sc, struct p54_pmalloc *pmalloc)
{
	unsigned int i;

	if(!sc->firmware_running)
		return EAGAIN;
	mtx_lock(&sc->pmalloc_state.lock);
	pmalloc->start_address = sc->pmalloc_state.start_address;
	pmalloc->end_address = sc->pmalloc_state.end_address;
	pmalloc->allocated_blocks = sc->pmalloc_state.allocated_blocks;
	for(i=0;i<P54_PMALLOC_MAX_BLOCKS;i++)
	{
		pmalloc->blocks[i].start_address = sc->pmalloc_state.blocks[i].start_address;
		pmalloc->blocks[i].end_address = sc->pmalloc_state.blocks[i].end_address;
		pmalloc->blocks[i].headroom = sc->pmalloc_state.blocks[i].headroom;
		switch(sc->pmalloc_state.blocks[i].purpose) {
		case P54U_PMALLOC_PURPOSE_CONTROL:
			pmalloc->blocks[i].purpose = P54_PMALLOC_PURPOSE_CONTROL;
			break;
		case P54U_PMALLOC_PURPOSE_EEPROM:
			pmalloc->blocks[i].purpose = P54_PMALLOC_PURPOSE_EEPROM;
			break;
		case P54U_PMALLOC_PURPOSE_80211MGMT:
			pmalloc->blocks[i].purpose = P54_PMALLOC_PURPOSE_80211MGMT;
			break;
		case P54U_PMALLOC_PURPOSE_80211DATA:
			pmalloc->blocks[i].purpose = P54_PMALLOC_PURPOSE_80211DATA;
			break;
		case P54U_PMALLOC_PURPOSE_80211BEACON:
			pmalloc->blocks[i].purpose = P54_PMALLOC_PURPOSE_80211BEACON;
			break;
		case P54U_PMALLOC_PURPOSE_FWRESERVED:
			pmalloc->blocks[i].purpose = P54_PMALLOC_PURPOSE_FWRESERVED;
			break;
		case P54U_PMALLOC_PURPOSE_MISC:
			pmalloc->blocks[i].purpose = P54_PMALLOC_PURPOSE_MISC;
			break;
		}
	}
	mtx_unlock(&sc->pmalloc_state.lock);

	return 0;
}

static int p54u_ioctl_p54frame(struct p54u_softc *sc, struct p54_frame *frame)
{
	struct mbuf *m;
	uint32_t address;

	if(suser(curthread) != 0)
		return EPERM;
	if(!sc->firmware_running)
		return EAGAIN;

	MGETHDR(m, M_TRYWAIT, MT_DATA);
	if(m == NULL)
		return ENOMEM;

	if(frame->size > MHLEN)
	{
		MCLGET(m, M_TRYWAIT);
		if((m->m_flags & M_EXT) == 0) {
			m_freem(m);
			return ENOMEM;
		}
	}
	m->m_pkthdr.len = m->m_len = 0;
	m_copyback(m, 0, frame->size, &frame->data[0]);
	address = frame->address;

	if(!p54u_enqueue_frame(sc, m, 0, address))
		return ENOMEM;
	return 0;
}

static int p54u_ioctl_p54abortaddr(struct p54u_softc *sc, uint32_t *address)
{
	struct mbuf *m;

	if(suser(curthread) != 0)
		return EPERM;
	if(!sc->firmware_running)
		return EAGAIN;

#ifdef WITH_SOFTMAC
	if(P54U_IS_SOFTMAC(sc))
		m = p54u_lmac_abortaddress(sc, sc->control_address, *address);
#endif
#ifdef WITH_FREEMAC
	if(P54U_IS_FREEMAC(sc))
		m = p54u_fmac_abortaddress(sc, sc->control_address, *address);
#endif

	if(m == NULL)
		return ENOMEM;

	if(!p54u_enqueue_frame(sc, m, 0, sc->control_address))
		return ENOMEM;

	p54u_pmfree(&sc->pmalloc_state, *address);

	p54u_handle_device_available(sc);

	return 0;
}

static int p54u_ioctl_p54memread(struct p54u_softc *sc, struct p54_memread *memread)
{
	int r;
	int i;
	uint32_t memwin_baseaddr, memwin_offset;
	uint32_t b;

	if(suser(curthread) != 0)
		return EPERM;
	if((memread->length % sizeof(b)) != 0)
		return EINVAL;
	memwin_baseaddr = memread->startaddr;
	memwin_offset = 0;

	p54u_tx_lock(sc);
	p54u_dev_writel(sc, ISL38XX_DIR_MEM_BASE_REG, memwin_baseaddr);

	for(i=0;i<memread->length;i+=sizeof(b))
	{
		b = p54u_dev_readl(sc, (ISL38XX_DIRECT_MEM_WIN+memwin_offset));
		if((r = copyout(&b, ((char *)memread->storage)+i, sizeof(b))) != 0) {
			p54u_tx_unlock(sc);
			return r;
		}
		memwin_offset += sizeof(b);
		if(memwin_offset == ISL38XX_MEMORY_WINDOW_SIZE) {
			memwin_offset = 0;
			memwin_baseaddr += ISL38XX_MEMORY_WINDOW_SIZE;
			p54u_dev_writel(sc, ISL38XX_DIR_MEM_BASE_REG, memwin_baseaddr);
		}
		if(sc->err) {
			p54u_tx_unlock(sc);
			return EIO;
		}
	}
	p54u_tx_unlock(sc);

	return 0;
}

static int p54u_ioctl_p54crashdump(struct p54u_softc *sc, struct p54_crashdump *crashdump)
{
	int r;

	if(sc->crash_data == NULL)
	{
		/* No crash dump, set size to zero and return */
		crashdump->size = 0;
		return 0;
	}
	if((r = copyout(sc->crash_data, crashdump->storage, sc->crash_data_size > crashdump->size ? crashdump->size : sc->crash_data_size)) != 0)
		return r;
	crashdump->size = sc->crash_data_size;
	return 0;
}

static int p54u_devnode_ioctl(struct cdev *dev, u_long cmd, caddr_t data, int flags, struct thread *p)
{
	struct p54u_softc *sc;
	int r;

	sc = p54u_get_sc(P54U_UNIT(dev));
	if(!sc)
		return EIO;

	switch(cmd)
	{
	case P54RESET:
		r = p54u_ioctl_reset(sc);
		break;
	case P54FIRMWARE:
		r = p54u_ioctl_firmware(sc, (struct p54_firmware *)data);
		break;

	case P54HWINFO:
		r = p54u_ioctl_p54hwinfo(sc, (struct p54_hwinfo *)data);
		break;
	case P54FWINFO:
		r = p54u_ioctl_p54fwinfo(sc, (struct p54_fwinfo *)data);
		break;
	case P54EEPROMDUMP:
		r = p54u_ioctl_p54eepromdump(sc, (char **)data);
		break;

	case P54STATS:
		r = p54u_ioctl_p54stats(sc, (struct p54_stats *)data);
		break;
	case P54LEDS:
		r = p54u_ioctl_p54leds(sc, (struct p54_leds *)data);
		break;

	case P54PMALLOC:
		r = p54u_ioctl_p54pmalloc(sc, (struct p54_pmalloc *)data);
		break;
	case P54FRAME:
		r = p54u_ioctl_p54frame(sc, (struct p54_frame *)data);
		break;
	case P54ABORTADDR:
		r = p54u_ioctl_p54abortaddr(sc, (uint32_t *)data);
		break;
	case P54MEMREAD:
		r = p54u_ioctl_p54memread(sc, (struct p54_memread *)data);
		break;

	case P54CRASHDUMP:
		r = p54u_ioctl_p54crashdump(sc, (struct p54_crashdump *)data);
		break;
	}

	return r;
}

static struct cdevsw devnode_cdevsw = {
	.d_version =  D_VERSION,
	.d_flags =    D_NEEDGIANT,
	.d_open =     p54u_devnode_open,
	.d_close =    p54u_devnode_close,
	.d_poll =     p54u_devnode_poll,
	.d_read =     p54u_devnode_read,
	.d_write =    p54u_devnode_write,
	.d_ioctl =    p54u_devnode_ioctl,
	.d_name =     "p54u_devnode",
};

unsigned int p54u_devnode_dequeue(struct p54u_softc *sc, char *buffer, unsigned int size)
{
	unsigned int i;

	for(i=0;i<size;i++)
	{
		if(p54u_devnode_inq(&sc->devnode_softc.txq) == 0)
			break;
		buffer[i] = p54u_devnode_qconsume(&sc->devnode_softc.txq);
		wakeup_one(&sc->devnode_softc.txq);
		selwakeup(&sc->devnode_softc.selp);
	}
	return i;
}

unsigned int p54u_devnode_enqueue(struct p54u_softc *sc, const char *buffer, unsigned int size)
{
	unsigned int i;

	for(i=0;i<size;i++)
	{
		if(p54u_devnode_inq(&sc->devnode_softc.rxq) == P54U_DEVNODE_RXQSIZE)
			break;
		p54u_devnode_qproduce(&sc->devnode_softc.rxq, buffer[i]);
		wakeup_one(&sc->devnode_softc.rxq);
		selwakeup(&sc->devnode_softc.selp);
	}
	return i;
}

int p54u_devnode_init(struct p54u_softc *sc)
{
	P54U_DBG_GEN("Creating device node");

	sc->devnode_softc.is_open = 0;
	p54u_devnode_initq(&sc->devnode_softc.txq, P54U_DEVNODE_TXQSIZE);
	p54u_devnode_initq(&sc->devnode_softc.rxq, P54U_DEVNODE_RXQSIZE);
	sc->devnode_softc.dev = make_dev(&devnode_cdevsw, p54u_devunit(sc), UID_ROOT, GID_OPERATOR, 0666, "p54u%d", p54u_devunit(sc));
	if(sc->devnode_softc.dev == NULL)
		return 0;
	return 1;
}

int p54u_devnode_free(struct p54u_softc *sc)
{
	P54U_DBG_GEN("Removing device node");

	if(sc->devnode_softc.dev != NULL)
		destroy_dev(sc->devnode_softc.dev);

	/* Fail any ongoing uio operations */
	wakeup(&sc->devnode_softc.txq);
	wakeup(&sc->devnode_softc.rxq);
	selwakeup(&sc->devnode_softc.selp);

	p54u_devnode_cleanupq(&sc->devnode_softc.txq);
	p54u_devnode_cleanupq(&sc->devnode_softc.rxq);
	return 1;
}
