/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
  Copyright (C) 2004, 2005 Jean-Baptiste Note <jean-baptiste.note@m4x.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "config.h"

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/endian.h>
#include <sys/lock.h>
#include <sys/sema.h>
#include <sys/mbuf.h>

#ifdef WITH_SOFTMAC
#include "softmac.h"
#endif
#ifdef WITH_FREEMAC
#include "freemac.h"
#endif
#include "pmalloc.h"
#include "output_layer.h"
#include "eeprom.h"
#include "pda.h"
#include "p54u.h"

int p54u_eeprom_parse(struct p54u_softc *sc, uint8_t *eeprom_data, unsigned int eeprom_size)
{
	struct p54u_eeprom_header *eeprom;
	unsigned int pda_offset;

	eeprom = (struct p54u_eeprom_header *)eeprom_data;
	if(le32toh(eeprom->signature) != P54U_EEPROM_SIGNATURE)
		P54U_WARN("Unusual EEPROM signature: 0x%08x", le32toh(eeprom->signature));
	P54U_DBG_GEN("EEPROM firmware length: %d", le16toh(eeprom->firmware_length));
	pda_offset = sizeof(struct p54u_eeprom_header)+eeprom->firmware_length+4; /* +4 to skip firmware CRC */
	P54U_DBG_GEN("PDA at offset 0x%04x", pda_offset);
	if(pda_offset >= eeprom_size) {
		P54U_ERR("Computed PDA offset beyond end of EEPROM");
		return 0;
	}
	return p54u_pda_parse(sc, eeprom_data+pda_offset, eeprom_size-pda_offset);
}

/* Reads and parses the whole EEPROM */
int p54u_eeprom_read(struct p54u_softc *sc)
{
	int i, len, e;
	struct mbuf *m;
	uint32_t size;
	uint32_t address;

	P54U_DBG_GEN("Reading EEPROM");

	i = P54U_EEPROM_SIZE;
	while(i>0) {
		len = i > P54U_EEPROM_BLOCK ? P54U_EEPROM_BLOCK : i;
#ifdef WITH_SOFTMAC
		if(P54U_IS_SOFTMAC(sc)) {
			size = p54u_lmac_eeprom_readreq_size(len);
			/* is tailroom needed or not ? */
			address = p54u_pmalloc(&sc->pmalloc_state, size+P54U_LMAC_TAILROOM, P54U_LMAC_HEADROOM, P54U_PMALLOC_PURPOSE_EEPROM);
			if(address == P54U_PMALLOC_FAILURE) {
				P54U_ERR("Failed to allocate Prism memory space for EEPROM read");
				sc->err = 1;
				return 0;
			}
			P54U_DBG_GEN("Obtained space for EEPROM at address 0x%08x", address);
			m = p54u_lmac_eeprom_readreq(sc, address, P54U_EEPROM_SIZE-i, len);
			if(m == NULL) {
				P54U_ERR("Unable to generate EEPROM frame");
				p54u_pmfree(&sc->pmalloc_state, address);
				sc->err = 1;
				return 0;
			}
			p54u_enqueue_frame(sc, m, 1, address);
		}
#endif
#ifdef WITH_FREEMAC
		if(P54U_IS_FREEMAC(sc)) {
			m = p54u_fmac_eeprom_readreq(sc, P54U_EEPROM_SIZE-i, len);
			if(m == NULL) {
				P54U_ERR("Unable to generate EEPROM frame");
				p54u_pmfree(&sc->pmalloc_state, sc->control_address);
				sc->err = 1;
				return 0;
			}
			p54u_enqueue_frame(sc, m, 1, sc->control_address);
		}
#endif

		e = sema_timedwait(&sc->eeprom_read_sema, P54U_TRANSFER_TIMEOUT_HZ);
		if(e != 0) {
			P54U_ERR("Error waiting for EEPROM data");
			return 0;
		}
		i -= P54U_EEPROM_BLOCK;
	}
	P54U_DBG_GEN("EEPROM read complete, parsing contents");
	if(!p54u_eeprom_parse(sc, sc->eeprom_data, P54U_EEPROM_SIZE)) {
		P54U_ERR("Error in EEPROM contents");
		return 0;
	}
	P54U_DBG_GEN("EEPROM parsing successfully completed");
	return 1;
}
