/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
  Copyright (C) 2004 Feyd
  Copyright (C) 2004, 2005 Jean-Baptiste Note <jean-baptiste.note@m4x.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "config.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/endian.h>
#include <sys/bus.h>
#include <sys/lock.h>
#include <sys/sema.h>

#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>

#include "transport.h"

#include "isl38xx_common.h"
#include "isl38xx_host.h"
#define ISLPTR(x) x
#include "isl38xx_hardware.h"

#include "firmware.h"
#include "p54u.h"

int p54u_parse_bra(struct p54u_softc *sc, const char *data, unsigned int len)
{
	unsigned int run_length = 0;
	unsigned int max_offset = len;
	int end = 0;
	const struct p54u_bra_option *option;
	unsigned int next_option_offset;

	int seen_fwtype = 0;
#ifdef WITH_SOFTMAC
	int seen_version = 0;
	int seen_exposedif = 0;
	int seen_description = 0;
#endif

	/* Find the beginning of the BRA */
	while(run_length < max_offset) {
		if(*((const uint32_t *)(data+run_length)) == 0)
			break;
		run_length += sizeof(uint32_t);
	}

	while(run_length < max_offset) {
		if(*((const uint32_t *)(data+run_length)) != 0)
			break;
		run_length += sizeof(uint32_t);
	}

	P54U_DBG_GEN("Found start of boot record area at offset %d", run_length);

	/* Retreive and analyze its information */
	while(!end) {
		if(run_length >= max_offset) {
			P54U_ERR("Attempting to access beyond end of firmware during BRA parsing, aborting");
			sc->err = 1;
			return 0;
		}
		option = (const struct p54u_bra_option *)(data+run_length);
		next_option_offset = le32toh(option->length)*sizeof(uint32_t);
		if(run_length+next_option_offset > max_offset) {
			P54U_ERR("BRA option too long (0x%08x), aborting", next_option_offset);
			sc->err = 1;
			return 0;
		}

		switch(le32toh(option->code)) {

		case P54U_BRA_CODE_FWTYPE:
			P54U_DBG_GEN("BRA: Firmware type");
			if(option->length == (P54U_FWTYPE_SIZE/sizeof(uint32_t))) {
				memcpy(&sc->firmware_type_string[0], option->data, P54U_FWTYPE_SIZE);
				if(memcmp(option->data, P54U_FWTYPE_SIG_LM86, P54U_FWTYPE_SIZE) == 0)
					sc->firmware_type = P54U_FWTYPE_LM86;
				else if(memcmp(option->data, P54U_FWTYPE_SIG_LM87, P54U_FWTYPE_SIZE) == 0)
					sc->firmware_type = P54U_FWTYPE_LM87;
				else if(memcmp(option->data, P54U_FWTYPE_SIG_FMAC, P54U_FWTYPE_SIZE) == 0)
					sc->firmware_type = P54U_FWTYPE_FMAC;
				else {
					P54U_ERR("Unsupported firmware type : %c%c%c%c", option->data[0], option->data[1], option->data[2], option->data[3]);
					sc->err = 1;
					return 0;
				}
				seen_fwtype = 1;
			} else {
				P54U_ERR("Unexpected size for firmware type in BRA");
				sc->err = 1;
				return 0;
			}
			break;

		case P54U_BRA_CODE_VERSION:
			P54U_DBG_GEN("BRA: Version");
#ifdef WITH_SOFTMAC
			strncpy(&sc->firmware_version_string[0], option->data, next_option_offset > P54U_MAX_VERSION_STRING_SIZE ? P54U_MAX_BUILD_STRING_SIZE : next_option_offset);
			seen_version = 1;
#else
			P54U_WARN("BRA contains a version, but SoftMAC support is disabled");
#endif

			break;

		case P54U_BRA_CODE_DEPENDENT_IF:
			P54U_DBG_GEN("BRA: Dependent Interface");
			P54U_WARN("BRA contains a dependent interface");
#ifdef WITH_DEBUG_GEN
			if(debug_gen)
				p54u_data_debug(sc, 0, option->data, next_option_offset);
#endif

			break;

		case P54U_BRA_CODE_EXPOSED_IF: {
				P54U_DBG_GEN("BRA: Exposed Interface");
#ifdef WITH_SOFTMAC
				const struct p54u_bra_exposed_if *exposed_if = (const struct p54u_bra_exposed_if *)option->data;
				if((le16toh(exposed_if->role) != P54U_BRA_IF_ROLE_SERVER) || (le16toh(exposed_if->if_id) != P54U_BRA_IF_ID_LMAC_IF)) {
					P54U_ERR("Firmware exposed interface is not a LMAC server");
					sc->err = 1;
					return 0;
				}
				sc->firmware_variant = le16toh(exposed_if->variant);
				sc->firmware_bottom_compat = le16toh(exposed_if->bottom_compat);
				sc->firmware_top_compat = le16toh(exposed_if->top_compat);
				/* FIXME: Just guessing what those fields are, according to the */
				/* differences in protocol */
				/* And "0.1" and "1.1" look strange, maybe that's rather split in 2x1 byte */
				/* We need more logs with new firmwares */
				sc->firmware_physical_variant = le16toh(exposed_if->physical_variant);
				sc->firmware_physical_bottom_compat = le16toh(exposed_if->physical_bottom_compat);
				sc->firmware_physical_top_compat = le16toh(exposed_if->physical_top_compat);
				seen_exposedif = 1;
#else
				P54U_WARN("BRA contains an exposed interface description, but SoftMAC support is disabled");
#endif

				break;
			}

		case P54U_BRA_CODE_DESCR: {
				P54U_DBG_GEN("BRA: Description");
				const struct p54u_bra_lmac_description *lmac_description = (const struct p54u_bra_lmac_description *)option->data;
				if((!seen_exposedif) && ((!seen_fwtype) || (sc->firmware_type != P54U_FWTYPE_FMAC))) {
					P54U_ERR("Interface description before interface identification in BRA");
					sc->err = 1;
					return 0;
				}
				sc->frame_space_start = le32toh(lmac_description->frame_space_start);
				sc->frame_space_end = le32toh(lmac_description->frame_space_end);
				memcpy(&sc->rate_table[0], &lmac_description->rates[0], P54U_LMAC_MAXRATES);
				sc->rate_count = 0;
				while((sc->rate_count < P54U_LMAC_MAXRATES) && (sc->rate_table[sc->rate_count] != 0))
					sc->rate_count++;
				seen_description = 1;
				break;
		}
		case P54U_BRA_CODE_FINAL:
			P54U_DBG_GEN("BRA: Final");
			end = 1;
			break;

		default:
			P54U_WARN("BRA option 0x%08x is unknown", le32toh(option->code));
#ifdef WITH_DEBUG_GEN
			if(debug_gen)
				p54u_data_debug(sc, 0, option->data, next_option_offset);
#endif
			break;
		}
		run_length += sizeof(*option)+next_option_offset;
	}


	if(!seen_fwtype) {
		P54U_ERR("Software identification is missing in BRA");
		sc->err = 1;
		return 0;
	}

	/* we don't use the macros here (see comments in p54u.h) */
	if((sc->firmware_type == P54U_FWTYPE_LM86) || (sc->firmware_type == P54U_FWTYPE_LM87)) {
#ifdef WITH_SOFTMAC
		P54U_DBG_GEN("Build string for firmware is : %s", (const char *)(data+run_length));
		strncpy(&sc->firmware_build_string[0], data+run_length, P54U_MAX_BUILD_STRING_SIZE);
		if(!seen_version) {
			P54U_WARN("Version is missing in BRA");
			strcpy(&sc->firmware_version_string[0], "Unknown");
		}
		if(!seen_exposedif) {
			P54U_ERR("Exposed interface is missing in BRA");
			sc->err = 1;
			return 0;
		}
		if(!seen_description) {
			P54U_ERR("Interface description is missing in BRA");
			sc->err = 1;
			return 0;
		}
		sc->physical_transport_version = 0;
		if((sc->firmware_physical_bottom_compat <= P54U_LMAC_PHYSICAL_V1) && (P54U_LMAC_PHYSICAL_V1 <= sc->firmware_physical_top_compat))
			sc->physical_transport_version = 1;
		if((sc->firmware_physical_bottom_compat <= P54U_LMAC_PHYSICAL_V2) && (P54U_LMAC_PHYSICAL_V2 <= sc->firmware_physical_top_compat))
			sc->physical_transport_version = 2;
		/* FIXME: If anyone has a log of a ISL3886 with physical transport 1.1, please send */
		if((sc->physical_transport_version == 0) || ((sc->physical_transport_version == 2) && (sc->device_type != P54U_DEVTYPE_USB2))) {
			P54U_ERR("Unsupported physical transport version (%d.%d), please use another firmware", (sc->firmware_physical_variant & 0xff00) >> 8, sc->firmware_physical_variant & 0xff);
			sc->err = 1;
			return 0;
		}
#else
		P54U_ERR("Firmware is SoftMAC, but p54u is compiled without SoftMAC support");
		sc->err = 1;
		return 0;
#endif

	}

	if(sc->firmware_type == P54U_FWTYPE_FMAC)
	{
#ifdef WITH_FREEMAC
		P54U_WARN("FreeMAC firmware detected, enabling experimental support");
#else
		P54U_ERR("Firmware is FreeMAC, but p54u is compiled without FreeMAC support");
		sc->err = 1;
		return 0;
#endif

	}

	return 1;
}

#ifdef WITH_USB

static void p54u_make_x2_header(struct p54u_x2_header *x2, uint32_t firmware_startaddr, uint32_t firmware_size)
{
	memcpy(&x2->signature[0], P54U_X2_SIGNATURE, P54U_X2_SIGNATURE_SIZE);
	x2->startaddr = htole32(firmware_startaddr);
	x2->size = htole32(firmware_size);
	x2->crc = htole32(crc32(((char *)x2)+P54U_X2_SIGNATURE_SIZE, sizeof(struct p54u_x2_header)-P54U_X2_SIGNATURE_SIZE-sizeof(uint32_t)));
}

/* with version 2 devices, data need some formatting :
 * escape 0x7e to 0x7d 0x5e
 * escape 0x7d to 0x7d 0x5d
 */
static unsigned int firmware_escape_write_buffer(char *output, const char *input, unsigned int *write)
{
	unsigned int i, j;

	i = 0;
	j = 0;
	while(j < *write) {
		switch(input[i]) {
		case '~':
			if((j-*write) < 2) {
				*write = j;
				return i;
			}
			output[j++] = '}';
			output[j++] = '^';
			break;
		case '}':
			if((j-*write) < 2) {
				*write = j;
				return i;
			}
			output[j++] = '}';
			output[j++] = ']';
			break;
		default:
			output[j++] = input[i];
		}
		i++;
	}
	*write = j;
	return i;
}

#endif /* WITH_USB */

/* Uploads and starts a new firmware, assuming the device is in the appropriate state
 * (use p54u_initialize_device() before). Blocks and only returns once the firmware
 * is running or the process has failed. Takes care of all the protocol, with support
 * for 3886 and 3887.
 */
int p54u_upload_firmware(struct p54u_softc *sc, const char *fw, unsigned int fwsize)
{
	switch(sc->device_type) {
#ifdef WITH_USB
		case P54U_DEVTYPE_USB1: {
			uint32_t reg;
			const char *data;
			uint32_t j;
			unsigned int remaining, len;
	
			P54U_DBG_GEN("Loading the firmware");
	
			data = fw;
			remaining = fwsize;
	
			P54U_DBG_GEN("Initializing IRQs");
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_EN_REG, 0);
			if(sc->err)
				return 0;
	
			/* Ack the irqs */
			reg = p54u_dev_readl(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_IDENT_REG);
			p54u_dev_writel(sc, (ISL38XX_HARDWARE_REG+ISL38XX_INT_ACK_REG), reg);
			if(sc->err)
				return 0;
			P54U_DBG_GEN("IRQ initialization done");
	
			P54U_DBG_GEN("Setting up the memory window");
			p54u_dev_writel(sc, (ISL38XX_HARDWARE_REG+ISL38XX_DIR_MEM_BASE_REG), P54U_FIRMWARE_DMA_ENGINE_BASE);

			P54U_DBG_GEN("Enabling the DMA RX engine");
			p54u_dev_writel(sc, ISL38XX_DIRECT_MEM_WIN+REG38XX_RXDMA_CONTROL-P54U_FIRMWARE_DMA_ENGINE_BASE, 1);
	
			j = 0;
			while(remaining > 0) {
				len = remaining > P54U_FIRMWARE_BLOCK ? P54U_FIRMWARE_BLOCK : remaining;
	
				memcpy(sc->bus.usb.data_tx.buf, data, len);
				if(p54u_bulk_transfer(sc, &sc->bus.usb.data_tx, &len, 0) != USBD_NORMAL_COMPLETION) {
					P54U_ERR("Unable to send firmware block");
					sc->err = 1;
					return 0;
				}
	
				data += len;
				remaining -= len;
	
				p54u_dev_writel(sc, ISL38XX_DIRECT_MEM_WIN+REG38XX_RXDMA_MAXCNT-P54U_FIRMWARE_DMA_ENGINE_BASE, len);
	
				p54u_dev_writel(sc, (ISL38XX_DIRECT_MEM_WIN+REG38XX_RXDMA_BASE-P54U_FIRMWARE_DMA_ENGINE_BASE), (ISL38XX_DEV_FIRMWARE_ADDRESS+j));
	
				p54u_dev_writel(sc, (ISL38XX_HARDWARE_REG+ISL38XX_DMA_MASTER_ADDRESS), NET2280_BASE_ENDPOINTS);
				p54u_dev_writel(sc, (ISL38XX_HARDWARE_REG+ISL38XX_DMA_MASTER_LEN), len/4);
				p54u_dev_writel(sc, (ISL38XX_HARDWARE_REG+ISL38XX_DMA_MASTER_CONTROL), 4);
				reg = p54u_dev_readl(sc, ISL38XX_DIRECT_MEM_WIN+REG38XX_RXDMA_STATUS-P54U_FIRMWARE_DMA_ENGINE_BASE);
				if(reg != (ISL38XX_DMA_STATUS_READY|ISL38XX_DMA_STATUS_DONE))
					P54U_WARN("DMA transfer of a firmware block may have failed");
				if(remaining > 0)
					p54u_brg_writel(sc, NET2280_EPA_STAT, NET2280_FIFO_FLUSH);
				if(sc->err) {
					P54U_ERR("Unable to write to registers during upload");
					return 0;
				}
				j += P54U_FIRMWARE_BLOCK;
			}
	
			reg = p54u_dev_readl(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG);
			reg |= ISL38XX_CTRL_STAT_RAMBOOT;
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG, reg);
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG, reg|ISL38XX_CTRL_STAT_RESET);
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG, reg);
	
			if(sc->err) {
				P54U_ERR("Unable to write to registers after upload");
				return 0;
			}
			P54U_DBG_GEN("Firmware uploaded, setting up the device");
	
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_EN_REG, ISL38XX_INT_INIT);

			p54u_brg_writel(sc, NET2280_IRQSTAT1, NET2280_PCI_INTA_INTERRUPT);
			p54u_brg_writel(sc, NET2280_USBIRQENB1, NET2280_PCI_INTA_INTERRUPT_ENABLE|NET2280_USB_INTERRUPT_ENABLE);
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_DEV_INT_REG, ISL38XX_INT_RESET);
	
			if(sema_timedwait(&sc->init_sema, P54U_TRANSFER_TIMEOUT_HZ) != 0) {
				P54U_ERR("Firmware startup failed");
				sc->err = 1;
				return 0;
			}
	
			P54U_DBG_GEN("Firmware startup succeeded");
			return 1;
		}
		case P54U_DEVTYPE_USB2: {
			unsigned int remaining, read, written;
			const char *data;
			uint32_t crc;
	
			P54U_DBG_GEN("Loading the firmware");
	
			/* Send the "download firmware" command to the BootROM */
			*((char *)(sc->bus.usb.data_tx.buf)+0) = '<';
			*((char *)(sc->bus.usb.data_tx.buf)+1) = '\r';
			written = 2;
			if(p54u_bulk_transfer(sc, &sc->bus.usb.data_tx, &written, 0) != USBD_NORMAL_COMPLETION) {
				P54U_ERR("Could not send start command");
				sc->err = 1;
				return 0;
			}
	
			P54U_DBG_GEN("Sending the X2 header");
			p54u_make_x2_header((struct p54u_x2_header *)sc->bus.usb.data_tx.buf, ISL38XX_DEV_FIRMWARE_ADDRESS, fwsize);
			written = sizeof(struct p54u_x2_header);
			if(p54u_bulk_transfer(sc, &sc->bus.usb.data_tx, &written, 0) != USBD_NORMAL_COMPLETION) {
				P54U_ERR("Could not send the X2 header");
				sc->err = 1;
				return 0;
			}
	
			data = fw;
			remaining = fwsize;
			P54U_DBG_GEN("Sending the firmware code");
			while(remaining > 0) {
				written = remaining > P54U_FIRMWARE_BLOCK ? P54U_FIRMWARE_BLOCK : remaining;
				read = firmware_escape_write_buffer(sc->bus.usb.data_tx.buf, data, &written);
				if(p54u_bulk_transfer(sc, &sc->bus.usb.data_tx, &written, 0) != USBD_NORMAL_COMPLETION) {
					P54U_ERR("Could not send firmware block, %d bytes were remaining", remaining);
					sc->err = 1;
					return 0;
				}
				data += read;
				remaining -= read;
			}
			crc = crc32(fw, fwsize);
			P54U_DBG_GEN("Firmware sent, starting it");
	
			/* Send the CRC, then the "go" command to the boot ROM */
			*((uint32_t *)(sc->bus.usb.data_tx.buf)) = htole32(crc);
			*((char *)(sc->bus.usb.data_tx.buf)+sizeof(uint32_t)+0) = 'g';
			*((char *)(sc->bus.usb.data_tx.buf)+sizeof(uint32_t)+1) = '\r';
			written = sizeof(uint32_t)+2;
			if(p54u_bulk_transfer(sc, &sc->bus.usb.data_tx, &written, 0) != USBD_NORMAL_COMPLETION) {
				P54U_ERR("Could not send GO command");
				sc->err = 1;
				return 0;
			}
			P54U_DBG_GEN("Send command OK");
	
			do {
				read = sc->bus.usb.data_rx.size;
				if(p54u_bulk_transfer(sc, &sc->bus.usb.data_rx, &read, USBD_SHORT_XFER_OK) != USBD_NORMAL_COMPLETION) {
					P54U_ERR("Could not get device response to firmware");
					sc->err = 1;
					return 0;
				}
				if(memcmp(sc->bus.usb.data_rx.buf, "ERROR", 5) == 0) {
					P54U_ERR("Got ERROR response to firmware upload");
					sc->err = 1;
					return 0;
				}
			} while(*(char *)sc->bus.usb.data_rx.buf != 'g');
	
			P54U_DBG_GEN("Firmware startup succeeded");
	
			return 1;
		}
#endif
#ifdef WITH_PCI
		case P54U_DEVTYPE_PCI: {
			uint32_t reg;
			unsigned int block_length;
			unsigned int position, remaining;
			unsigned int i;
		
			P54U_DBG_GEN("Writing firmware to the chip's RAM");
			/* Write firmware to the chip's RAM using the memory window */
			position = 0;
			remaining = fwsize;
			
			while(remaining > 0) {
				block_length = remaining > ISL38XX_MEMORY_WINDOW_SIZE ? ISL38XX_MEMORY_WINDOW_SIZE : remaining;
	
				p54u_dev_writel(sc, ISL38XX_DIR_MEM_BASE_REG, ISL38XX_DEV_FIRMWARE_ADDRESS+position);
				DELAY(ISL38XX_WRITEIO_DELAY);
				for(i=0;i<block_length;i+=4)
					p54u_dev_writel(sc, ISL38XX_DIRECT_MEM_WIN+i, *((const unsigned int *)(fw+position+i)));
				position += block_length;
				remaining -= block_length;
			}
			
			P54U_DBG_GEN("Firmware written, resetting device");
			
			/* Clear Reset and Clkrun, set Ramboot */
			reg = p54u_dev_readl(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG);
			reg &= ~(ISL38XX_CTRL_STAT_CLKRUN|ISL38XX_CTRL_STAT_RESET);
			reg |= ISL38XX_CTRL_STAT_RAMBOOT;
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG, reg);
			DELAY(ISL38XX_WRITEIO_DELAY);
			/* Setting the Reset bit latches the host override and Ramboot bits */
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG, reg|ISL38XX_CTRL_STAT_RESET);
			DELAY(ISL38XX_WRITEIO_DELAY);
			/* Clearing the Reset bit starts everything */
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG, reg);
			DELAY(ISL38XX_WRITEIO_DELAY);
			
			if(sc->err) {
				P54U_ERR("Unable to write to registers after upload");
				return 0;
			}
			
			P54U_DBG_GEN("Firmware uploaded, starting it with DMA control block at 0x%08x", sc->bus.pci.cb_dma_addr);
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_EN_REG, ISL38XX_INT_INIT);
			DELAY(ISL38XX_WRITEIO_DELAY);
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_GEN_PURP_COM_REG_1, sc->bus.pci.cb_dma_addr);
			DELAY(ISL38XX_WRITEIO_DELAY);
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_DEV_INT_REG, ISL38XX_INT_RESET);
			DELAY(ISL38XX_WRITEIO_DELAY);
			P54U_DBG_GEN("Done.");
			
			if(sema_timedwait(&sc->init_sema, P54U_TRANSFER_TIMEOUT_HZ) != 0) {
				P54U_ERR("Firmware startup failed");
				sc->err = 1;
				return 0;
			}
	
			
			return 1;
		}
#endif
	}
	return 0;
}

/* Performs the full firmware setup :
 * - parses the boot record area, and fills the appropriate fields in the softc
 * - sends the firmware to the device and starts it
 */
int p54u_init_firmware(struct p54u_softc *sc, char *fw, unsigned int size)
{
#ifdef WITH_DEBUG_GEN
	P54U_DBG_GEN("Firmware init (%d bytes)", size);
	if(debug_gen)
		p54u_data_debug(sc, 0, fw, size > 0x100 ? 0x100 : size);
#endif

	if(!p54u_parse_bra(sc, fw, size))
		return 0;
	if(!p54u_upload_firmware(sc, fw, size))
		return 0;
	sc->firmware_running = 1;
	return 1;
}
