/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "config.h"

#ifdef WITH_FREEMAC

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/endian.h>
#include <sys/mbuf.h>

#include "freemac.h"
#include "transport.h"
#include "input_layer.h"
#include "p54u.h"

#ifdef WITH_USB

/* Synchronous function that sends a FreeMAC frame contained in mbuf m at the specified address */
/* Must aquire appropriate locks before called */
int p54u_fmac_usb_transport(struct p54u_softc *sc, struct mbuf *m, uint32_t address)
{
	int len;
	int ok;

	switch(sc->device_type) {
		case P54U_DEVTYPE_USB1: {
			/* VERSION 1 DEVICES : Encapsulate data in a p54u_fmac_physical_tx structure which will contain the Prism address and the length */
			struct p54u_fmac_physical_tx *p54u_tx;
	
			/* Backup the payload length */
			len = m->m_pkthdr.len;
			M_PREPEND(m, sizeof(struct p54u_fmac_physical_tx), M_TRYWAIT);
			if(m == NULL)
				return 0;
			m = m_pullup(m, sizeof(struct p54u_fmac_physical_tx));
			if(m == NULL)
				return 0;
			p54u_tx = mtod(m, struct p54u_fmac_physical_tx *);
			bzero(p54u_tx, sizeof(struct p54u_fmac_physical_tx));
	
			p54u_tx->address = htole32(address);
			p54u_tx->size = htole16(len);
	
			ok = p54u_usb_transport(sc, m);
			break;
		}
		case P54U_DEVTYPE_USB2: {
			/* VERSION 2 DEVICES */
			/* TODO */
			ok = 0;
			/*
			struct p54u_lmac_87_v11_physical_tx *p54u_tx;
			
			M_PREPEND(m, sizeof(struct p54u_lmac_87_v11_physical_tx), M_TRYWAIT);
			if(m == NULL) return 0;
			m = m_pullup(m, sizeof(struct p54u_lmac_87_v11_physical_tx));
			if(m == NULL) return 0;
			p54u_tx = mtod(m, struct p54u_lmac_87_v11_physical_tx *);
			bzero(p54u_tx, sizeof(struct p54u_lmac_87_v11_physical_tx));
			
			p54u_tx->address = htole32(address);
			
			ok = p54u_usb_transport(sc, m);
			*/
			break;
		}
	}

	if(!ok)
		sc->err = 1;

	return ok;
}

#endif

/************************/
/*** Output functions ***/
/************************/

struct mbuf *p54u_fmac_abortaddress(struct p54u_softc *sc, uint32_t req_id, uint32_t aborted_address)
{
	/* TODO */
	return NULL;
}

struct mbuf *p54u_fmac_leds(struct p54u_softc *sc, uint16_t mode, uint16_t perm_setting, uint16_t temp_setting, uint16_t duration)
{
	struct mbuf *m;
	struct p54u_fmac_leds *frame;
	uint16_t perm_setting2, temp_setting2;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if(m == NULL)
		return NULL;
	m->m_pkthdr.len = m->m_len = sizeof(struct p54u_fmac_leds);
	frame = mtod(m, struct p54u_fmac_leds *);

	if(sc->leds_inverted)
	{
		perm_setting2 = perm_setting^(P54U_FMAC_LED_A|P54U_FMAC_LED_B);
		temp_setting2 = temp_setting^(P54U_FMAC_LED_A|P54U_FMAC_LED_B);
	} else
	{
		perm_setting2 = perm_setting;
		temp_setting2 = temp_setting;
	}

	frame->packet_type = htole16(P54U_FMAC_TYPE_LEDS);
	frame->mode = htole16(mode);
	frame->perm_setting = htole16(perm_setting2);
	frame->temp_setting = htole16(temp_setting2);
	frame->duration = htole16(duration);

	return m;
}

struct mbuf *p54u_fmac_eeprom_readreq(struct p54u_softc *sc, uint16_t offset, uint16_t length)
{
	struct mbuf *m;
	struct p54u_fmac_eeprom *frame;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if(m == NULL)
		return NULL;
	MCLGET(m, M_DONTWAIT);
	if((m->m_flags & M_EXT) == 0)
	{
		m_freem(m);
		return NULL;
	}
	m->m_pkthdr.len = m->m_len = sizeof(struct p54u_fmac_eeprom)+length;
	frame = mtod(m, struct p54u_fmac_eeprom *);
	/* bzero is not required, but uncomment if you want clean logs */
	bzero(frame, sizeof(struct p54u_fmac_eeprom)+length);

	frame->packet_type = htole16(P54U_FMAC_TYPE_EEPROM);
	frame->offset = htole16(offset);
	frame->length = htole16(length);

	return m;
}

struct mbuf *p54u_fmac_stats_readreq(struct p54u_softc *sc, uint32_t req_id)
{
	/* TODO */
	return NULL;
}

/***********************/
/*** Input functions ***/
/***********************/

static void p54u_fmac_input_eeprom(struct p54u_softc *sc, struct mbuf *m)
{
	struct p54u_fmac_eeprom *eeprom;
	uint16_t offset, length;

	m = m_pullup(m, sizeof(struct p54u_fmac_eeprom));
	if(m == NULL) {
		P54U_WARN("Unable to pull up mbuf for EEPROM header");
		return;
	}
	eeprom = mtod(m, struct p54u_fmac_eeprom *);

	offset = le16toh(eeprom->offset);
	length = le16toh(eeprom->length);
	m_adj(m, sizeof(struct p54u_fmac_eeprom));
	if(m->m_pkthdr.len < length) {
		P54U_WARN("mbuf too short for announced length");
		m_freem(m);
		return;
	}
	m_adj(m, -(m->m_pkthdr.len-length));

	p54u_handle_eeprom(sc, m, offset);
}

void p54u_fmac_input(struct p54u_softc *sc, struct mbuf *m)
{
	uint16_t *frame_type;

	m_adj(m, 0x10); /* TODO: take physical header into account, if it's really needed */
	m = m_pullup(m, sizeof(uint16_t));
	if(m == NULL) {
		P54U_WARN("Unable to pull up mbuf for frame type");
		return;
	}
	frame_type = mtod(m, uint16_t *);
	P54U_DBG_GEN("Received frame from FreeMAC, type 0x%04x", *frame_type);
	switch(*frame_type) {
		case P54U_FMAC_TYPE_EEPROM:
			p54u_fmac_input_eeprom(sc, m);
			break;
		default:
			P54U_WARN("Unknown frame type 0x%04x", *frame_type);
			m_freem(m);
			break;
	}
}

void p54u_fmac_init(struct p54u_softc *sc)
{
	P54U_DBG_GEN("Setting up FreeMAC protocol layer");
}

void p54u_fmac_free(struct p54u_softc *sc)
{
	P54U_DBG_GEN("Shutting down FreeMAC protocol layer");
}

#endif /* WITH_FREEMAC */
