/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "config.h"

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/lock.h>
#include <sys/sema.h>
#include <sys/endian.h>
#include <sys/socket.h>

#include <net/bpf.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>
#include <net/if_types.h>

#include <net80211/ieee80211_var.h>
#include <net80211/ieee80211_proto.h>
#include <net80211/ieee80211_ioctl.h>
#ifdef WITH_BPF
#include <net80211/ieee80211_radiotap.h>
#endif

#include "input_layer.h"
#include "netif.h"
#include "eeprom.h"
#include "p54u.h"

void p54u_handle_device_available(struct p54u_softc *sc)
{
	struct ifnet *ifp = sc->sc_ic.ic_ifp;

	if(sc->if_initialized)
	{
		mtx_lock(&sc->network_lock);
		if(ifp->if_drv_flags & IFF_DRV_OACTIVE) {
			ifp->if_drv_flags &= ~IFF_DRV_OACTIVE;
			p54u_start(ifp);
		}
		mtx_unlock(&sc->network_lock);
	}
}

void p54u_handle_stats(struct p54u_softc *sc, unsigned int rx_packets, unsigned int time, unsigned int unk1, unsigned int unk2, unsigned int unk3, unsigned int unk4, unsigned int unk5, unsigned int unk6, unsigned int unk7)
{
	sc->stats_rx_packets = rx_packets;
	sc->stats_time = time;
	sc->stats_unk1 = unk1;
	sc->stats_unk2 = unk2;
	sc->stats_unk3 = unk3;
	sc->stats_unk4 = unk4;
	sc->stats_unk5 = unk5;
	sc->stats_unk6 = unk6;
	sc->stats_unk7 = unk7;

	sema_post(&sc->stats_read_sema);
}

int p54u_handle_eeprom(struct p54u_softc *sc, struct mbuf *m, int offset)
{
	if((m->m_pkthdr.len+offset) > P54U_EEPROM_SIZE)
	{
		P54U_ERR("EEPROM readback beyond end of memory");
		sc->err = 1;
		return 0;
	}
	m_copydata(m, 0, m->m_pkthdr.len, &sc->eeprom_data[offset]);
	m_freem(m);

	sema_post(&sc->eeprom_read_sema);

	return 1;
}

int p54u_handle_80211rx(struct p54u_softc *sc, struct mbuf *m, unsigned int rate, unsigned int freq, unsigned int antsignal, unsigned int timestamp)
{
	struct ieee80211com *ic = &sc->sc_ic;
	struct ifnet *ifp = ic->ic_ifp;
	struct ieee80211_node *ni;
	struct ieee80211_frame_min *wh;

	if(!sc->if_initialized)
	{
		P54U_WARN("802.11 RX with interface not ready");
		m_freem(m);
		return 0;
	}

	if(m->m_pkthdr.len < (sizeof(struct ieee80211_frame)+IEEE80211_CRC_LEN))
	{
		P54U_DBG_NET("Frame too short (%d), dropped", m->m_pkthdr.len);
		m_freem(m);
		return 0;
	}

	m_adj(m, -IEEE80211_CRC_LEN);

	m = m_pullup(m, sizeof(struct ieee80211_frame));
	if(m == NULL)
	{
		P54U_WARN("Unable to pull up mbuf for 802.11 header");
		return 0;
	}

	m->m_pkthdr.rcvif = ifp;

#ifdef WITH_BPF
	if(sc->sc_drvbpf != NULL) {
		struct p54u_rx_radiotap_header *tap = &sc->sc_rxtap;

		tap->wr_flags = 0;
		tap->wr_chan_freq = htole16(freq);
		tap->wr_chan_flags = htole16(IEEE80211_CHAN_G);
		tap->rate = rate;
		tap->wr_antsignal = antsignal;

		bpf_mtap2(sc->sc_drvbpf, tap, sc->sc_rxtap_len, m);
	}
#endif

	wh = mtod(m, struct ieee80211_frame_min *);

	/* Update the RX data rate with data frames only */
	if((wh->i_fc[0] & IEEE80211_FC0_TYPE_MASK) == IEEE80211_FC0_TYPE_DATA)
		sc->rxrate = rate;

	/* FIXME: workaround for net80211 bug */
	if((wh->i_fc[0] & IEEE80211_FC0_TYPE_MASK) != IEEE80211_FC0_TYPE_DATA) {
		m = m_pullup(m, m->m_pkthdr.len);
		if(m == NULL) {
			P54U_WARN("Unable to pull up mbuf for mgmt frame");
			return 0;
		}
	}

	if(((wh->i_fc[0] & IEEE80211_FC0_TYPE_MASK) == IEEE80211_FC0_TYPE_DATA) && (ic->ic_state == IEEE80211_S_RUN))
		p54u_led_event(sc);

	ni = ieee80211_find_rxnode(ic, wh);

	ieee80211_input(ic, m, ni, antsignal, timestamp);

	ieee80211_free_node(ni);

	return 1;
}
