/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
  Copyright (C) 2004, 2005 Jean-Baptiste Note <jean-baptiste.note@m4x.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "config.h"

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/endian.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/mutex.h>
#include <sys/sema.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/mbuf.h>
#include <sys/sockio.h>

#ifdef WITH_BPF
#include <net/bpf.h>
#endif
#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>
#include <net/if_types.h>

#include <net80211/ieee80211_var.h>
#include <net80211/ieee80211_proto.h>
#include <net80211/ieee80211_ioctl.h>
#ifdef WITH_BPF
#include <net80211/ieee80211_radiotap.h>
#endif

#include <p54_ioctl.h>

#include "isl38xx_host.h"
#include "pmalloc.h"
#ifdef WITH_SOFTMAC
#include "softmac.h"
#endif
#ifdef WITH_FREEMAC
#include "freemac.h"
#endif
#include "netif.h"
#include "p54u.h"

#ifdef WITH_SOFTMAC
static int p54u_lmac_configure(struct p54u_softc *sc, enum ieee80211_state nstate)
{
	struct ieee80211com *ic = &sc->sc_ic;
	struct ifnet *ifp = ic->ic_ifp;
	struct ieee80211_node *ni = ic->ic_bss;
	const uint8_t *bssid;
	uint16_t mode = P54U_TX_CONTROL_FILTER_NOTYPE;
	uint16_t chan_m1, chan_m2;
	uint16_t magic2 = P54U_TX_CONTROL_FILTER_MAGIC2_FORM2;
	uint32_t magic3 = P54U_TX_CONTROL_FILTER_MAGIC3_FORM1;
	uint32_t magic4 = P54U_TX_CONTROL_FILTER_MAGIC4_FORM1;
	uint16_t magic8 = 0;
	uint16_t magic9 = 0;
	struct mbuf *m;

	if(nstate != IEEE80211_S_INIT)
	{
		/* compute the elements needed for the rx/tx filter */

		if(nstate == IEEE80211_S_SCAN) {
			bssid = ifp->if_broadcastaddr;
			mode = P54U_TX_CONTROL_FILTER_NOTYPE;
		} else {
			bssid = ni->ni_bssid;
		}

		chan_m1 = P54U_TX_CONTROL_CHANNEL_MAGIC1_TX;
		chan_m2 = P54U_TX_CONTROL_CHANNEL_MAGIC2_TX;

		if((nstate == IEEE80211_S_AUTH) || (nstate == IEEE80211_S_RUN)) {
			switch(ic->ic_opmode) {
			case IEEE80211_M_STA:
				magic2 = P54U_TX_CONTROL_FILTER_MAGIC2_STA;
				magic3 = P54U_TX_CONTROL_FILTER_MAGIC3_STA;
				magic4 = P54U_TX_CONTROL_FILTER_MAGIC4_STA;
				magic8 = P54U_TX_CONTROL_FILTER_MAGIC8_STA;
				mode = P54U_TX_CONTROL_FILTER_STA;
				break;
			case IEEE80211_M_AHDEMO:
			case IEEE80211_M_IBSS:
				mode = P54U_TX_CONTROL_FILTER_ADHOC;
				break;
			case IEEE80211_M_HOSTAP:
				mode = P54U_TX_CONTROL_FILTER_HOSTAP;
				break;
			case IEEE80211_M_MONITOR:
				magic2 = P54U_TX_CONTROL_FILTER_MAGIC2_MONITOR;
				magic3 = P54U_TX_CONTROL_FILTER_MAGIC3_MONITOR;
				magic4 = P54U_TX_CONTROL_FILTER_MAGIC4_MONITOR;
				mode = P54U_TX_CONTROL_FILTER_MONITOR;
				break;
			}
		}

		/* should also set rate, etc but i don't know yet how to set them */
		m = p54u_lmac_channel_change(sc, sc->control_address, ieee80211_mhz2ieee(ic->ic_curchan->ic_freq, IEEE80211_CHAN_2GHZ), chan_m1, chan_m2);
		if(m == NULL) {
			P54U_WARN("Unable to configure device (channel)");
			return 0;
		}
		p54u_enqueue_frame(sc, m, 1, sc->control_address);

		m = p54u_lmac_set_filter(sc, sc->control_address, mode, bssid, &ic->ic_myaddr[0], sc->rxhw, magic2, magic3, magic4, sc->reserved_address, magic8, magic9);
		if(m == NULL) {
			P54U_WARN("Unable to configure device (filter)");
			return 0;
		}
		p54u_enqueue_frame(sc, m, 1, sc->control_address);
	}
	return 1;
}
#endif

#ifdef WITH_FREEMAC
static int p54u_fmac_configure(struct p54u_softc *sc, enum ieee80211_state nstate)
{
	/* TODO */
	return 0;
}
#endif

static void p54u_init(void *priv)
{
	struct p54u_softc *sc = (struct p54u_softc *)priv;
	struct ieee80211com *ic = &sc->sc_ic;
	struct ifnet *ifp = ic->ic_ifp;
	struct ifaddr *ifa;
	struct sockaddr_dl *sdl;

	mtx_lock(&sc->network_lock);

	P54U_DBG_NET("p54u_init");

	ifa = ifaddr_byindex(ifp->if_index);
	sdl = (struct sockaddr_dl *)ifa->ifa_addr;
	IEEE80211_ADDR_COPY(ic->ic_myaddr, LLADDR(sdl));

	if (ic->ic_opmode != IEEE80211_M_MONITOR) {
		if(ic->ic_roaming != IEEE80211_ROAMING_MANUAL)
			ieee80211_new_state(ic, IEEE80211_S_SCAN, -1);
	} else {
		ieee80211_new_state(ic, IEEE80211_S_RUN, -1);
	}

	mtx_unlock(&sc->network_lock);
}

int p54u_update_leds(struct p54u_softc *sc, enum ieee80211_state state)
{
	struct mbuf *m;

	if(!sc->leds_auto)
		return 1;
	sc->last_led_event = ticks;
	switch(state)
	{
	case IEEE80211_S_SCAN:
#ifdef WITH_SOFTMAC

		if(P54U_IS_SOFTMAC(sc))
			m = p54u_lmac_leds(sc, sc->control_address, P54U_LMAC_LED_MODE_SET, 0, P54U_LMAC_LED_A|P54U_LMAC_LED_B, P54U_SCAN_LEDFLASH);
#endif
#ifdef WITH_FREEMAC

		if(P54U_IS_FREEMAC(sc))
			m = p54u_fmac_leds(sc, P54U_FMAC_LED_MODE_SET, 0, P54U_FMAC_LED_A|P54U_FMAC_LED_B, P54U_SCAN_LEDFLASH);
#endif

		break;
	case IEEE80211_S_RUN:
#ifdef WITH_SOFTMAC

		if(P54U_IS_SOFTMAC(sc))
			m = p54u_lmac_leds(sc, sc->control_address, P54U_LMAC_LED_MODE_SET, P54U_LMAC_LED_A|P54U_LMAC_LED_B, P54U_LMAC_LED_A|P54U_LMAC_LED_B, 0);
#endif
#ifdef WITH_FREEMAC

		if(P54U_IS_FREEMAC(sc))
			m = p54u_fmac_leds(sc, P54U_FMAC_LED_MODE_SET, P54U_FMAC_LED_A|P54U_FMAC_LED_B, 0, 0);
#endif

		break;
	default:
#ifdef WITH_SOFTMAC

		if(P54U_IS_SOFTMAC(sc))
			m = p54u_lmac_leds(sc, sc->control_address, P54U_LMAC_LED_MODE_SET, 0, 0, 0);
#endif
#ifdef WITH_FREEMAC

		if(P54U_IS_FREEMAC(sc))
			m = p54u_fmac_leds(sc, P54U_FMAC_LED_MODE_SET, 0, 0, 0);
#endif

		break;
	}
	if(m == NULL)
		return 0;
	p54u_enqueue_frame(sc, m, 0, sc->control_address);
	return 1;
}

int p54u_led_event(struct p54u_softc *sc)
{
	struct mbuf *m;

	if(!sc->leds_auto)
		return 1;
	if((ticks - sc->last_led_event) < 2*P54U_ACT_LEDFLASH)
		return 1;
	sc->last_led_event = ticks;
#ifdef WITH_SOFTMAC

	if(P54U_IS_SOFTMAC(sc))
		m = p54u_lmac_leds(sc, sc->control_address, P54U_LMAC_LED_MODE_SET, P54U_LMAC_LED_A|P54U_LMAC_LED_B, 0, P54U_ACT_LEDFLASH);
#endif
#ifdef WITH_FREEMAC

	if(P54U_IS_FREEMAC(sc))
		m = p54u_fmac_leds(sc, P54U_FMAC_LED_MODE_SET, P54U_FMAC_LED_A|P54U_FMAC_LED_B, 0, P54U_ACT_LEDFLASH);
#endif

	if(m == NULL)
		return 0;
	p54u_enqueue_frame(sc, m, 0, sc->control_address);

	return 1;
}

static int p54u_ioctl(struct ifnet *ifp, u_long cmd, caddr_t data)
{
	struct p54u_softc *sc = ifp->if_softc;
	struct ieee80211com *ic = &sc->sc_ic;
	int r;

	mtx_lock(&sc->network_lock);

	switch(cmd)
	{
	case SIOCSIFFLAGS:
		if((!(sc->sc_if_flags & IFF_UP)) && (ifp->if_flags & IFF_UP))
			p54u_init(sc);
		sc->sc_if_flags = ifp->if_flags;
		r = 0;
		break;
	default:
		r = ieee80211_ioctl(ic, cmd, data);
		break;
	}

	if((r == ENETRESET) && (ifp->if_flags & IFF_UP) && (ifp->if_drv_flags & IFF_DRV_RUNNING))
	{
		p54u_init(sc);
		r = 0;
	}
	mtx_unlock(&sc->network_lock);
	return r;
}

void p54u_start(struct ifnet *ifp)
{
	struct p54u_softc *sc = ifp->if_softc;
	struct ieee80211com *ic = &sc->sc_ic;
	struct mbuf *m;
	struct ieee80211_node *ni = NULL;
	struct ieee80211_frame *wh;
	struct ether_header *eh;
#ifdef DEBUG_NET
	int i;
#endif

	unsigned int size;
	uint32_t device_headroom;
	uint32_t address;

	mtx_lock(&sc->network_lock);

	P54U_DBG_NET("start");

	while(1) {
		/*
		* Poll the management queue for frames; they
		* have priority over normal data frames.
		*/
		IF_POLL(&ic->ic_mgtq, m);
		if(m == NULL) {
			/* No MGMT frames. Look for normal frames. Should only be sent in RUN state */
			if(ic->ic_state != IEEE80211_S_RUN)
				break;
			/* Try to dequeue a data frame to send. If there is no such packet, return. */
			IFQ_DRV_DEQUEUE(&ifp->if_snd, m);
			if(m == NULL)
				break;

			P54U_DBG_NET("802.11 DATA frame TX");

#ifdef WITH_SOFTMAC
			if(P54U_IS_SOFTMAC(sc))
				device_headroom = P54U_LMAC_HEADROOM;
#endif
#ifdef WITH_FREEMAC
			if(P54U_IS_FREEMAC(sc))
				device_headroom = P54U_FMAC_HEADROOM;
#endif

			size = p54u_lmac_80211tx_size(m->m_pkthdr.len + roundup(sizeof(struct ieee80211_qosframe), sizeof(uint32_t))); /* Allocate maximum size */
			address = p54u_pmalloc(&sc->pmalloc_state, size+P54U_LMAC_TAILROOM, device_headroom, P54U_PMALLOC_PURPOSE_80211DATA);
			if(address == P54U_PMALLOC_FAILURE) {
				P54U_DBG_NET("No Prism memory space for DATA frame");
				IFQ_DRV_PREPEND(&ifp->if_snd, m);
				ifp->if_drv_flags |= IFF_DRV_OACTIVE;
				break;
			}

			P54U_DBG_NET("Obtained space for DATA frame at 0x%08x", address);


			/* Find the node for the destination so we can do things like power save and fast frames aggregation. */
			if(m->m_len < sizeof(struct ether_header) && (m = m_pullup(m, sizeof(struct ether_header))) == NULL) {
				ic->ic_stats.is_tx_nobuf++;
				ifp->if_oerrors++;
				break;
			}
			eh = mtod(m, struct ether_header *);
			ni = ieee80211_find_txnode(ic, eh->ether_dhost);
			if(ni == NULL) {
				P54U_ERR("find_txnode failed");
				m_freem(m);
				ifp->if_oerrors++;
				break;
			}

			if(ieee80211_classify(ic, m, ni)) {
				P54U_ERR("classification failure");
				m_freem(m);
				ifp->if_oerrors++;
				ieee80211_free_node(ni);
				break;
			}

			ifp->if_opackets++;
#ifdef WITH_BPF
			BPF_MTAP(ifp, m);
#endif

			m = ieee80211_encap(ic, m, ni);
			if(m == NULL) {
				P54U_ERR("encapsulation failure");
				ifp->if_oerrors++;
				ieee80211_free_node(ni);
				continue;
			}

			wh = mtod(m, struct ieee80211_frame *);

#ifdef WITH_BPF
			if(ic->ic_rawbpf != NULL)
				bpf_mtap(ic->ic_rawbpf, m);
#endif

			if(wh->i_fc[1] & IEEE80211_FC1_WEP) {
				struct ieee80211_key *k;

				k = ieee80211_crypto_encap(ic, ni, m);
				if(k == NULL) {
					P54U_ERR("crypto failure");
					ifp->if_oerrors++;
					m_freem(m);
					ieee80211_free_node(ni);
					continue;
				}
			}


#ifdef WITH_BPF
			if(sc->sc_drvbpf != NULL) {
				struct p54u_tx_radiotap_header *tap = &sc->sc_txtap;

				tap->wt_flags = 0;
				tap->wt_chan_freq = htole16(ic->ic_curchan->ic_freq);
				tap->wt_chan_flags = htole16(IEEE80211_CHAN_G);

				bpf_mtap2(sc->sc_drvbpf, tap, sc->sc_txtap_len, m);
			}
#endif

			ieee80211_free_node(ni);

#ifdef WITH_SOFTMAC
			if(P54U_IS_SOFTMAC(sc))
				m = p54u_lmac_80211tx(sc, address, m, 0x10, 0x4);
#endif
#ifdef WITH_FREEMAC
			if(P54U_IS_FREEMAC(sc)) {
				/* TODO */
				m_freem(m);
				m = NULL;
			}
#endif

			if(m == NULL) {
				P54U_WARN("Unable to generate device frame for 802.11 DATA TX");
				continue;
			}

			if(m->m_pkthdr.len > size)
				P54U_WARN("BUG: packet larger than expected (%d > %d)", m->m_pkthdr.len, size);

			p54u_enqueue_frame(sc, m, 0, address);
			P54U_DBG_NET("DATA frame sent to device");

			p54u_led_event(sc);

		} else {
			P54U_DBG_NET("802.11 MGMT frame TX");

			/* Check that we will be able to send this frame */
#ifdef WITH_SOFTMAC
			if(P54U_IS_SOFTMAC(sc))
				device_headroom = P54U_LMAC_HEADROOM;
#endif
#ifdef WITH_FREEMAC
			if(P54U_IS_FREEMAC(sc))
				device_headroom = P54U_FMAC_HEADROOM;
#endif

			size = p54u_lmac_80211tx_size(m->m_pkthdr.len);
			address = p54u_pmalloc(&sc->pmalloc_state, size+P54U_LMAC_TAILROOM, device_headroom, P54U_PMALLOC_PURPOSE_80211MGMT);
			if(address == P54U_PMALLOC_FAILURE) {
				P54U_DBG_NET("No memory space for MGMT frame");
				ifp->if_drv_flags |= IFF_DRV_OACTIVE;
				break;
			}

			P54U_DBG_NET("Obtained space for MGMT frame at 0x%08x", address);

			/* We'll only fail on hard error condition past this point */

			/* We can remove the frame from the queue */
			IF_DEQUEUE(&ic->ic_mgtq, m);

			/* Pass the frame to possible BPF listeners */
#ifdef WITH_BPF
			if(ic->ic_rawbpf != NULL)
				bpf_mtap(ic->ic_rawbpf, m);
			if(sc->sc_drvbpf != NULL) {
				struct p54u_tx_radiotap_header *tap = &sc->sc_txtap;

				tap->wt_flags = 0;
				tap->wt_chan_freq = htole16(ic->ic_curchan->ic_freq);
				tap->wt_chan_flags = htole16(ic->ic_curchan->ic_flags);

				bpf_mtap2(sc->sc_drvbpf, tap, sc->sc_txtap_len, m);
			}
#endif
			/* FIXME: probe requests */
			if(ic->ic_state == IEEE80211_S_SCAN) {
				P54U_DBG_GEN("Not sending frame (probe request)");
				p54u_pmfree(&sc->pmalloc_state, address);
				m_freem(m);
				continue;
			}

#ifdef WITH_SOFTMAC
			if(P54U_IS_SOFTMAC(sc))
				m = p54u_lmac_80211tx(sc, address, m, 0x00, 0x4);
#endif
#ifdef WITH_FREEMAC

			if(P54U_IS_FREEMAC(sc)) {
				/* TODO */
				m_freem(m);
				m = NULL;
			}
#endif

			if(m == NULL) {
				P54U_WARN("Unable to generate device frame for 802.11 MGMT TX");
				continue;
			}

			p54u_enqueue_frame(sc, m, 0, address);
			P54U_DBG_NET("MGMT frame sent to device");
		}
	}
	mtx_unlock(&sc->network_lock);
}

static void p54u_next_scan(void *arg)
{
	struct p54u_softc *sc = arg;

	/* callout will be reset in the state transition function */
	if(sc->sc_ic.ic_state == IEEE80211_S_SCAN)
		ieee80211_next_scan(&sc->sc_ic);
}

static int p54u_beacon_alloc(struct p54u_softc *sc, struct ieee80211_node *ni)
{
	struct ieee80211com *ic = &sc->sc_ic;
	struct mbuf *m;
	unsigned int size;
	uint32_t address;

	m = ieee80211_beacon_alloc(ic, ni, &sc->sc_boff);
	if(m == NULL) {
		P54U_ERR("Unable to generate beacon frame");
		return 0;
	}

#ifdef WITH_SOFTMAC
	if(P54U_IS_SOFTMAC(sc))
	{
		size = p54u_lmac_80211tx_size(m->m_pkthdr.len);
		address = p54u_pmalloc(&sc->pmalloc_state, size+P54U_LMAC_TAILROOM, P54U_LMAC_HEADROOM, P54U_PMALLOC_PURPOSE_80211BEACON);
		P54U_DBG_NET("Allocated beacon at address 0x%08x", address);
		if(address == P54U_PMALLOC_FAILURE) {
			P54U_WARN("Unable to allocate Prism memory space for beacon");
			m_freem(m);
			return 0;
		}
		m = p54u_lmac_80211tx(sc, address, m, P54U_LMAC_ID1_FLAGS_BEACON, P54U_FRAME_BEACON);
	}
#endif
#ifdef WITH_FREEMAC
	if(P54U_IS_FREEMAC(sc))
	{
		/* TODO */
		m_freem(m);
		m = NULL;
	}
#endif

	if(m == NULL)
	{
		P54U_WARN("Unable to generate beacon frame");
		return 0;
	}
	p54u_enqueue_frame(sc, m, 0, address);
	P54U_DBG_NET("Beacon sent to device");

	return 1;
}

static int p54u_newstate(struct ieee80211com *ic, enum ieee80211_state nstate, int arg)
{
	struct ifnet *ifp = ic->ic_ifp;
	struct p54u_softc *sc = ifp->if_softc;
	int r;
	uint32_t address;
	struct mbuf *m;

	mtx_lock(&sc->network_lock);

	P54U_DBG_NET("new state: %s -> %s", ieee80211_state_name[ic->ic_state], ieee80211_state_name[nstate]);

	p54u_update_leds(sc, nstate);

	/* Stop beacon (if any) */
	address = p54u_pmfind(&sc->pmalloc_state, P54U_PMALLOC_PURPOSE_80211BEACON);
	if(address != P54U_PMALLOC_FAILURE)
	{
#ifdef WITH_SOFTMAC
		if(P54U_IS_SOFTMAC(sc))
			m = p54u_lmac_abortaddress(sc, sc->control_address, address);
#endif
#ifdef WITH_FREEMAC

		if(P54U_IS_FREEMAC(sc))
			m = p54u_fmac_abortaddress(sc, sc->control_address, address);
#endif

		if(m != NULL)
			p54u_enqueue_frame(sc, m, 0, sc->control_address);
		else
			P54U_WARN("Unable to allocate memory to free beacon");
	}

	/* Reconfigure the device */
#ifdef WITH_SOFTMAC
	if(P54U_IS_SOFTMAC(sc))
		p54u_lmac_configure(sc, nstate);
#endif
#ifdef WITH_FREEMAC

	if(P54U_IS_FREEMAC(sc))
		p54u_fmac_configure(sc, nstate);
#endif

	/* Setup beacon frame for adhoc or hostap mode. */
	if((nstate == IEEE80211_S_RUN) && ((ic->ic_opmode == IEEE80211_M_IBSS) | (ic->ic_opmode == IEEE80211_M_HOSTAP)))
		p54u_beacon_alloc(sc, ic->ic_bss);

	/* Invoke the parent method to complete the work */
	r = (*sc->sc_newstate)(ic, nstate, arg);

	mtx_unlock(&sc->network_lock);

	if(nstate == IEEE80211_S_SCAN)
		callout_reset(&sc->sc_scan_ch, (P54U_SCAN_DELAY*hz)/1000, p54u_next_scan, sc);
	if(nstate == IEEE80211_S_INIT)
		callout_drain(&sc->sc_scan_ch);

	return r;
}

static void p54u_watchdog(struct ifnet *ifp)
{
	struct p54u_softc *sc = ifp->if_softc;
	struct ieee80211com *ic = &sc->sc_ic;

	mtx_lock(&sc->network_lock);
	ieee80211_watchdog(ic);
	mtx_unlock(&sc->network_lock);
}

static int p54u_media_change(struct ifnet *ifp)
{
	struct p54u_softc *sc = ifp->if_softc;
	int r;

	mtx_lock(&sc->network_lock);

	P54U_DBG_NET("media change");
	r = ieee80211_media_change(ifp);
	if(r == ENETRESET)
	{
		p54u_init(sc);
		r = 0;
	}

	mtx_unlock(&sc->network_lock);

	return r;
}

static void p54u_media_status(struct ifnet *ifp, struct ifmediareq *imr)
{
	struct p54u_softc *sc = ifp->if_softc;
	struct ieee80211com *ic = &sc->sc_ic;

	imr->ifm_status = IFM_AVALID;
	imr->ifm_active = IFM_IEEE80211;
	if(ic->ic_state == IEEE80211_S_RUN)
		imr->ifm_status |= IFM_ACTIVE;

	imr->ifm_active |= ieee80211_rate2media(ic, sc->rxrate, IEEE80211_MODE_11G);
	switch (ic->ic_opmode)
	{
	case IEEE80211_M_STA:
		break;
	case IEEE80211_M_IBSS:
		imr->ifm_active |= IFM_IEEE80211_IBSS;
		break;
	case IEEE80211_M_AHDEMO: // unsupported
		imr->ifm_active |= IFM_IEEE80211_IBSS | IFM_FLAG0;
		break;
	case IEEE80211_M_HOSTAP:
		imr->ifm_active |= IFM_IEEE80211_HOSTAP;
		break;
	case IEEE80211_M_MONITOR:
		imr->ifm_active |= IFM_IEEE80211_MONITOR;
		break;
	}
	imr->ifm_active |= IFM_IEEE80211_11G; // TODO: real b/g
}

static int p54u_key_alloc(struct ieee80211com *ic, const struct ieee80211_key *k, ieee80211_keyix *keyix, ieee80211_keyix *rxkeyix)
{
	struct p54u_softc *sc = ic->ic_ifp->if_softc;

	/* Force software encryption */
	/* The chipset can do hardware encryption, reverse engineer it if you're interested... */
	if(!(k->wk_flags & IEEE80211_KEY_SWCRYPT))
		return 0;
	return sc->sc_key_alloc(ic, k, keyix, rxkeyix);
}

static void p54u_setup_freqs(struct p54u_softc *sc)
{
	struct ieee80211com *ic = &sc->sc_ic;
	unsigned int i;

	for(i=1;i<P54U_NUM_CHANNELS+1;i++)
	{
		ic->ic_channels[i].ic_freq = ieee80211_ieee2mhz(i, IEEE80211_CHAN_2GHZ);
		ic->ic_channels[i].ic_flags = IEEE80211_MODE_11B|IEEE80211_MODE_11G;
		P54U_DBG_NET("Set up channel IEEE %d, freq %dMHz", i, ic->ic_channels[i].ic_freq);
	}
}

static int p54u_setup_rates(struct p54u_softc *sc, unsigned int mode)
{
	struct ieee80211com *ic = &sc->sc_ic;
	struct ieee80211_rateset *rs;
	int i, maxrates;

	/* FIXME: should check mode */
	rs = &ic->ic_sup_rates[mode];
	maxrates = sc->rate_count;
	if(maxrates > IEEE80211_RATE_MAXSIZE)
	{
		P54U_DBG_NET("802.11 stack rate table too small");
		maxrates = IEEE80211_RATE_MAXSIZE;
	}
	for(i=0;i<maxrates;i++)
		rs->rs_rates[i] = sc->rate_table[i] & IEEE80211_RATE_VAL;
	rs->rs_nrates = maxrates;

	return 0;
}

int p54u_netif_init(struct p54u_softc *sc)
{
	struct ieee80211com *ic = &sc->sc_ic;
	struct ifnet *ifp;

	P54U_DBG_NET("Setting up the network interface");

	sc->leds_auto = 1;
	sc->rxrate = 108;

	callout_init(&sc->sc_scan_ch, CALLOUT_MPSAFE);

	ifp = ic->ic_ifp = if_alloc(IFT_ETHER);
	if(ifp == NULL)
	{
		P54U_ERR("Unable to create network interface");
		return 0;
	}

	ifp->if_softc = sc;
	if_initname(ifp, "p54u", p54u_devunit(sc));

	ic->ic_phytype = IEEE80211_T_DS;
	ic->ic_opmode = IEEE80211_M_STA;
	ic->ic_state = IEEE80211_S_INIT;

	/* set device capabilities */
	ic->ic_caps = IEEE80211_C_MONITOR | IEEE80211_C_IBSS | IEEE80211_C_HOSTAP | IEEE80211_C_WEP | IEEE80211_C_WPA;

	/* set MAC address */
	IEEE80211_ADDR_COPY(&ic->ic_myaddr[0], &sc->dev_addr[0]);

	sc->sc_if_flags = ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	ifp->if_drv_flags = IFF_DRV_RUNNING;
	//ifp->if_flags |= IFF_DEBUG;
	ifp->if_init = p54u_init;
	ifp->if_ioctl = p54u_ioctl;
	ifp->if_start = p54u_start;
	ifp->if_watchdog = p54u_watchdog;
	IFQ_SET_MAXLEN(&ifp->if_snd, IFQ_MAXLEN);
	ifp->if_snd.ifq_drv_maxlen = IFQ_MAXLEN;
	IFQ_SET_READY(&ifp->if_snd);
	ic->ic_ifp = ifp;

	p54u_setup_freqs(sc);
	p54u_setup_rates(sc, IEEE80211_MODE_11B);
	p54u_setup_rates(sc, IEEE80211_MODE_11G);

	/* Indicate we need the 802.11 header padded to a 32-bit boundary for 4-address and QoS frames. */
	ic->ic_flags |= IEEE80211_F_DATAPAD;
	ic->ic_state = IEEE80211_S_INIT;

	ieee80211_ifattach(ic);

	/* override state transition machine */
	sc->sc_newstate = ic->ic_newstate;
	ic->ic_newstate = p54u_newstate;
	sc->sc_key_alloc = ic->ic_crypto.cs_key_alloc;
	ic->ic_crypto.cs_key_alloc = p54u_key_alloc;

	ieee80211_media_init(ic, p54u_media_change, p54u_media_status);

#ifdef WITH_BPF

	P54U_DBG_NET("Enabling BPF");
	bpfattach2(ifp, DLT_IEEE802_11_RADIO, sizeof(struct ieee80211_frame)+64, &sc->sc_drvbpf);

	sc->sc_rxtap_len = sizeof(sc->sc_rxtapu);
	sc->sc_rxtap.wr_ihdr.it_len = htole16(sc->sc_rxtap_len);
	sc->sc_rxtap.wr_ihdr.it_present = htole32(P54U_RX_RADIOTAP_PRESENT);

	sc->sc_txtap_len = sizeof(sc->sc_txtapu);
	sc->sc_txtap.wt_ihdr.it_len = htole16(sc->sc_txtap_len);
	sc->sc_txtap.wt_ihdr.it_present = htole32(P54U_TX_RADIOTAP_PRESENT);
#endif

	P54U_DBG_NET("Network interface created");

	return 1;
}

int p54u_netif_free(struct p54u_softc *sc)
{
	struct ieee80211com *ic = &sc->sc_ic;
	struct ifnet *ifp = ic->ic_ifp;

	P54U_DBG_NET("Shutting down 802.11 stack");
	ieee80211_new_state(ic, IEEE80211_S_INIT, -1);
#ifdef WITH_BPF

	bpfdetach(ic->ic_ifp);
#endif

	ieee80211_ifdetach(ic);
	if_free(ifp);
	return 1;
}
