/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "config.h"

#include <sys/types.h>
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/queue.h>
#include <sys/taskqueue.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/lock.h>
#include <sys/mutex.h>

#include "firmware.h"
#ifdef WITH_SOFTMAC
#include "softmac.h"
#endif
#ifdef WITH_FREEMAC
#include "freemac.h"
#endif
#include "isl38xx_common.h"
#include "output_layer.h"
#include "p54u.h"

#ifdef WITH_USB

static void p54u_usb_output_task(void *context, int pending)
{
	struct p54u_softc *sc = (struct p54u_softc *)context;
	struct p54u_output_frame *bf;

	while(1) {
		mtx_lock(&sc->bus.usb.output_lock);
		bf = TAILQ_FIRST(&sc->bus.usb.output_queue);
		if(bf == NULL) {
			mtx_unlock(&sc->bus.usb.output_lock);
			break;
		}
		TAILQ_REMOVE(&sc->bus.usb.output_queue, bf, bf_list);
		mtx_unlock(&sc->bus.usb.output_lock);

		p54u_tx_lock(sc);
		mtx_lock(&Giant);
#ifdef WITH_SOFTMAC
		if(P54U_IS_SOFTMAC(sc))
			p54u_lmac_usb_transport(sc, bf->m, bf->address);
#endif
#ifdef WITH_FREEMAC
		if(P54U_IS_FREEMAC(sc))
			p54u_fmac_usb_transport(sc, bf->m, bf->address);
#endif

		mtx_unlock(&Giant);
		p54u_tx_unlock(sc);

		free(bf, M_P54U);
	}
}

#endif

int p54u_enqueue_frame(struct p54u_softc *sc, struct mbuf *m, int hipri, uint32_t address)
{
#ifdef WITH_USB
	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_USB2)) {
		struct p54u_output_frame *bf;
	
		bf = malloc(sizeof(struct p54u_output_frame), M_P54U, M_NOWAIT);
		if(bf == NULL) {
			m_freem(m);
			return 0;
		}
	
		bf->m = m;
		bf->address = address;
	
		mtx_lock(&sc->bus.usb.output_lock);
		TAILQ_INSERT_TAIL(&sc->bus.usb.output_queue, bf, bf_list);
		taskqueue_enqueue(taskqueue_thread, &sc->bus.usb.output_task);
		mtx_unlock(&sc->bus.usb.output_lock);
	}
#endif
#ifdef WITH_PCI
	if(sc->device_type == P54U_DEVTYPE_PCI) {
		int ok;
		
		p54u_tx_lock(sc);
		ok = p54u_queue_submit(sc, hipri ? &sc->bus.pci.data_tx_hipri : &sc->bus.pci.data_tx, m, address);
		p54u_tx_unlock(sc);
		if(!ok) return 0;
	}
#endif

	return 1;
}

#ifdef WITH_USB

int p54u_output_init(struct p54u_softc *sc)
{
	P54U_DBG_GEN("Setting up output layer");
	mtx_init(&sc->bus.usb.output_lock, "p54u", "out q", MTX_DEF);
	TAILQ_INIT(&sc->bus.usb.output_queue);
	TASK_INIT(&sc->bus.usb.output_task, 0, p54u_usb_output_task, sc);
	return 1;
}

/* You must make sure that p54u_enqueue_frame() will never be called again once this function has been called */
int p54u_output_free(struct p54u_softc *sc)
{
	struct p54u_output_frame *bf;
	
	P54U_DBG_GEN("Shutting down output layer");
	/* Drain the output queue */
	mtx_lock(&sc->bus.usb.output_lock);
	while(1) {
		bf = TAILQ_FIRST(&sc->bus.usb.output_queue);
		if(bf == NULL)
			break;
		TAILQ_REMOVE(&sc->bus.usb.output_queue, bf, bf_list);
		m_freem(bf->m);
		free(bf, M_P54U);
	}
	mtx_unlock(&sc->bus.usb.output_lock);

	mtx_destroy(&sc->bus.usb.output_lock);
	return 1;
}
#endif
