/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
  Copyright (C) 2004 Feyd
  Copyright (C) 2004, 2005 Jean-Baptiste Note <jean-baptiste.note@m4x.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "config.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/sema.h>
#include <sys/mutex.h>
#include <sys/interrupt.h>
#include <sys/time.h>
#include <sys/endian.h>
#include <sys/socket.h>
#include <sys/mbuf.h>
#include <sys/selinfo.h>
#include <sys/sysctl.h>
#include <sys/rman.h>

#ifdef WITH_USB
#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>
#endif

#ifdef WITH_PCI
#include <machine/bus.h>
#include <machine/resource.h>
#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>
#endif

#include <net/bpf.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>
#include <net/if_types.h>

#include <net80211/ieee80211_var.h>
#include <net80211/ieee80211_proto.h>
#include <net80211/ieee80211_ioctl.h>
#ifdef WITH_BPF
#include <net80211/ieee80211_radiotap.h>
#endif

#include "isl38xx_common.h"
#include "isl38xx_host.h"
#ifdef WITH_SOFTMAC
#include "softmac.h"
#endif
#ifdef WITH_FREEMAC
#include "freemac.h"
#endif
#include "pcidefs.h"
#include "transport.h"
#include "pmalloc.h"
#include "output_layer.h"
#include "devnode.h"
#include "p54u.h"

MALLOC_DEFINE(M_P54U, "p54u", "Buffers for the p54u driver");

void p54u_delay(int ms)
{
	tsleep(&ms, PZERO, "p54ud", (ms*hz+999)/1000+1);
}

int debug_gen;
int debug_dumpxfers;
int debug_net;

#if ((defined WITH_DEBUG_GEN) || (defined WITH_DEBUG_NET))
SYSCTL_NODE(_hw, OID_AUTO, p54u, CTLFLAG_RW, 0, "Prism54 devices");
#endif

#ifdef WITH_DEBUG_GEN
SYSCTL_INT(_hw_p54u, OID_AUTO, debug_gen, CTLFLAG_RW, &debug_gen, 0, "General debugging");
SYSCTL_INT(_hw_p54u, OID_AUTO, debug_dumpxfers, CTLFLAG_RW, &debug_dumpxfers, 0, "Dump all transfers");
#endif
#ifdef WITH_DEBUG_NET
SYSCTL_INT(_hw_p54u, OID_AUTO, debug_net, CTLFLAG_RW, &debug_net, 0, "Debug network interface");
#endif

struct p54u_softc *p54u_devices[P54U_MAX_DEVICES] = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, };

int p54u_alloc_device(struct p54u_softc *sc)
{
	int i;
	
	for(i=0;i<P54U_MAX_DEVICES;i++) {
		if(p54u_devices[i] == NULL) {
			p54u_devices[i] = sc;
			return i;
		}
	}
	return -1;
}

int p54u_devunit(struct p54u_softc *sc)
{
	int i;
	
	for(i=0;i<P54U_MAX_DEVICES;i++) {
		if(p54u_devices[i] == sc) return i;
	}
	return -1;
}

struct p54u_softc *p54u_get_sc(int unit)
{
	if((unit < 0) || (unit >= P54U_MAX_DEVICES)) return NULL;
	return p54u_devices[unit];
}

void p54u_free_device(int unit)
{
	p54u_devices[unit] = NULL;
}

#if ((defined WITH_DEBUG_GEN) || (defined WITH_DEBUG_NET))

void p54u_data_debug(struct p54u_softc *sc, unsigned int ep, const void *data, int len)
{
	int k, l, in;
	struct timeval t;
	unsigned long int time;
	char linebuffer[256]; /* We send full lines at each printf(), otherwise syslog gets confused */
	char *currentpos;

	microtime(&t);
	timevalsub(&t, &sc->start_time);
	time = t.tv_sec*1000+t.tv_usec/1000;
#ifdef WITH_USB
	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_USB2)) in = (ep & UE_DIR_IN) != 0;
#endif
#ifdef WITH_PCI
	if(sc->device_type == P54U_DEVTYPE_PCI) in = (ep == 2);
#endif
	for(k=0;k<len;k+=16) {
		currentpos = &linebuffer[0];
		currentpos += sprintf(currentpos, "p54u%d: ", p54u_devunit(sc));
		if(k == 0) {
			if(ep != 0) {
#ifdef WITH_USB
				if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_USB2)) currentpos += sprintf(currentpos, "[%s %lums ep %02x]\np54u%d: ", in ? "IN " : "OUT",  time, ep, p54u_devunit(sc));
#endif
#ifdef WITH_PCI
				if(sc->device_type == P54U_DEVTYPE_PCI) currentpos += sprintf(currentpos, "[%s %lums]\np54u%d: ", in ? "IN " : "OUT",  time, p54u_devunit(sc));
#endif
			}
			else
				currentpos += sprintf(currentpos, "[%lums]\np54u%d: ", time, p54u_devunit(sc));
		}
		currentpos += sprintf(currentpos, "%08x:", k);

		for(l=0;(l<16)&&(k+l<len);++l)
			currentpos += sprintf(currentpos, " %02x", ((const unsigned char *)data)[k+l]);

		printf("%s\n", &linebuffer[0]);
	}
}

#endif

static int p54u_init_led(struct p54u_softc *sc)
{
	struct mbuf *m;

#ifdef WITH_SOFTMAC
	if(P54U_IS_SOFTMAC(sc))
		m = p54u_lmac_leds(sc, sc->control_address, P54U_LMAC_LED_MODE_SET, 0, 0, 0);
#endif
#ifdef WITH_FREEMAC
	if(P54U_IS_FREEMAC(sc))
		m = p54u_fmac_leds(sc, P54U_FMAC_LED_MODE_SET, 0, 0, 0);
#endif

	if(m == NULL)
		return 0;
	p54u_enqueue_frame(sc, m, 0, sc->control_address);
	return 1;
}

/* Bring the device up to the point where we have a running firmware, */
/* all protocols initialized, PDA read and parsed, and network interface created */
int p54u_boot(struct p54u_softc *sc, char *fw, unsigned int size)
{
	/* resources freed on USB detach : we can just return from the function in case of failure */
	P54U_DBG_GEN("Booting the device");

	if(sc->firmware_running) {
		P54U_DBG_GEN("Firmware already running");
		return 0;
	}
	if(sc->first_boot) {
		if(!p54u_initialize_device(sc))
			return 0;
	}
	sc->first_boot = 0;
	if(((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_PCI)) && (!sc->irq_initialized))	{
		/* Set up IRQs now, because we need them to check firmware startup */
		if(!p54u_irq_init(sc))
			return 0;
		sc->irq_initialized = 1;
	}

	if(!p54u_init_firmware(sc, fw, size))
		return 0;

	/* Ok, the firmware's running and the hardware's ready. */

	/* Set up UARTPCI right now, so that we can debug */
	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_PCI))
		p54u_uart_start(sc);

	/* Protocol initialization */
#ifdef WITH_SOFTMAC
	if(P54U_IS_SOFTMAC(sc))
		p54u_lmac_init(sc);
#endif
#ifdef WITH_FREEMAC
	if(P54U_IS_FREEMAC(sc))
		p54u_fmac_init(sc);
#endif

	sc->protocol_initialized = 1;

	/* We're now ready to accept frames from the device */
#ifdef WITH_USB
	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_USB2)) {
		P54U_DBG_GEN("Submitting initial DATA RX xfer");
		if(!p54u_data_rx_submit(sc))
			return 0;
	}
#endif

#ifdef WITH_FREEMAC
	/* FIXME */
	if(P54U_IS_FREEMAC(sc))	{
		sc->frame_space_start = 0x00022000;
		sc->frame_space_end = 0x00027000;
	}
#endif

	/* Set up host->firmware communication */
#ifdef WITH_SOFTMAC
	if(P54U_IS_SOFTMAC(sc))	{
		sc->frame_space_end -= P54U_LMAC_RESERVED_SIZE+1;
		sc->reserved_address = sc->frame_space_end+1;
	}
#endif
	P54U_DBG_GEN("Reserved space is at 0x%08x", sc->reserved_address);

	P54U_DBG_GEN("Setting up Prism memory allocator (0x%08x-0x%08x)", sc->frame_space_start, sc->frame_space_end);
	p54u_pmstate_init(&sc->pmalloc_state, sc->frame_space_start, sc->frame_space_end);
	sc->pmalloc_state_initialized = 1;

#ifdef WITH_SOFTMAC
	if(P54U_IS_SOFTMAC(sc))
		sc->control_size = P54U_LMAC_CONTROL_BUFFER_SIZE;
#endif
#ifdef WITH_FREEMAC
	if(P54U_IS_FREEMAC(sc))
		sc->control_size = P54U_FMAC_CONTROL_BUFFER_SIZE;
#endif

	sc->control_address = p54u_pmalloc(&sc->pmalloc_state, sc->control_size, 0, P54U_PMALLOC_PURPOSE_CONTROL);
	if(sc->control_address == P54U_PMALLOC_FAILURE)
	{
		P54U_ERR("Unable to allocate space for control frames (corrupted or incompatible firmware ?)");
		return 0;
	}
	P54U_DBG_GEN("Allocated space for control frames at 0x%08x", sc->control_address);

#ifdef WITH_USB
	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_USB2)) {
		if(!p54u_output_init(sc))
			return 0;
		sc->bus.usb.output_initialized = 1;
	}
#endif
#ifdef WITH_PCI
	if(sc->device_type == P54U_DEVTYPE_PCI) {
		uint32_t reg;
		
		if(!p54u_alloc_queues(sc)) return 0;
		sc->bus.pci.queues_initialized = 1;

		/* Catch UPDATE interrupt */
		reg = p54u_dev_readl(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_EN_REG);
		reg |= ISL38XX_INT_UPDATE;
		p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_EN_REG, reg);
	}
#endif

	P54U_DBG_GEN("Hardware and protocols ready");

	if((sc->firmware_type == P54U_FWTYPE_FMAC) && (sc->device_type == P54U_DEVTYPE_PCI)) return 1;
	if(!p54u_eeprom_read(sc))
		return 0;
	sc->eeprom_available = 1;
	if(!p54u_init_led(sc))
		return 0;
	if(!p54u_netif_init(sc))
		return 0;
	sc->if_initialized = 1;

	P54U_DBG_GEN("Device boot completed");

	return 1;
}

static int p54u_init(struct p54u_softc *sc)
{
	sx_init(&sc->tx_sx, "tx_sx");
	sema_init(&sc->init_sema, 0, "init_sema");
	sema_init(&sc->eeprom_read_sema, 0, "eeprom_read_sema");
	sema_init(&sc->stats_read_sema, 0, "stats_read_sema");
	mtx_init(&sc->network_lock, "net_mtx", MTX_NETWORK_LOCK, MTX_DEF|MTX_RECURSE);
	sc->sync_initialized = 1;

#ifdef WITH_USB
	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_USB2)) {
		if(!p54u_alloc_standard_pipes(sc))
			return 0;
	}
#endif

	if(!p54u_devnode_init(sc))
		return 0;
	sc->devnode_initialized = 1;

	sc->first_boot = 1;

	return 1;
}

static int p54u_shutdown(struct p54u_softc *sc)
{
	P54U_DBG_GEN("Terminating the driver");

	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_PCI)) {
		p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_DEV_INT_REG, ISL38XX_INT_ABORT);
		p54u_delay(50);
	}
	
#ifdef WITH_USB
	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_USB2)) usbd_abort_pipe(sc->bus.usb.data_rx.handle);
#endif
	if(sc->irq_initialized) {
		p54u_irq_free(sc);
		sc->irq_initialized = 0;
	}

	if(sc->if_initialized) {
		p54u_netif_free(sc);
		sc->if_initialized = 0;
	}

	if(sc->protocol_initialized) {
#ifdef WITH_SOFTMAC
		if(P54U_IS_SOFTMAC(sc))
			p54u_lmac_free(sc);
#endif
#ifdef WITH_FREEMAC
		if(P54U_IS_FREEMAC(sc))
			p54u_fmac_free(sc);
#endif

		sc->protocol_initialized = 0;
	}

	if(sc->pmalloc_state_initialized) {
		P54U_DBG_GEN("Shutting down Prism memory allocator");
		p54u_pmstate_free(&sc->pmalloc_state);
		sc->pmalloc_state_initialized = 0;
	}

#ifdef WITH_USB
	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_USB2)) {
		if(sc->bus.usb.output_initialized) {
			p54u_output_free(sc);
			sc->bus.usb.output_initialized = 0;
		}
	}
#endif

#ifdef WITH_PCI
	if(sc->device_type == P54U_DEVTYPE_PCI) {
		if(sc->bus.pci.queues_initialized) {
			P54U_DBG_GEN("Cleaning up DMA buffers");
			p54u_free_queues(sc);
			sc->bus.pci.queues_initialized = 0;
		}
	}
#endif
	
	if(sc->devnode_initialized) {
		p54u_devnode_free(sc);
		sc->devnode_initialized = 0;
	}

#ifdef WITH_USB
	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_USB2))
		p54u_free_standard_pipes(sc);
#endif

	sc->firmware_running = 0;
	sc->eeprom_available = 0;

	P54U_DBG_GEN("Cleaning up");
	free(sc->crash_data, M_P54U);
	if(sc->sync_initialized) {
		mtx_destroy(&sc->network_lock);
		sema_destroy(&sc->stats_read_sema);
		sema_destroy(&sc->eeprom_read_sema);
		sema_destroy(&sc->init_sema);
		sx_destroy(&sc->tx_sx);
	}
	return 1;
}

int p54u_reset(struct p54u_softc *sc)
{
	P54U_DBG_GEN("Performing warm reset of the device");

	/* Clear error condition : the reason for reset may be a previous fatal error */
	sc->err = 0;

	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_PCI)) {
		p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_DEV_INT_REG, ISL38XX_INT_ABORT);
		p54u_delay(50);
	}
	
	/* Abort DATA and INT xfers now */
#ifdef WITH_USB
	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_USB2))
		if(sc->bus.usb.data_rx.handle != 0) usbd_abort_pipe(sc->bus.usb.data_rx.handle);
#endif
	if(sc->irq_initialized) {
		p54u_irq_free(sc);
		sc->irq_initialized = 0;
	}

	/* Shutdown software */
	if(sc->if_initialized) {
		p54u_netif_free(sc);
		sc->if_initialized = 0;
	}

	if(sc->protocol_initialized) {
#ifdef WITH_SOFTMAC
		if(P54U_IS_SOFTMAC(sc))
			p54u_lmac_free(sc);
#endif
#ifdef WITH_FREEMAC
		if(P54U_IS_FREEMAC(sc))
			p54u_fmac_free(sc);
#endif

		sc->protocol_initialized = 0;
	}

	if(sc->pmalloc_state_initialized) {
		P54U_DBG_GEN("Shutting down Prism memory allocator");
		p54u_pmstate_free(&sc->pmalloc_state);
		sc->pmalloc_state_initialized = 0;
	}

#ifdef WITH_USB
	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_USB2)) {
		if(sc->bus.usb.output_initialized) {
			p54u_output_free(sc);
			sc->bus.usb.output_initialized = 0;
		}
	}
#endif
	free(sc->crash_data, M_P54U);
	sc->crash_data = NULL;
	
#ifdef WITH_PCI
	if(sc->device_type == P54U_DEVTYPE_PCI) {
		if(sc->bus.pci.queues_initialized) {
			P54U_DBG_GEN("Cleaning up DMA buffers");
			p54u_free_queues(sc);
			sc->bus.pci.queues_initialized = 0;
		}
	}
#endif

#ifdef WITH_USB
	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_USB2)) {
		/* Free and open the endpoints again to reset them */
		P54U_DBG_GEN("Resetting endpoints");
		p54u_free_standard_pipes(sc);
		if(!p54u_alloc_standard_pipes(sc))
			return 0;
	}
#endif

	if(!p54u_initialize_device(sc))
		return 0;

	sc->eeprom_available = 0;

	return 1;
}

MODULE_DEPEND(p54u, wlan, 1, 1, 1);

#ifdef WITH_USB

#define p54u_usb_softc p54u_softc
 
static const struct p54u_device_entry p54u_usb_table[] = {
	{0x5041, 0x2234, "Linksys WUSB54G",                           P54U_DEVTYPE_USB1, 0},
	{0x5041, 0x2235, "Linksys WUSB54G Portable",                  P54U_DEVTYPE_USB1, 0},
	{0x1915, 0x2234, "Linksys WUSB54G OEM",                       P54U_DEVTYPE_USB1, 0},
	{0x1915, 0x2235, "Linksys WUSB54G Portable OEM",              P54U_DEVTYPE_USB1, 0},
	{0x0506, 0x0a11, "3COM 3CRWE254G72",                          P54U_DEVTYPE_USB1, 0},
	{0x2001, 0x3701, "DLink DWL-G120",                            P54U_DEVTYPE_USB1, 0},
	{0x2001, 0x3703, "DLink DWL-G122",                            P54U_DEVTYPE_USB1, 0},
	{0x0cde, 0x0006, "Medion 40900, Roper Europe",                P54U_DEVTYPE_USB1, 0},
	{0x0846, 0x4200, "Netgear WG121",                             P54U_DEVTYPE_USB1, 0},
	{0x0846, 0x4210, "Netgear WG121 (alternative version)",       P54U_DEVTYPE_USB1, 0},
	{0x0846, 0x4220, "Netgear WG111",                             P54U_DEVTYPE_USB1, 0},
	{0x0707, 0xee06, "SMC 2862W-G",                               P54U_DEVTYPE_USB1, 0},
	{0x124a, 0x4023, "Shuttle PN15/Airvast WM168g/IOGear GWU513", P54U_DEVTYPE_USB1, 0},
	{0x083a, 0x4502, "Siemens Gigaset USB Adapter 54",            P54U_DEVTYPE_USB1, 0},
	{0x083a, 0x5501, "Philips CPWUA054/00",                       P54U_DEVTYPE_USB1, 0},
	{0x0846, 0x4240, "Netgear WG111 version 2",                   P54U_DEVTYPE_USB2, 1},
	{0x0cde, 0x0008, "Sagem XG703A ",                             P54U_DEVTYPE_USB2, 0},
	{0x09aa, 0x1000, "Spinnaker Proto board",                     P54U_DEVTYPE_USB2, 0},
	{0x413c, 0x8102, "Spinnaker DUT",                             P54U_DEVTYPE_USB2, 0},
	{0x050d, 0x7050, "Belkin F5D7050 ver 1000",                   P54U_DEVTYPE_USB2, 0},
	{0x0915, 0x2000, "Cohiba Proto board",                        P54U_DEVTYPE_USB2, 0},
	{0x0915, 0x2002, "Cohiba Proto board",                        P54U_DEVTYPE_USB2, 0},
	{0x0572, 0x2000, "Cohiba Proto board",                        P54U_DEVTYPE_USB2, 0},
	{0x0572, 0x2002, "Cohiba Proto board",                        P54U_DEVTYPE_USB2, 0},
	{0x413c, 0x8104, "Cohiba Proto board",                        P54U_DEVTYPE_USB2, 0},
	{0x0d8e, 0x3762, "DLink DWL-G120 version 2",                  P54U_DEVTYPE_USB2, 0},
	{0x1435, 0x0427, "Inventel UR054G",                           P54U_DEVTYPE_USB2, 0},
	{0x083a, 0x4521, "Siemens Gigaset USB Adapter 54 version 2",  P54U_DEVTYPE_USB2, 0},
	{0x0707, 0xee13, "SMC 2862W-G version 2",                     P54U_DEVTYPE_USB2, 0},
	{0x0bf8, 0x1009, "Fujitsu E-5400 USB D1700",                  P54U_DEVTYPE_USB2, 0},
	{0x0baf, 0x0118, "U.S. Robotics U5",                          P54U_DEVTYPE_USB2, 0},
	{0x13b1, 0x000a, "Linksys WUSB54G version 2",                 P54U_DEVTYPE_USB2, 0},
};

USB_DECLARE_DRIVER(p54u_usb);

USB_MATCH(p54u_usb)
{
	USB_MATCH_START(p54u_usb, uaa);
	unsigned int i;

	if(uaa->iface == NULL)
		return UMATCH_NONE;

	for(i=0;i<sizeof(p54u_usb_table)/sizeof(p54u_usb_table[0]);i++) {
		if((p54u_usb_table[i].vendor == uaa->vendor) && (p54u_usb_table[i].product == uaa->product))
			return UMATCH_VENDOR_PRODUCT;
	}
	return UMATCH_NONE;
}

USB_ATTACH(p54u_usb)
{
	USB_ATTACH_START(p54u_usb, sc, uaa);
	usbd_interface_handle iface = uaa->iface;
	const char *devinfo;
	unsigned int i;

	bzero(sc, sizeof(struct p54u_softc));
#if ((defined WITH_DEBUG_GEN) || (defined WITH_DEBUG_NET))
	microtime(&sc->start_time);
#endif
	if(p54u_alloc_device(sc) == -1) {
		USB_ATTACH_ERROR_RETURN;
	}

	sc->bus.usb.sc_udev = uaa->device;
	sc->bus.usb.sc_iface = iface;

	for(i=0;i<sizeof(p54u_usb_table)/sizeof(p54u_usb_table[0]);i++)
		if((p54u_usb_table[i].vendor == uaa->vendor) && (p54u_usb_table[i].product == uaa->product))
			break;

	devinfo = &p54u_usb_table[i].name[0];
	sc->device_type = p54u_usb_table[i].type;
	sc->leds_inverted = p54u_usb_table[i].leds_inverted;

#define sc_dev bus.usb.sc_dev
	USB_ATTACH_SETUP;
#undef sc_dev

	if(!p54u_init(sc)) {
		P54U_ERR("Failed to initialize the driver");
		p54u_shutdown(sc);
		USB_ATTACH_ERROR_RETURN;
	}

	USB_ATTACH_SUCCESS_RETURN;
}

USB_DETACH(p54u_usb)
{
	USB_DETACH_START(p54u_usb, sc);
	p54u_shutdown(sc);
	p54u_free_device(p54u_devunit(sc));
	return 0;
}

DRIVER_MODULE(p54u_usb, uhub, p54u_usb_driver, p54u_usb_devclass, usbd_driver_load, 0);

#endif

#ifdef WITH_PCI

static const struct p54u_device_entry p54u_pci_table[] = {
	{0x1260, 0x3877, "Generic ISL3877",            P54U_DEVTYPE_PCI, 0},
	{0x1260, 0x3886, "Generic ISL3886",            P54U_DEVTYPE_PCI, 0},
	{0x1260, 0x3888, "Generic ISL3888",            P54U_DEVTYPE_PCI, 0},
	{0x1260, 0x3890, "Generic ISL3890",            P54U_DEVTYPE_PCI, 0},
	{0x10b7, 0x6001, "3COM 3CRWE154G72",           P54U_DEVTYPE_PCI, 0},
};

static int p54u_pci_probe(device_t dev)
{
	unsigned int i;
	
	for(i=0;i<sizeof(p54u_pci_table)/sizeof(p54u_pci_table[0]);i++) {
		if((p54u_pci_table[i].vendor == pci_get_vendor(dev)) && (p54u_pci_table[i].product == pci_get_device(dev))) {
			device_set_desc(dev, p54u_pci_table[i].name);
			return BUS_PROBE_DEFAULT;
		}
	}

	return ENXIO;
}

static int p54u_pci_detach(device_t dev);

static void p54u_dma_map_addr(void *arg, bus_dma_segment_t *segs, int nseg, int error)
{
	uint32_t *paddr;

	if(error) return;

	paddr = arg;
	*paddr = segs->ds_addr;
}

static int p54u_pci_attach(device_t dev)
{
	struct p54u_softc *sc;
	unsigned int i;
	unsigned char latency_timer;
	int rid;
	
	sc = device_get_softc(dev);
	bzero(sc, sizeof(struct p54u_softc));
#if ((defined WITH_DEBUG_GEN) || (defined WITH_DEBUG_NET))
	microtime(&sc->start_time);
#endif
	
	if(p54u_alloc_device(sc) == -1)
		return ENOMEM;

	sc->bus.pci.dev = dev;
	
	for(i=0;i<sizeof(p54u_pci_table)/sizeof(p54u_pci_table[0]);i++)
		if((p54u_pci_table[i].vendor == pci_get_vendor(dev)) && (p54u_pci_table[i].product == pci_get_device(dev)))
			break;
	
	sc->device_type = p54u_pci_table[i].type;
	sc->leds_inverted = p54u_pci_table[i].leds_inverted;
	
	latency_timer = pci_read_config(dev, PCI_LATENCY_TIMER, 1);
	P54U_DBG_GEN("latency timer %x", latency_timer);
	if(latency_timer < PCIDEVICE_LATENCY_TIMER_MIN) {
		 latency_timer = PCIDEVICE_LATENCY_TIMER_MIN;
		 pci_write_config(dev, PCI_LATENCY_TIMER, latency_timer, 1);
	}
	
	/* Get memory-mapped registers */
	rid = PCI_BASE_ADDRESS_0;
	sc->bus.pci.memspace = bus_alloc_resource_any(dev, SYS_RES_MEMORY, &rid, RF_ACTIVE);
	
	if(sc->bus.pci.memspace == NULL) return ENXIO;
	
	sc->bus.pci.btag = rman_get_bustag(sc->bus.pci.memspace);
	sc->bus.pci.bhandle = rman_get_bushandle(sc->bus.pci.memspace);
	
	/* Get IRQ */
	rid = 0;
	sc->bus.pci.interrupt = bus_alloc_resource_any(dev, SYS_RES_IRQ, &rid, RF_SHAREABLE|RF_ACTIVE);
	if(sc->bus.pci.interrupt == NULL) {
		p54u_pci_detach(dev);
		return ENXIO;
	}
	
	/* Set up DMA */
	pci_enable_busmaster(dev);
	if(bus_dma_tag_create(NULL, 4, 0, BUS_SPACE_MAXADDR_32BIT, BUS_SPACE_MAXADDR, NULL, NULL, sizeof(struct isl38xx_cb), 1, sizeof(struct isl38xx_cb), 0, NULL, NULL, &sc->bus.pci.cb_tag)) {
		p54u_pci_detach(dev);
		return ENXIO;
	}
	sc->bus.pci.cb_tag_created = 1;
	
	if(bus_dmamem_alloc(sc->bus.pci.cb_tag, (void **)&sc->bus.pci.cb, BUS_DMA_NOWAIT|BUS_DMA_ZERO|BUS_DMA_COHERENT, &sc->bus.pci.cb_map)) {
		p54u_pci_detach(dev);
		return ENXIO;
	}
	
	if(bus_dmamap_load(sc->bus.pci.cb_tag, sc->bus.pci.cb_map, sc->bus.pci.cb, sizeof(struct isl38xx_cb), p54u_dma_map_addr, &sc->bus.pci.cb_dma_addr, BUS_DMA_NOWAIT)) {
		p54u_pci_detach(dev);
		return ENXIO;
	}
	
	bus_dmamap_sync(sc->bus.pci.cb_tag, sc->bus.pci.cb_map, BUS_DMASYNC_PREREAD);
		
	if(!p54u_init(sc)) {
		P54U_ERR("Failed to initialize the driver");
		p54u_shutdown(sc);
		return ENXIO;
	}
	
	return 0;
}

static int p54u_pci_detach(device_t dev)
{
	struct p54u_softc *sc;
	
	sc = device_get_softc(dev);
	p54u_shutdown(sc);
	if(sc->bus.pci.cb_dma_addr != 0)
		bus_dmamap_unload(sc->bus.pci.cb_tag, sc->bus.pci.cb_map);
	if(sc->bus.pci.cb)
		bus_dmamem_free(sc->bus.pci.cb_tag, sc->bus.pci.cb, sc->bus.pci.cb_map);
	if(sc->bus.pci.cb_tag_created) bus_dma_tag_destroy(sc->bus.pci.cb_tag);
	if(sc->bus.pci.interrupt)
		bus_release_resource(dev, SYS_RES_IRQ, 0, sc->bus.pci.interrupt);
	if(sc->bus.pci.memspace)
		bus_release_resource(dev, SYS_RES_MEMORY, PCI_BASE_ADDRESS_0, sc->bus.pci.memspace);
	p54u_free_device(p54u_devunit(sc));
	return 0;
}

/*
 * Stop all chip I/O so that the kernel's probe routines don't
 * get confused by errant DMAs when rebooting.
 */
static void p54u_pci_shutdown(device_t dev)
{
	p54u_pci_detach(dev);
}

static device_method_t p54u_pci_methods[] = {
	DEVMETHOD(device_probe,         p54u_pci_probe),
	DEVMETHOD(device_attach,        p54u_pci_attach),
	DEVMETHOD(device_detach,        p54u_pci_detach),
	DEVMETHOD(device_shutdown,      p54u_pci_shutdown),
	{ 0, 0 }
};

static driver_t p54u_pci_driver = {
	"p54u_pci",
	p54u_pci_methods,
	sizeof(struct p54u_softc)
};

static devclass_t p54u_pci_devclass;

DRIVER_MODULE(p54u_pci, pci, p54u_pci_driver, p54u_pci_devclass, 0, 0);

#endif
