/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
  Copyright (C) 2004 Feyd
  Copyright (C) 2004, 2005 Jean-Baptiste Note <jean-baptiste.note@m4x.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef __P54U_H
#define __P54U_H

#include <sys/types.h>
#include <sys/param.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/sema.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/queue.h>
#include <sys/taskqueue.h>
#include <sys/mutex.h>
#include <sys/socket.h>
#include <sys/mbuf.h>
#include <sys/module.h>
#include <sys/bus.h>

#ifdef WITH_USB
#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#endif

#ifdef WITH_PCI
#include <machine/bus.h>
#include <machine/resource.h>
#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>
#endif

#include <net/bpf.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>
#include <net/if_types.h>

#include <net80211/ieee80211_var.h>
#include <net80211/ieee80211_proto.h>
#include <net80211/ieee80211_ioctl.h>
#ifdef WITH_BPF
#include <net80211/ieee80211_radiotap.h>
#endif

#include "isl38xx_host.h"
#include "transport.h"
#include "firmware.h"
#include "output_layer.h"
#include "pda.h"
#include "netif.h"
#include "pmalloc.h"
#include "devnode.h"

MALLOC_DECLARE(M_P54U);

#define P54U_NUM_CHANNELS            14
#define P54U_MAX_ID_SIZE             17

#define p54u_tx_lock(_sc) do { \
	P54U_DBG_GEN("TX lock line %d", __LINE__); \
	sx_xlock(&_sc->tx_sx); \
	P54U_DBG_GEN("TX lock OK"); \
} while(0)
#define p54u_tx_unlock(_sc) do { \
	P54U_DBG_GEN("TX unlock line %d", __LINE__); \
	sx_xunlock(&_sc->tx_sx); \
	P54U_DBG_GEN("TX unlock OK"); \
} while(0)

/* We use ifdef's here to ensure that we won't forget them later in the code */
/* we don't use these macros in p54u_parse_bra() to be able to display relevant error messages when firmware is unsupported */
#ifdef WITH_SOFTMAC
#define P54U_IS_SOFTMAC(_sc) ((_sc->firmware_type == P54U_FWTYPE_LM86) || (_sc->firmware_type == P54U_FWTYPE_LM87))
#endif
#ifdef WITH_FREEMAC
#define P54U_IS_FREEMAC(_sc) (_sc->firmware_type == P54U_FWTYPE_FMAC)
#endif

#define P54U_MAX_DEVICE_NAME_SIZE 64

enum {
	P54U_DEVTYPE_USB1 = 0,
	P54U_DEVTYPE_USB2 = 1,
	P54U_DEVTYPE_PCI = 2,
};

struct p54u_device_entry {
	uint16_t vendor;
	uint16_t product;
	char name[P54U_MAX_DEVICE_NAME_SIZE];
	int type;
	int leds_inverted;
};

#ifdef WITH_PCI

enum {
	P54U_QUEUE_TYPE_RX,
	P54U_QUEUE_TYPE_TX,
};

struct p54u_dma_membuf {
	struct mbuf *m;
	bus_dmamap_t map;
};


/* Each queue has three indexes:
*   free_index/device_curr_frag/driver_curr_frag
* All indexes are ever-increasing, but interpreted modulo the
* device queue size when used.
*   free_index <= device_curr_frag <= driver_curr_frag  at all times
* For rx queues, [free_index, device_curr_frag) contains fragments
* that the interrupt processing needs to handle (owned by driver).
* [device_curr_frag, driver_curr_frag) is the free space in the
* rx queue, waiting for data (owned by device).  The driver
* increments driver_curr_frag to indicate to the device that more
* buffers are available.
* If device_curr_frag == driver_curr_frag, no more rx buffers are
* available, and the rx DMA engine of the device is halted.
* For tx queues, [free_index, device_curr_frag) contains fragments
* where tx is done; they need to be freed (owned by driver).
* [device_curr_frag, driver_curr_frag) contains the frames
* that are being transferred (owned by device).  The driver
* increments driver_curr_frag to indicate that more tx work
* needs to be done.
*/

struct p54u_queue {
	bus_dma_tag_t tag;
	int type;
	unsigned int size;
	/* those three indexes must be interpreted modulo the queue size when used to index arrays */
	unsigned int free_index;
	/* those are pointers in the control block, so little endian */
	uint32_t *device_index;
	uint32_t *driver_index;
	struct isl38xx_fragment *fragments;
	struct p54u_dma_membuf *membufs;
};

#endif

/* Structure to hold all of our device specific stuff */
struct p54u_softc {
	union {
#ifdef WITH_USB
		struct {
			USBBASEDEVICE sc_dev;
			usbd_device_handle sc_udev;
			usbd_interface_handle sc_iface;
			
			/* USB transfers */
			/* Endpoints */
			struct p54u_pipe data_tx;
			struct p54u_pipe data_rx;
			/* The following pipes are for controlling the NET2280 - unused on v2 devices */
			struct p54u_pipe dev_tx;
			struct p54u_pipe dev_rx;
			struct p54u_pipe brg_tx;
			struct p54u_pipe brg_rx;
			struct p54u_pipe int_rx;
			
			/* Mutex to protect output_queue and output_task */
			struct mtx output_lock;
			TAILQ_HEAD(, p54u_output_frame) output_queue;
			struct task output_task;
			int output_initialized;
		} usb;
#endif
#ifdef WITH_PCI
		struct {
			device_t dev;
			struct resource *memspace;
			bus_space_handle_t bhandle; /* bus space handle */
			bus_space_tag_t btag; /* bus space tag */
			struct resource *interrupt;
			void *interrupt_handler;
			
			int cb_tag_created;
			bus_dma_tag_t cb_tag;
			struct isl38xx_cb *cb; /* control block, used for DMA transfers */
			bus_dmamap_t cb_map;
			uint32_t cb_dma_addr;
			
			struct p54u_queue data_rx;
			struct p54u_queue data_tx;
			struct p54u_queue data_rx_hipri;
			struct p54u_queue data_tx_hipri;
			int queues_initialized;
		} pci;
#endif
	} bus;

#if ((defined WITH_DEBUG_GEN) || (defined WITH_DEBUG_NET))
	struct timeval start_time; /* Keeps track of the time when the device was connected */
#endif

	int err; /* Set on fatal error (requiring power cycle of the device) */

	int device_type;

	int sync_initialized;
	int first_boot;
	int firmware_running;
	/* Boot record area information */
	char firmware_type_string[P54U_FWTYPE_SIZE];                /* SoftMAC and FreeMAC */
	int firmware_type;
#ifdef WITH_SOFTMAC
	char firmware_version_string[P54U_MAX_VERSION_STRING_SIZE]; /* SoftMAC only */
	char firmware_build_string[P54U_MAX_BUILD_STRING_SIZE];     /* SoftMAC only */
	uint16_t firmware_variant;                                  /* SoftMAC only */
	uint16_t firmware_bottom_compat;                            /* SoftMAC only */
	uint16_t firmware_top_compat;                               /* SoftMAC only */
	uint16_t firmware_physical_variant;                         /* SoftMAC only */
	uint16_t firmware_physical_bottom_compat;                   /* SoftMAC only */
	uint16_t firmware_physical_top_compat;                      /* SoftMAC only */
	int physical_transport_version;
#endif

	int rate_count;
	uint8_t rate_table[P54U_LMAC_MAXRATES];                  /* Read from the BRA on SoftMAC - what to do on FreeMAC is still to be defined */

	/* Start and end addresses of the memory space for frame storage on the Prism chipset */
	/* Read from the BRA on SoftMAC, sent by the firmware itself on FreeMAC */
	uint32_t frame_space_start;
	uint32_t frame_space_end;

	/* Device node */
	int devnode_initialized;
	struct p54u_devnode_softc devnode_softc;

	/* Input */
	/* Version 1 and PCI devices only */
	int irq_initialized;
	struct task irq_task;
	int uart_cts;
	int ongoing_frame;
	uint16_t ongoing_frame_fullsize;
	struct mbuf *ongoing_frame_data;
	
	
	/* Output */
	/* SX lock to synchronize register writes between
	 * the output thread which uses p54u_reg_rw() with USB version 1 and PCI devices
	 * and ioctl calls that use p54u_reg_rw()
	 */
	struct sx tx_sx;
	struct p54u_pmalloc_state pmalloc_state;
	int pmalloc_state_initialized;
	uint32_t control_size;
	uint32_t control_address;
	uint32_t reserved_size;
	uint32_t reserved_address;

	/* Firmware startup synchronization */
	struct sema init_sema;

	/* Set if the protocol layer (in softmac.c or freemac.c) has been initialized */
	int protocol_initialized;

	/* EEPROM */
	int eeprom_available;
	struct sema eeprom_read_sema;
	uint8_t eeprom_data[P54U_EEPROM_SIZE];
	char manufacturing_part_number[P54U_MAX_ID_SIZE];
	char nic_serial_number[P54U_MAX_ID_SIZE];
	uint8_t *dev_addr; /* pointer to the MAC address in eeprom_data */
	uint16_t rxhw; /* some rx/tx filter value, describes a hardware component */
	uint32_t platform_id;
	uint16_t antenna1_gain;
	uint16_t antenna2_gain;
	struct p54u_pda_output_limits_rev0 *pda_output_limits_rev0; /* pointer to this structure in eeprom_data */
	struct p54u_pda_pa_curve_data_rev0 *pda_pa_curve_data_rev0; /* pointer to this structure in eeprom_data */
	unsigned int n_iq_autocal;
	struct p54u_pda_iq_autocal_channel *iq_autocal;

	/* LEDs */
	int leds_inverted; /* FIXME: find out where this is in the PDA */
	int leds_auto;
	int last_led_event;

	/* Stats report */
	struct sema stats_read_sema;
	unsigned int stats_rx_packets;
	unsigned int stats_time;
	unsigned int stats_unk1;
	unsigned int stats_unk2;
	unsigned int stats_unk3;
	unsigned int stats_unk4;
	unsigned int stats_unk5;
	unsigned int stats_unk6;
	unsigned int stats_unk7;

	/* Networking */
	int if_initialized;
	struct mtx network_lock;
	struct ieee80211com sc_ic;
	struct ieee80211_beacon_offsets sc_boff;
	int sc_if_flags;
	struct callout sc_scan_ch;
	int (*sc_newstate)(struct ieee80211com *, enum ieee80211_state, int);
	int (*sc_key_alloc)(struct ieee80211com *, const struct ieee80211_key *, ieee80211_keyix *, ieee80211_keyix *);
	uint8_t rxrate;

#ifdef WITH_BPF
	struct bpf_if *sc_drvbpf;

	union {
		struct p54u_rx_radiotap_header th;
		uint8_t pad[64];
	} sc_rxtapu;
#define sc_rxtap        sc_rxtapu.th

	int sc_rxtap_len;

	union {
		struct p54u_tx_radiotap_header th;
		uint8_t pad[64];
	} sc_txtapu;
#define sc_txtap        sc_txtapu.th

	int sc_txtap_len;
#endif
	/* Crash dumps */
	int crash_data_size;
	char *crash_data;
};

void p54u_delay(int ms);

#define P54U_MAX_DEVICES 8

extern struct p54u_softc *p54u_devices[P54U_MAX_DEVICES];

int p54u_alloc_device(struct p54u_softc *sc);
int p54u_devunit(struct p54u_softc *sc);
struct p54u_softc *p54u_get_sc(int unit);
void p54u_free_device(int unit);

#ifdef WITH_DEBUG_GEN
extern int debug_gen;
extern int debug_dumpxfers;
#define P54U_DBG_GEN(format, arg...) do { if(debug_gen) printf("p54u%d: DEBUG_GEN: ("__FILE__ ") " format "\n", p54u_devunit(sc), ##arg); } while(0)
#else
#define P54U_DBG_GEN(format, arg...)
#endif

#ifdef WITH_DEBUG_NET
extern int debug_net;
#define P54U_DBG_NET(format, arg...) do { if(debug_net) printf("p54u%d: DEBUG_NET: ("__FILE__ ") " format "\n", p54u_devunit(sc), ##arg); } while(0)
#else
#define P54U_DBG_NET(format, arg...)
#endif

#define P54U_WARN(format, arg...) printf("p54u%d: WARNING: ("__FILE__ ") " format "\n", p54u_devunit(sc), ##arg)
#define P54U_ERR(format, arg...) printf("p54u%d: ERROR: ("__FILE__ ") " format "\n", p54u_devunit(sc), ##arg)

#if ((defined WITH_DEBUG_GEN) || (defined WITH_DEBUG_NET))
void p54u_data_debug(struct p54u_softc *sc, unsigned int ep, const void *data, int len);
#endif

int p54u_boot(struct p54u_softc *sc, char *fw, unsigned int size);
int p54u_reset(struct p54u_softc *sc);

#endif /* __P54U_H */
