/*
  Copyright (C) 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
  Copyright (C) 2005 Nokia Corporation (Samuel Ortiz <samuel.ortiz@nokia.com>)
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "config.h"

#include <sys/types.h>
#include <sys/endian.h>
#include <sys/systm.h>
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/module.h>

#include "pda.h"
#include "p54u.h"

static uint16_t crc16(uint16_t crc, uint8_t byte)
{
	uint16_t t;

	crc ^= byte;
	t = (crc << 8) ^ (crc << 12);
	crc = t ^ (t >> 5) ^ (t >> 12) ^ (crc >> 8);

	return crc;
}

int p54u_pda_parse(struct p54u_softc *sc, uint8_t *pda_data, unsigned int pda_length)
{
	unsigned int i;
	int end_of_pda;
	unsigned int current_pointer;
	uint16_t crc;
	struct p54u_pda_record *record;
	unsigned int payload_size;

	P54U_DBG_GEN("Parsing PDA (%d bytes long)", pda_length);

	if(pda_length < sizeof(struct p54u_pda_record))
	{
		P54U_ERR("PDA length too small for a single record");
		return 0;
	}

	end_of_pda = 0;
	current_pointer = 0;
	crc = 0xffff;
	while(!end_of_pda)
	{
		/* Get a record */
		record = (struct p54u_pda_record *)(pda_data+current_pointer);
		payload_size = (le16toh(record->size)-1)*2; /* -1 to remove the size field */
		P54U_DBG_GEN("PDR code 0x%04x, payload size 0x%04x", le16toh(record->code), payload_size);
		/* We use this check because it is likely that current_pointer+... overflows */
		/* if we are encountering really corrupted data */
		if((payload_size >= pda_length) || (current_pointer+payload_size+sizeof(struct p54u_pda_record)) >= pda_length) {
			P54U_ERR("PDR beyond end of PDA");
			return 0;
		}
		/* Update CRC */
		for(i=0;i<sizeof(struct p54u_pda_record);i++)
			crc = crc16(crc, *(((uint8_t *)(record))+i));
		if(le16toh(record->code) != P54U_PDR_END)
			for(i=0;i<payload_size;i++)
				crc = crc16(crc, record->payload[i]);
		/* Parse the record */
		switch(le16toh(record->code)) {
		case P54U_PDR_END: {
				uint16_t *payload;

				P54U_DBG_GEN("PDA end, %d bytes", payload_size);
				if(payload_size < sizeof(uint16_t)) {
					P54U_ERR("Not enough payload for CRC");
					return 0;
				}
				payload = (uint16_t *)&record->payload[0];
				crc = ~crc & 0xffff;
				P54U_DBG_GEN("CRC: computed 0x%04x, expected 0x%04x", crc, le16toh(*payload));
				if(le16toh(*payload) != crc)
					P54U_WARN("PDA CRC16 failed - continuing anyway, but expect problems");
				end_of_pda = 1;
				break;
			}
		case P54U_PDR_MANUFACTURING_PART_NUMBER:
			P54U_DBG_GEN("Manufacturing part number, %d bytes", payload_size);
			memcpy(sc->manufacturing_part_number, &record->payload[0], payload_size > P54U_MAX_ID_SIZE ? P54U_MAX_ID_SIZE : payload_size);
			break;
		case P54U_PDR_PDA_VERSION:
			P54U_DBG_GEN("Version, %d bytes", payload_size);
			break;
		case P54U_PDR_NIC_SERIAL_NUMBER:
			P54U_DBG_GEN("NIC serial number, %d bytes", payload_size);
			memcpy(sc->nic_serial_number, &record->payload[0], payload_size > P54U_MAX_ID_SIZE ? P54U_MAX_ID_SIZE : payload_size);
			break;

		case P54U_PDR_MAC_ADDRESS:
			P54U_DBG_GEN("MAC address, %d bytes", payload_size);
			if(payload_size != 6) {
				P54U_ERR("Incorrect MAC address size (%d)", payload_size);
				return 0;
			}
			sc->dev_addr = &record->payload[0];
			break;
		case P54U_PDR_REGULATORY_DOMAIN_LIST:
			P54U_DBG_GEN("Regulatory domain list, %d bytes", payload_size);
			break;
		case P54U_PDR_TEMPERATURE_TYPE:
			P54U_DBG_GEN("Temperature type, %d bytes", payload_size);
			break;

		case P54U_PDR_PRISM_PCI_IDENTIFIER:
			P54U_DBG_GEN("PCI identifier, %d bytes", payload_size);
			break;

		case P54U_PDR_COUNTRY_INFORMATION:
			P54U_DBG_GEN("Country information, %d bytes", payload_size);
			break;
		case P54U_PDR_INTERFACE_LIST: {
				struct p54u_pda_interface *payload;
				int i;

				P54U_DBG_GEN("Interface list, %d bytes", payload_size);
				payload = (struct p54u_pda_interface *)&record->payload[0];
				for(i=0;i<payload_size/sizeof(struct p54u_pda_interface);i++) {
					P54U_DBG_GEN("Interface %d: role 0x%02x, id 0x%02x, variant 0x%02x, btm compat 0x%02x, top compat 0x%02x",
						    i,
						    le16toh(payload[i].role),
						    le16toh(payload[i].if_id),
						    le16toh(payload[i].variant),
						    le16toh(payload[i].bottom_compat),
						    le16toh(payload[i].top_compat));
					if(payload[i].if_id == 0x000f) sc->rxhw = le16toh(payload[i].variant) & 0x0007;
				}
				
				break;
			}
		case P54U_PDR_HARDWARE_PLATFORM_COMPONENT_ID: {
				uint32_t *payload;

				P54U_DBG_GEN("Component ID, %d bytes", payload_size);

				if(payload_size < sizeof(uint32_t)) {
					P54U_ERR("Not enough payload for component ID");
					return 0;
				}
				payload = (uint32_t *)&record->payload[0];
				sc->platform_id = *payload;
				break;
			}
		case P54U_PDR_OEM_NAME:
			P54U_DBG_GEN("OEM name");
			break;
		case P54U_PDR_PRODUCT_NAME:
			P54U_DBG_GEN("Product name");
			break;
		case P54U_PDR_UTF8_OEM_NAME:
			P54U_DBG_GEN("OEM name (UTF8)");
			break;
		case P54U_PDR_UTF8_PRODUCT_NAME:
			P54U_DBG_GEN("Product name (UTF8)");
			break;
		case P54U_PDR_COUNTRY_LIST:
			P54U_DBG_GEN("Country list, %d bytes", payload_size);
			break;
		case P54U_PDR_DEFAULT_COUNTRY:
			P54U_DBG_GEN("Default country, %d bytes", payload_size);
			break;

		case P54U_PDR_ANTENNA_GAIN: {
				struct p54u_pda_antenna_gain *payload;

				P54U_DBG_GEN("Antenna gain, %d bytes", payload_size);

				if(payload_size != sizeof(struct p54u_pda_antenna_gain)) {
					P54U_ERR("Unexpected size of antenna gain PDR (%d)", payload_size);
					return 0;
				}
				payload = (struct p54u_pda_antenna_gain *)&record->payload[0];
				sc->antenna1_gain = le16toh(payload->antenna1);
				sc->antenna2_gain = le16toh(payload->antenna2);
				break;
			}

		case P54U_PDR_PRISM_INDIGO_PA_CALIBRATION_DATA:
			P54U_DBG_GEN("Indigo calibration, %d bytes", payload_size);
			break;
		case P54U_PDR_RSSI_LINEAR_APPROXIMATION:
			P54U_DBG_GEN("RSSI linear approx, %d bytes", payload_size);
			break;

		/* was freq3 */
		case P54U_PDR_PRISM_PA_CAL_OUTPUT_POWER_LIMITS: {
				uint8_t revision;
				unsigned int expected_size;

				P54U_DBG_GEN("Output power limits, %d bytes", payload_size);
				if(payload_size < 1) {
					P54U_ERR("Payload too short for output power limits");
					return 0;
				}
				revision = record->payload[0];
				if(revision != P54U_PDR1903_REV0) {
					P54U_ERR("Unknown revision for output limits (%d)", revision);
					return 0;
				}
				sc->pda_output_limits_rev0 = (struct p54u_pda_output_limits_rev0 *)&record->payload[0];
				if(payload_size < sizeof(struct p54u_pda_output_limits_rev0)) {
					P54U_ERR("Payload is too short for struct p54u_pda_output_limits_rev0");
					return 0;
				}
				expected_size = sizeof(struct p54u_pda_output_limits_rev0)+sc->pda_output_limits_rev0->n_pa_levels*sizeof(struct p54u_pda_output_limits_channel_rev0);
				if(payload_size != expected_size) {
					P54U_ERR("Unexpected size of output power limits (%d instead of %d)", payload_size, expected_size);
					return 0;
				}
				break;
			}

		/* was freq4 */
		case P54U_PDR_PRISM_PA_CAL_CURVE_DATA: {
				uint8_t revision;
				unsigned int expected_size;

				P54U_DBG_GEN("PA calibration, %d bytes", payload_size);
				if(payload_size < 1) {
					P54U_ERR("Payload too short for PA calibration");
					return 0;
				}
				revision = record->payload[0];
				if(revision != P54U_PDR1904_REV0) {
					P54U_ERR("Unknown revision for PA calibration data (%d)", revision);
					return 0;
				}
				sc->pda_pa_curve_data_rev0 = (struct p54u_pda_pa_curve_data_rev0 *)&record->payload[0];
				if(payload_size < sizeof(struct p54u_pda_pa_curve_data_rev0)) {
					P54U_ERR("Payload is too short for struct p54u_pda_pa_curve_data_rev0");
					return 0;
				}
				expected_size = sizeof(struct p54u_pda_pa_curve_data_rev0)
					+ sc->pda_pa_curve_data_rev0->n_channels*(sizeof(struct p54u_pda_pa_curve_data_channel_rev0) + sc->pda_pa_curve_data_rev0->points_per_channel*sizeof(struct p54u_pda_pa_curve_data_sample_rev0));
				if(payload_size != expected_size) {
					P54U_ERR("Unexpected size of PA calibration data (%d instead of %d)", payload_size, expected_size);
					return 0;
				}
				break;
			}

		case P54U_PDR_RSSI_LINEAR_APPROXIMATION_DUAL_BAND:
			P54U_DBG_GEN("RSSI linear approx (dual band), %d bytes", payload_size);
			break;

			/* was freq6 */
		case P54U_PDR_PRISM_ZIF_TX_IQ_CALIBRATION: {
				P54U_DBG_GEN("ZIF IQ calibration, %d bytes", payload_size);
				
				sc->n_iq_autocal = payload_size/sizeof(struct p54u_pda_iq_autocal_channel);
				sc->iq_autocal = (struct p54u_pda_iq_autocal_channel *)&record->payload[0];

				break;
			}
		case P54U_PDR_REGULATORY_POWER_LIMITS:
			P54U_DBG_GEN("Regulatory power limits, %d bytes", payload_size);
			break;
		case P54U_PDR_RSSI_LINEAR_APPROXIMATION_EXTENDED:
			P54U_DBG_GEN("RSSI linear approx (extended), %d bytes", payload_size);
			break;
		case P54U_PDR_RADIATED_TRANSMISSION_CORRECTION:
			P54U_DBG_GEN("Radiated transmission correction, %d bytes", payload_size);
			break;
		case P54U_PDR_PRISM_TX_IQ_CALIBRATION:
			P54U_DBG_GEN("TX IQ calibration, %d bytes", payload_size);
			break;

		case P54U_PDR_BASEBAND_REGISTERS:
			P54U_DBG_GEN("Baseband registers, %d bytes", payload_size);
			break;
		case P54U_PDR_PER_CHANNEL_BASEBAND_REGISTERS:
			P54U_DBG_GEN("Per channel baseband registers, %d bytes", payload_size);
			break;

		default:
			P54U_WARN("PDA contains an unknown option (0x%04x, payload %d bytes). Please mail us an EEPROM dump of your device (see README file).", le16toh(record->code), payload_size);
			break;
		}
		current_pointer += payload_size+sizeof(struct p54u_pda_record);
	}
	return 1;
}
