/*
  Copyright (C) 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
  Copyright (C) 2005 Nokia Corporation (Samuel Ortiz <samuel.ortiz@nokia.com>)
  Copyright (C) 2006 Jean-Baptiste Note <jean-baptiste.note@m4x.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

/* pda.h : Parsing of the Production Data Area */

#ifndef __PDA_H
#define __PDA_H

/*
 * From N770:
 * this files defines the PDR codes used to build PDAs as defined in document
 * number 553155. The current implementation mirrors version 1.1 of the
 * document and lists only PDRs supported by the ARM platform.
 */

/* common and choice range (0x0000 - 0x0fff) */
#define P54U_PDR_END                                 0x0000
#define P54U_PDR_MANUFACTURING_PART_NUMBER           0x0001
#define P54U_PDR_PDA_VERSION                         0x0002
#define P54U_PDR_NIC_SERIAL_NUMBER                   0x0003

#define P54U_PDR_MAC_ADDRESS                         0x0101
#define P54U_PDR_REGULATORY_DOMAIN_LIST              0x0103
#define P54U_PDR_TEMPERATURE_TYPE                    0x0107

#define P54U_PDR_PRISM_PCI_IDENTIFIER                0x0402

/* ARM range (0x1000 - 0x1fff) */
#define P54U_PDR_COUNTRY_INFORMATION                 0x1000
#define P54U_PDR_INTERFACE_LIST                      0x1001
#define P54U_PDR_HARDWARE_PLATFORM_COMPONENT_ID      0x1002
#define P54U_PDR_OEM_NAME                            0x1003
#define P54U_PDR_PRODUCT_NAME                        0x1004
#define P54U_PDR_UTF8_OEM_NAME                       0x1005
#define P54U_PDR_UTF8_PRODUCT_NAME                   0x1006
#define P54U_PDR_COUNTRY_LIST                        0x1007
#define P54U_PDR_DEFAULT_COUNTRY                     0x1008

#define P54U_PDR_ANTENNA_GAIN                        0x1100

#define P54U_PDR_PRISM_INDIGO_PA_CALIBRATION_DATA    0x1901
#define P54U_PDR_RSSI_LINEAR_APPROXIMATION           0x1902
#define P54U_PDR_PRISM_PA_CAL_OUTPUT_POWER_LIMITS    0x1903
#define P54U_PDR_PRISM_PA_CAL_CURVE_DATA             0x1904
#define P54U_PDR_RSSI_LINEAR_APPROXIMATION_DUAL_BAND 0x1905
#define P54U_PDR_PRISM_ZIF_TX_IQ_CALIBRATION         0x1906
#define P54U_PDR_REGULATORY_POWER_LIMITS             0x1907
#define P54U_PDR_RSSI_LINEAR_APPROXIMATION_EXTENDED  0x1908
#define P54U_PDR_RADIATED_TRANSMISSION_CORRECTION    0x1909
#define P54U_PDR_PRISM_TX_IQ_CALIBRATION             0x190a

/* reserved range (0x2000 - 0x7fff) */

/* customer range (0x8000 - 0xffff) */
#define P54U_PDR_BASEBAND_REGISTERS                  0x8000
#define P54U_PDR_PER_CHANNEL_BASEBAND_REGISTERS      0x8001

/*
 * regulatory domain codes as defined in IEEE 802.11
 */
#define P54U_REGDOMAIN_FCC           0x10
#define P54U_REGDOMAIN_IC            0x20
#define P54U_REGDOMAIN_ETSI          0x30
#define P54U_REGDOMAIN_SPAIN         0x31
#define P54U_REGDOMAIN_FRANCE        0x32
#define P54U_REGDOMAIN_MKK           0x40    /* channel 14 only */
#define P54U_REGDOMAIN_MKK13         0x41    /* channel 1 to 13 */

#define P54U_MAX_ARM_PDR             32
#define P54U_PDR_SIZE_LENGTH         2
#define P54U_PDR_CODE_LENGTH         2


/* ARM PDA types */
struct p54u_pda_record {
	uint16_t size; /* in 16-bit words, including this header */
	uint16_t code;
	uint8_t payload[];
};


/*
 * PDR structures
 */

/* Interface list PDR 0x1001 */
 struct p54u_pda_interface {
	 uint16_t role;
	 uint16_t if_id;
	 uint16_t variant;
	 uint16_t bottom_compat;
	 uint16_t top_compat;
} __attribute__ ((packed));;

/* [RE] Antenna gain PDR 0x1100 */
struct p54u_pda_antenna_gain
{
	uint16_t antenna1;
	uint16_t antenna2;
} __attribute__ ((packed));


/* RSSI PDR 0x1908 */
struct p54u_pda_rssi_cal_channel {
	uint16_t frequency;
	uint16_t rssi_A;
	uint16_t rssi_B;
	uint16_t rssi_LNA;
} __attribute__ ((packed));

struct p54u_pda_rssi_cal {
	uint8_t pdr_revision;
	uint8_t xcvr_bmp;
	struct p54u_pda_rssi_cal_channel rssi_data[];
} __attribute__ ((packed));


/* IQ calibration PDR 0x1906*/
struct p54u_pda_iq_autocal_channel {
	uint16_t frequency;
	uint16_t iq_param_1;
	uint16_t iq_param_2;
	uint16_t iq_param_3;
	uint16_t iq_param_4;
} __attribute__ ((packed));

/* Output limits PDR 0x1903 */
/* RE */
#define P54U_PDR1903_REV0   0x00
struct p54u_pda_output_limits_channel_rev0 {
	uint16_t frequency;
	uint8_t limit_bpsk;
	uint8_t limit_qpsk;
	uint8_t limit_16qam;
	uint8_t limit_64qam;
	uint8_t rate_set_mask;
	uint8_t rate_set_size;
} __attribute__ ((packed));

struct p54u_pda_output_limits_rev0 {
	uint8_t pdr_revision;
	uint8_t n_pa_levels;
	struct p54u_pda_output_limits_channel_rev0 power[];
} __attribute__ ((packed));

/* PA Curve data PDR 0x1904 */
/* RE */
#define P54U_PDR1904_REV0    0x00
struct p54u_pda_pa_curve_data_sample_rev0 {
	uint8_t rf_power;
	uint8_t pa_detector;
	uint8_t pcv;
} __attribute__ ((packed));

struct p54u_pda_pa_curve_data_channel_rev0 {
	uint16_t frequency;
	struct p54u_pda_pa_curve_data_sample_rev0 curve_samples[];
} __attribute__ ((packed));

struct p54u_pda_pa_curve_data_rev0 {
	uint8_t cal_method_revision;
	uint8_t n_channels;
	uint8_t points_per_channel;
	uint8_t padding;
	struct p54u_pda_pa_curve_data_channel_rev0 curve_data[];
} __attribute__ ((packed));


/* Default country PDR 0x1008 */
struct p54u_pda_default_country {
	uint8_t regulatory_domain;
	uint8_t country_code[3];
} __attribute__ ((packed));


struct p54u_softc;

/* pda_length in bytes */
int p54u_pda_parse(struct p54u_softc *sc, uint8_t *pda_data, unsigned int pda_length);

#endif /* __PDA_H */
