/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

/* pmalloc.c : Prism Memory ALLOCator */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/lock.h>
#include <sys/mutex.h>

#include "pmalloc.h"

//#define DEBUG_PMALLOC

/* All addresses are inclusive */

static uint32_t p54u_pmalloc_unsafe(struct p54u_pmalloc_state *state, uint32_t size, uint32_t headroom, int purpose)
{
	unsigned int i;

	/* Sanity check */
	if(size == 0)
		return P54U_PMALLOC_FAILURE;

	size += headroom;

	/* Alignment */
	if((size % P54U_PMALLOC_ALIGNMENT) != 0)
		size += P54U_PMALLOC_ALIGNMENT - (size % P54U_PMALLOC_ALIGNMENT);

	/* We can't allocate anything because all blocks are used */
	if(state->allocated_blocks == P54U_PMALLOC_MAX_BLOCKS)
		return P54U_PMALLOC_FAILURE;

	/* Nothing's allocated. Just use the first block */
	if(state->allocated_blocks == 0)
	{
		if(size <= (state->end_address - state->start_address + 1)) {
#ifdef DEBUG_PMALLOC
			printf("allocating first block, address 0x%08x\n", state->start_address);
#endif

			state->blocks[0].start_address = state->start_address;
			state->blocks[0].end_address = state->start_address + size - 1;
			state->blocks[0].headroom = headroom;
			state->blocks[0].purpose = purpose;
			state->allocated_blocks = 1;
			return (state->blocks[0].start_address+headroom);
		} else {
			return P54U_PMALLOC_FAILURE;
		}
	}

	/* Some blocks allocated. Iterate through the block list, and find space between two blocks if possible */
	for(i=0;i<(state->allocated_blocks-1);i++)
	{
		if((state->blocks[i+1].start_address - state->blocks[i].end_address - 1) >= size) {
#ifdef DEBUG_PMALLOC
			printf("allocating between blocks %d (0x%08x-0x%08x) and %d (0x%08x-0x%08x), size = %d\n", i, state->blocks[i].start_address, state->blocks[i].end_address, i+1, state->blocks[i+1].start_address, state->blocks[i+1].end_address, size);
#endif
			/* Okay, we know that we can allocate memory between blocks at indexes i and i+1 */
			/* Move blocks to make room at index i+1 - this will always be possible */
			/* because we check first that state->allocated_blocks < P54U_PMALLOC_MAX_BLOCKS */
			bcopy(&state->blocks[i+1], &state->blocks[i+2], (state->allocated_blocks-i-1)*sizeof(struct p54u_pmalloc_block));
			/* Now, we just have to fill in the new block at index i+1 and update the allocated block count */
			state->blocks[i+1].start_address = state->blocks[i].end_address + 1;
			state->blocks[i+1].end_address = state->blocks[i].end_address + size;
			state->blocks[i+1].headroom = headroom;
			state->blocks[i+1].purpose = purpose;
			state->allocated_blocks++;
			return (state->blocks[i+1].start_address+headroom);
		}
	}

	/* If we have reached this point, we must try to allocate memory after the last block */
	/* all this is defined because 1 < state->allocated_blocks < P54U_PMALLOC_MAX_BLOCKS */
	if(size <= (state->end_address - state->blocks[state->allocated_blocks-1].end_address))
	{
#ifdef DEBUG_PMALLOC
		printf("allocating at the end, last block ends at 0x%08x, memory ends at 0x%08x ; new block is %d\n", state->blocks[state->allocated_blocks-1].end_address, state->end_address, state->allocated_blocks);
#endif

		state->blocks[state->allocated_blocks].start_address = state->blocks[state->allocated_blocks-1].end_address + 1;
		state->blocks[state->allocated_blocks].end_address = state->blocks[state->allocated_blocks-1].end_address + size;
		state->blocks[state->allocated_blocks].headroom = headroom;
		state->blocks[state->allocated_blocks].purpose = purpose;
		return (state->blocks[state->allocated_blocks++].start_address+headroom);
	} else
	{
		return P54U_PMALLOC_FAILURE;
	}
}

static int p54u_pmfree_unsafe(struct p54u_pmalloc_state *state, uint32_t address)
{
	unsigned int i;

#ifdef DEBUG_PMALLOC

	printf("freeing address 0x%08x\n", address);
#endif

	for(i=0;i<state->allocated_blocks;i++)
	{
		if((state->blocks[i].start_address + state->blocks[i].headroom) == address) {
			bcopy(&state->blocks[i+1], &state->blocks[i], (state->allocated_blocks-i-1)*sizeof(struct p54u_pmalloc_block));
			state->allocated_blocks--;
		}
	}

	return 0;
}

static uint32_t p54u_pmfind_unsafe(struct p54u_pmalloc_state *state, int purpose)
{
	unsigned int i;

	for(i=0;i<state->allocated_blocks;i++)
	{
		if(state->blocks[i].purpose == purpose)
			return state->blocks[i].start_address;
	}

	return P54U_PMALLOC_FAILURE;
}

/* Thread-safe wrappers for the above functions */

uint32_t p54u_pmalloc(struct p54u_pmalloc_state *state, uint32_t size, uint32_t headroom, int purpose)
{
	uint32_t ret;

	mtx_lock(&state->lock)
	;
	ret = p54u_pmalloc_unsafe(state, size, headroom, purpose);
	mtx_unlock(&state->lock)
	;

	return ret;
}

int p54u_pmfree(struct p54u_pmalloc_state *state, uint32_t address)
{
	int ret;

	mtx_lock(&state->lock)
	;
	ret = p54u_pmfree_unsafe(state, address);
	mtx_unlock(&state->lock)
	;

	return ret;
}

uint32_t p54u_pmfind(struct p54u_pmalloc_state *state, int purpose)
{
	uint32_t ret;

	mtx_lock(&state->lock)
	;
	ret = p54u_pmfind_unsafe(state, purpose);
	mtx_unlock(&state->lock)
	;

	return ret;
}

/* State initialization */

void p54u_pmstate_init(struct p54u_pmalloc_state *state, uint32_t start_address, uint32_t end_address)
{
	mtx_init(&state->lock, "pmalloc", "pmalloc", MTX_DEF)
	;
	state->start_address = start_address;
	state->end_address = end_address;
	state->allocated_blocks = 0;
}

void p54u_pmstate_free(struct p54u_pmalloc_state *state)
{
	mtx_destroy(&state->lock)
	;
}
