/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
  Copyright (C) 2004, 2005 Jean-Baptiste Note <jean-baptiste.note@m4x.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "config.h"

#ifdef WITH_SOFTMAC

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/endian.h>
#include <sys/socket.h>
#include <sys/mbuf.h>

#include <net/bpf.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>
#include <net/if_types.h>

#include <net80211/ieee80211_var.h>
#include <net80211/ieee80211_proto.h>
#include <net80211/ieee80211_ioctl.h>
#ifdef WITH_BPF
#include <net80211/ieee80211_radiotap.h>
#endif

#include "input_layer.h"
#include "pmalloc.h"
#include "pda.h"
#include "softmac.h"
#include "p54u.h"

#ifdef WITH_USB
/* Synchronous function that sends a LMAC frame contained in mbuf m at the specified address */
/* Must aquire appropriate locks before called */
int p54u_lmac_usb_transport(struct p54u_softc *sc, struct mbuf *m, uint32_t address)
{
	int len;
	int ok;
	int b;

	/* Abort all transfers if we already encountered an error */
	if(sc->err) {
		m_freem(m);
		return 0;
	}

	switch(sc->device_type) {
		case P54U_DEVTYPE_USB1: {
			/* VERSION 1 DEVICES : Encapsulate data in a p54u_lm86_v1_physical_tx structure which will contain the Prism address and the length */
			struct p54u_lmac_86_physical_tx *p54u_tx;
	
			/* Backup the payload length */
			len = m->m_pkthdr.len;
			M_PREPEND(m, sizeof(struct p54u_lmac_86_physical_tx), M_TRYWAIT);
			if(m == NULL)
				return 0;
			m = m_pullup(m, sizeof(struct p54u_lmac_86_physical_tx));
			if(m == NULL)
				return 0;
			p54u_tx = mtod(m, struct p54u_lmac_86_physical_tx *);
			bzero(p54u_tx, sizeof(struct p54u_lmac_86_physical_tx));
	
			p54u_tx->address = htole32(address);
			p54u_tx->length = htole16(len);
			p54u_tx->follower = htole16(0); /* seems to have no effect, at least with recent firmwares */
	
			ok = p54u_usb_transport(sc, m);
			break;
		}
		case P54U_DEVTYPE_USB2: {
			/* VERSION 2 DEVICES */
			if(sc->physical_transport_version == 1) {
				/* Physical transport version 1.0 : Send the Prism address, then data, in two separate URBs with no encapsulation */
				*((uint32_t *)sc->bus.usb.data_tx.buf) = htole32(address);
				b = sizeof(uint32_t);
				ok = (p54u_bulk_transfer(sc, &sc->bus.usb.data_tx, &b, 0) == USBD_NORMAL_COMPLETION);
	
				if(ok)
					p54u_usb_transport(sc, m);
			} else {
				/* Physical transport version 1.1 : Encapsulate data in a p54u_lm87_v2_physical_tx structure which will contain the Prism address and a CRC */
				struct p54u_lmac_87_v11_physical_tx *p54u_tx;
	
				M_PREPEND(m, sizeof(struct p54u_lmac_87_v11_physical_tx), M_TRYWAIT);
				if(m == NULL)
					return 0;
				m = m_pullup(m, sizeof(struct p54u_lmac_87_v11_physical_tx));
				if(m == NULL)
					return 0;
				p54u_tx = mtod(m, struct p54u_lmac_87_v11_physical_tx *);
				bzero(p54u_tx, sizeof(struct p54u_lmac_87_v11_physical_tx));
	
				p54u_tx->address = htole32(address);
				p54u_tx->signature = htole32(0); /* TODO: find out how to compute it */
	
				ok = p54u_usb_transport(sc, m);
			}
			break;
		}
	}

	if(!ok)
		sc->err = 1;

	return ok;
}

#endif

/************************/
/*** Output functions ***/
/************************/

/* These functions only generate a frame contained in a mbuf, without altering the hardware or software state */
/* except for 802.11 TX, which prepends the SoftMAC protocol headers to the existing mbuf chain */

struct mbuf *p54u_lmac_abortaddress(struct p54u_softc *sc, uint32_t req_id, uint32_t aborted_address)
{
	struct mbuf *m;
	struct p54u_lmac_abortaddress *frame;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if(m == NULL)
		return NULL;
	m->m_pkthdr.len = m->m_len = sizeof(struct p54u_lmac_abortaddress);
	frame = mtod(m, struct p54u_lmac_abortaddress *);

	frame->aborted_address = htole32(aborted_address);

	frame->header.id1.code = P54U_LMAC_ID1_CODE_CONTROL;
	frame->header.id1.flags = P54U_LMAC_ID1_FLAGS_CONTROL_NOREADBACK;
	frame->header.id1.length = htole16(P54U_LMAC_ID1LEN(struct p54u_lmac_abortaddress));
	frame->header.req_id = htole32(req_id);
	frame->header.type = htole16(P54U_LMAC_ID2_CONTROL_TYPE_ABORTADDR);
	frame->header.flags = htole16(0);

	return m;
}

struct mbuf *p54u_lmac_eeprom_readreq(struct p54u_softc *sc, uint32_t req_id, uint16_t offset, uint16_t length)
{
	struct mbuf *m;
	struct p54u_lmac_eeprom *frame;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if(m == NULL)
		return NULL;
	MCLGET(m, M_DONTWAIT);
	if((m->m_flags & M_EXT) == 0)
	{
		m_freem(m);
		return NULL;
	}
	m->m_pkthdr.len = m->m_len = sizeof(struct p54u_lmac_eeprom)+length;
	frame = mtod(m, struct p54u_lmac_eeprom *);
	/* bzero is not required, but uncomment if you want clean logs */
	/*bzero(frame, sizeof(struct p54u_lmac_eeprom)+length);*/

	frame->offset = htole16(offset);
	frame->len = htole16(length);

	frame->header.id1.code = P54U_LMAC_ID1_CODE_CONTROL;
	frame->header.id1.flags = 0;
	frame->header.id1.length = htole16(P54U_LMAC_ID1LEN(struct p54u_lmac_eeprom)+length);
	frame->header.req_id = htole32(req_id);
	frame->header.type = htole16(P54U_LMAC_ID2_CONTROL_TYPE_EEPROM);
	frame->header.flags = htole16(0);

	return m;
}

struct mbuf *p54u_lmac_leds(struct p54u_softc *sc, uint32_t req_id, uint16_t mode, uint16_t perm_setting, uint16_t temp_setting, uint16_t duration)
{
	struct mbuf *m;
	struct p54u_lmac_leds *frame;
	uint16_t perm_setting2, temp_setting2;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if(m == NULL)
		return NULL;
	m->m_pkthdr.len = m->m_len = sizeof(struct p54u_lmac_leds);
	frame = mtod(m, struct p54u_lmac_leds *);

	if(sc->leds_inverted) {
		perm_setting2 = perm_setting^(P54U_LMAC_LED_A|P54U_LMAC_LED_B);
		temp_setting2 = temp_setting^(P54U_LMAC_LED_A|P54U_LMAC_LED_B);
	} else {
		perm_setting2 = perm_setting;
		temp_setting2 = temp_setting;
	}

	frame->mode = htole16(mode);
	frame->led_setting_permanent = htole16(perm_setting2);
	frame->led_setting_temporary = htole16(temp_setting2);
	frame->temporary_setting_duration = htole16(duration);

	frame->header.id1.code = P54U_LMAC_ID1_CODE_CONTROL;
	frame->header.id1.flags = P54U_LMAC_ID1_FLAGS_CONTROL_NOREADBACK;
	frame->header.id1.length = htole16(P54U_LMAC_ID1LEN(struct p54u_lmac_leds));
	frame->header.req_id = htole32(req_id);
	frame->header.type = htole16(P54U_LMAC_ID2_CONTROL_TYPE_LEDS);
	frame->header.flags = htole16(0);

	return m;
}

struct mbuf *p54u_lmac_channel_change(struct p54u_softc *sc, uint32_t req_id, unsigned int channel, uint16_t magic1, uint16_t magic2)
{
	unsigned int frequency;
	struct mbuf *m;
	struct p54u_lmac_channel *frame;
	int i, index;
	struct p54u_pda_pa_curve_data_channel_rev0 *pa_curve_data;

	frequency = ieee80211_ieee2mhz(channel, IEEE80211_CHAN_2GHZ);
	channel--; /* get array index */
	if(channel >= P54U_NUM_CHANNELS)
		return NULL;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if(m == NULL)
		return NULL;
	m->m_pkthdr.len = m->m_len = sizeof(struct p54u_lmac_channel) 
		+ sc->pda_pa_curve_data_rev0->points_per_channel*sizeof(struct p54u_lmac_pa_curve_data_sample);
	frame = mtod(m, struct p54u_lmac_channel *);

	/* what are those ? */
	frame->magic1 = htole16(magic1);
	frame->magic2 = htole16(magic2);

	/* Find the IQ autocalibration data in the PDA and copy it to the frame */
	index = -1;
	for(i=0;i<sc->n_iq_autocal;i++) {
		if(le16toh(sc->iq_autocal[i].frequency) == frequency) {
			index = i;
			break;
		}
	}
	if(index == -1) {
		P54U_ERR("IQ autocalibration parameters at %dMHz not found in PDA", frequency);
		m_freem(m);
		return NULL;
	}
	memcpy(&frame->iq_autocal, &sc->iq_autocal[index], sizeof(struct p54u_lmac_iq_autocal_channel));
	
	/* Output limits, same thing */
	index = -1;
	for(i=0;i<sc->pda_output_limits_rev0->n_pa_levels;i++) {
		if(le16toh(sc->pda_output_limits_rev0->power[i].frequency) == frequency) {
			index = i;
			break;
		}
	}
	if(index == -1) {
		P54U_ERR("Output limits at %dMHz not found in PDA", frequency);
		m_freem(m);
		return NULL;
	}
	/* Endianness is OK, both LMAC and PDA are little-endian */
	frame->output_limits.limit_barker = sc->pda_output_limits_rev0->power[index].limit_bpsk;
	frame->output_limits.limit_bpsk = sc->pda_output_limits_rev0->power[index].limit_bpsk;
	frame->output_limits.limit_qpsk = sc->pda_output_limits_rev0->power[index].limit_qpsk;
	frame->output_limits.limit_16qam = sc->pda_output_limits_rev0->power[index].limit_16qam;
	frame->output_limits.limit_64qam = sc->pda_output_limits_rev0->power[index].limit_64qam;

	
	/* Find the PA curve data in the PDA, interpolate missing points and put the resulting data in the frame */
	index = -1;
	for(i=0;i<sc->pda_pa_curve_data_rev0->n_channels;i++) {
		pa_curve_data = (struct p54u_pda_pa_curve_data_channel_rev0 *)
			((uint8_t *)sc->pda_pa_curve_data_rev0
			+ sizeof(struct p54u_pda_pa_curve_data_rev0)
			+ i*(sizeof(struct p54u_pda_pa_curve_data_channel_rev0) + sc->pda_pa_curve_data_rev0->points_per_channel*sizeof(struct p54u_pda_pa_curve_data_sample_rev0)));
			if(le16toh(pa_curve_data->frequency) == frequency) {
				index = i;
				break;
			}
	}
	if(index == -1) {
		P54U_ERR("PA curve data at %dMHz not found in PDA", frequency);
		m_freem(m);
		return NULL;
	}
	/* no endianness conversion required (all fields are 8-bit) */
	frame->n_points_in_pa_curve = sc->pda_pa_curve_data_rev0->points_per_channel;
	for(i=0;i<sc->pda_pa_curve_data_rev0->points_per_channel;i++) {
		frame->curve_data[i].rf_power = pa_curve_data->curve_samples[i].rf_power;
		frame->curve_data[i].pa_detector = pa_curve_data->curve_samples[i].pa_detector;
		frame->curve_data[i].pcv_64qam = pa_curve_data->curve_samples[i].pcv;
		/* "invent" the points for the other modulations */
#define SUB(x,y) (uint8_t)((x) - (y)) > (x) ? 0 : (x) - (y)
		frame->curve_data[i].pcv_16qam = SUB(frame->curve_data[i].pcv_64qam, 12);
		frame->curve_data[i].pcv_qpsk = SUB(frame->curve_data[i].pcv_16qam, 12);
		frame->curve_data[i].pcv_bpsk = SUB(frame->curve_data[i].pcv_qpsk, 12);
		frame->curve_data[i].pcv_barker= SUB(frame->curve_data[i].pcv_bpsk, 14);
#undef SUB
		frame->curve_data[i].padding = 0;
	}


	bzero(&frame->padding1, sizeof(frame->padding1));

	frame->header.id1.code = P54U_LMAC_ID1_CODE_CONTROL;
	frame->header.id1.flags = P54U_LMAC_ID1_FLAGS_CONTROL_NOREADBACK;
	frame->header.id1.length = htole16(P54U_LMAC_ID1LEN(struct p54u_lmac_channel));
	frame->header.req_id = htole32(req_id);
	frame->header.type = htole16(P54U_LMAC_ID2_CONTROL_TYPE_CHANNEL);
	frame->header.flags = htole16(0);

	return m;
}

struct mbuf *p54u_lmac_set_filter(struct p54u_softc *sc, uint32_t req_id, uint16_t filter_type, const uint8_t *mac_source, const uint8_t *mac_destination, uint16_t rxhw, uint16_t magic2, uint32_t magic3, uint32_t magic4, uint32_t reserved_address, uint16_t magic8, uint16_t magic9)
{
	struct mbuf *m;
	struct p54u_lmac_filter *frame;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if(m == NULL)
		return NULL;
	m->m_pkthdr.len = m->m_len = sizeof(struct p54u_lmac_filter);
	frame = mtod(m, struct p54u_lmac_filter *);
	bzero(frame, sizeof(struct p54u_lmac_filter));

	frame->filter_type = htole16(filter_type);
	frame->rxhw = htole16(rxhw);
	frame->magic2 = htole16(magic2);
	frame->magic3 = htole32(magic3);
	frame->magic4 = htole32(magic4);

	frame->reserved_address = htole32(reserved_address);

	if(sc->device_type == P54U_DEVTYPE_USB2)
		frame->magic7 = htole32(P54U_TX_CONTROL_FILTER_MAGIC7_VER2);
	else
		frame->magic7 = htole32(P54U_TX_CONTROL_FILTER_MAGIC7_VER1);
	frame->magic8 = htole16(magic8);
	frame->magic9 = htole16(magic9);

	/* usually the device's MAC address */
	IEEE80211_ADDR_COPY(&frame->destination[0], mac_destination);
	/* usually the other MAC address */
	IEEE80211_ADDR_COPY(&frame->source[0], mac_source);

	frame->header.id1.code = P54U_LMAC_ID1_CODE_CONTROL;
	frame->header.id1.flags = P54U_LMAC_ID1_FLAGS_CONTROL_NOREADBACK;
	frame->header.id1.length = htole16(P54U_LMAC_ID1LEN(struct p54u_lmac_filter));
	frame->header.req_id = htole32(req_id);
	frame->header.type = htole16(P54U_LMAC_ID2_CONTROL_TYPE_FILTER);
	frame->header.flags = htole16(0);

	return m;
}

struct mbuf *p54u_lmac_80211tx(struct p54u_softc *sc, uint32_t req_id, struct mbuf *m, uint8_t id1_flags, uint32_t frame_type)
{
	struct p54u_lmac_80211tx *frame;
	uint16_t tx_len;

	tx_len = m->m_pkthdr.len;
	M_PREPEND(m, sizeof(struct p54u_lmac_80211tx), M_DONTWAIT);
	if(m == NULL)
		return NULL;
	m = m_pullup(m, sizeof(struct p54u_lmac_80211tx));
	if(m == NULL)
		return NULL;
	frame = mtod(m, struct p54u_lmac_80211tx *);
	bzero(frame, sizeof(struct p54u_lmac_80211tx));

	frame->frame_type = htole32(frame_type);
	frame->magic4 = htole32(0);
	frame->magic5 = htole32(0);
	frame->frame_padding = 4;

	frame->header.id1.code = P54U_LMAC_ID1_CODE_80211TXALIGN;
	frame->header.id1.flags = id1_flags;
	frame->header.id1.length = htole16(tx_len);
	frame->header.req_id = htole32(req_id);
	frame->header.type = htole16(P54U_LMAC_ID2_CONTROL_TYPE_DATANOACK);
	frame->header.flags = htole16(0);

	return m;
}

struct mbuf *p54u_lmac_stats_readreq(struct p54u_softc *sc, uint32_t req_id)
{
	struct mbuf *m;
	struct p54u_lmac_stats *frame;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if(m == NULL)
		return NULL;
	m->m_pkthdr.len = m->m_len = sizeof(struct p54u_lmac_stats);
	frame = mtod(m, struct p54u_lmac_stats *);
	/* bzero not required, but left for clean log */
	bzero(frame, sizeof(struct p54u_lmac_stats));

	frame->header.id1.code = P54U_LMAC_ID1_CODE_CONTROL;
	frame->header.id1.flags = 0;
	frame->header.id1.length = htole16(P54U_LMAC_ID1LEN(struct p54u_lmac_stats));
	frame->header.req_id = htole32(req_id);
	frame->header.type = htole16(P54U_LMAC_ID2_CONTROL_TYPE_STATS);
	frame->header.flags = htole16(0);

	return m;
}

/***********************/
/*** Input functions ***/
/***********************/

static void p54u_lmac_input_control_frequencydone(struct p54u_softc *sc, struct mbuf *m)
{
#ifdef WITH_DEBUG_GEN
	struct p54u_lmac_frequencydone *frequencydone;

	m = m_pullup(m, sizeof(struct p54u_lmac_frequencydone));
	if(m == NULL)
	{
		P54U_WARN("Unable to pull up mbuf for frequency done structure");
		return;
	}
	frequencydone = mtod(m, struct p54u_lmac_frequencydone *);

	P54U_DBG_GEN("Frequency sucessfully set to %d MHz", le16toh(frequencydone->frequency));
#endif

	m_freem(m);
}

static void p54u_lmac_input_control_stats(struct p54u_softc *sc, struct mbuf *m)
{
	struct p54u_lmac_stats *stats;

	m = m_pullup(m, sizeof(struct p54u_lmac_stats));
	if(m == NULL)
	{
		P54U_WARN("Unable to pull up mbuf for stats structure");
		return;
	}
	stats = mtod(m, struct p54u_lmac_stats *);

	p54u_handle_stats(sc, le32toh(stats->rx_packets), le32toh(stats->time), le32toh(stats->unk1), le32toh(stats->unk2), le32toh(stats->unk3), le32toh(stats->unk4), le32toh(stats->unk5), le32toh(stats->unk6), le32toh(stats->unk7));

	m_freem(m);
}

static void p54u_lmac_input_control_txdone(struct p54u_softc *sc, struct mbuf *m)
{
#ifdef WITH_DEBUG_GEN
	struct p54u_lmac_txdone *txdone;

	m = m_pullup(m, sizeof(struct p54u_lmac_txdone));
	if(m == NULL)
	{
		P54U_WARN("Unable to pull up mbuf for TX report structure");
		return;
	}
	txdone = mtod(m, struct p54u_lmac_txdone *);
	P54U_DBG_GEN("TX done : address 0x%08x, status 0x%04x, ack strength 0x%04x, sequence 0x%04x, unknown 0x%04x", le32toh(txdone->header.req_id), le32toh(txdone->status), le16toh(txdone->ack_rssi), le16toh(txdone->sequence), le16toh(txdone->unknown));
#endif

	m_freem(m);
}

static void p54u_lmac_input_control_eeprom(struct p54u_softc *sc, struct mbuf *m)
{
	struct p54u_lmac_eeprom *eeprom;
	uint16_t offset;

	m = m_pullup(m, sizeof(struct p54u_lmac_eeprom));
	if(m == NULL) {
		P54U_WARN("Unable to pull up mbuf for EEPROM header");
		return;
	}
	eeprom = mtod(m, struct p54u_lmac_eeprom *);

	offset = le16toh(eeprom->offset);
	m_adj(m, sizeof(struct p54u_lmac_eeprom));

	p54u_handle_eeprom(sc, m, offset);
}

static void p54u_lmac_input_control(struct p54u_softc *sc, struct mbuf *m)
{
	struct p54u_lmac_id2_control *id2;

	m = m_pullup(m, sizeof(struct p54u_lmac_id2_control));
	if(m == NULL) {
		P54U_WARN("Unable to pull up mbuf for control id2 header");
		return;
	}
	id2 = mtod(m, struct p54u_lmac_id2_control *);

	/* Now that the readback or TX has been done, the Prism address can be re-used. */
	/* We never free the control address. */
	if(le32toh(id2->req_id) != sc->control_address)	{
		p54u_pmfree(&sc->pmalloc_state, le32toh(id2->req_id));
		p54u_handle_device_available(sc);
	}

	switch(le16toh(id2->type)) {
	case P54U_LMAC_ID2_CONTROL_TYPE_FREQDONE:
		p54u_lmac_input_control_frequencydone(sc, m);
		break;
	case P54U_LMAC_ID2_CONTROL_TYPE_STATS:
		p54u_lmac_input_control_stats(sc, m);
		break;
	case P54U_LMAC_ID2_CONTROL_TYPE_TXDONE:
		p54u_lmac_input_control_txdone(sc, m);
		break;
	case P54U_LMAC_ID2_CONTROL_TYPE_EEPROM:
		p54u_lmac_input_control_eeprom(sc, m);
		break;
	default:
		P54U_WARN("Received control frame with unknown type 0x%04x, ignored", le16toh(id2->type));
		m_freem(m);
		break;
	}
}

static void p54u_lmac_input_80211(struct p54u_softc *sc, struct mbuf *m)
{
	struct p54u_lmac_id2_80211rx *id2;
	unsigned int freq, rate, antsignal, timestamp;

	m = m_pullup(m, sizeof(struct p54u_lmac_id2_80211rx));
	if(m == NULL) {
		P54U_WARN("Unable to pull up mbuf for 802.11 RX id2 header");
		return;
	}
	id2 = mtod(m, struct p54u_lmac_id2_80211rx *);

	if(!(id2->id1.flags & P54U_LMAC_ID1_FLAGS_RX_FCS_OK)) {
		P54U_DBG_GEN("Received packet with bad FCS, dropped");
		m_freem(m);
		return;
	}

	freq = htole16(id2->freq);
	antsignal = id2->antsignal;
	rate = sc->rate_table[id2->rate % sc->rate_count];
	timestamp = le32toh(id2->timestamp);

	m_adj(m, sizeof(struct p54u_lmac_id2_80211rx));

	p54u_handle_80211rx(sc, m, rate, freq, antsignal, timestamp);
}

static void p54u_lmac_input_frame(struct p54u_softc *sc, struct mbuf *m)
{
	struct p54u_lmac_id1 *id1;

	m = m_pullup(m, sizeof(struct p54u_lmac_id1));
	if(m == NULL) {
		P54U_WARN("Unable to pull up mbuf for id1 header");
		return;
	}
	id1 = mtod(m, struct p54u_lmac_id1 *);

	switch(id1->code) {
	case P54U_LMAC_ID1_CODE_CONTROL:
		p54u_lmac_input_control(sc, m);
		break;
	case P54U_LMAC_ID1_CODE_80211DATARX:
	case P54U_LMAC_ID1_CODE_80211MGMTRX:
		p54u_lmac_input_80211(sc, m);
		break;
	default:
		P54U_WARN("Received frame with unknown code 0x%02x, flags 0x%02x, ignored", id1->code, id1->flags);
		m_freem(m);
		break;
	}
}

/* Returns 1 if the frame's complete and ready to be parsed, 0 otherwise */
/* Positions sc->ongoing_frame_data, the passed mbuf should not be used anymore after the call */
static int p54u_reassemble_frame(struct p54u_softc *sc, struct mbuf *m)
{
	if(sc->device_type == P54U_DEVTYPE_USB1) {
#ifdef WITH_USB
		struct p54u_lmac_86_physical_rx *physical_rx;
		
		/* Version 1 devices sometimes send a frame over several URBs */
		/* we need to get the full size in the physical RX header, then reassemble the frame */
		if(sc->ongoing_frame) {
			/* We're in the middle of a frame : append the new URB to it */
			m_cat(sc->ongoing_frame_data, m);
			m->m_flags &= ~M_PKTHDR;
			m_fixhdr(sc->ongoing_frame_data);
			/* Is the frame complete ? */
			if((sc->ongoing_frame_data->m_pkthdr.len >= sc->ongoing_frame_fullsize) && ((sc->ongoing_frame_data->m_pkthdr.len - sc->ongoing_frame_fullsize) < 4)) {
				P54U_DBG_GEN("Frame reassembled");
				sc->ongoing_frame = 0;
				m_adj(sc->ongoing_frame_data, -(sc->ongoing_frame_data->m_pkthdr.len-sc->ongoing_frame_fullsize)); /* Strip padding */
				return 1;
			}
			/* Did we overflow the size indicated in the physical RX header ? */
			if(sc->ongoing_frame_data->m_pkthdr.len > sc->ongoing_frame_fullsize) {
				P54U_WARN("Read more data than indicated in the physical RX header, frame(s) dropped");
				m_freem(sc->ongoing_frame_data);
				sc->ongoing_frame = 0;
				return 0;
			}
			return 0;
		} else {
			/* New frame. Parse the physical RX header */
			sc->ongoing_frame_data = m;
			sc->ongoing_frame_data = m_pullup(sc->ongoing_frame_data, sizeof(struct p54u_lmac_86_physical_rx));
			if(sc->ongoing_frame_data == NULL) {
				P54U_WARN("Unable to pull up mbuf for physical RX header");
				return 0;
			}
			physical_rx = mtod(sc->ongoing_frame_data, struct p54u_lmac_86_physical_rx *);
			sc->ongoing_frame_fullsize = le16toh(physical_rx->length);
			m_adj(sc->ongoing_frame_data, sizeof(struct p54u_lmac_86_physical_rx)); /* We don't need this header anymore : strip it */
			//P54U_DBG_GEN("RX URB payload %d, header specified size %d", sc->ongoing_frame_data->m_pkthdr.len, sc->ongoing_frame_fullsize);
			m_adj(sc->ongoing_frame_data, -(sc->ongoing_frame_data->m_pkthdr.len-sc->ongoing_frame_fullsize)); /* Strip padding */
			/* >= to take padding into account */
			if(sc->ongoing_frame_data->m_pkthdr.len >= sc->ongoing_frame_fullsize) {
				/* Full frame */
				return 1;
			} else {
				/* Fragmented frame */
				sc->ongoing_frame = 1;
				return 0;
			}
		}
#else
		P54U_ERR("p54u_reassemble_frame with USB device, with USB support disabled");
		return 1;
#endif
	} else {
		/* 3887 sends one frame per URB, and no padding */
		/* TODO: PCI fragmentation */
		sc->ongoing_frame_data = m;
		return 1;
	}
}

void p54u_lmac_input(struct p54u_softc *sc, struct mbuf *m)
{
	if(!p54u_reassemble_frame(sc, m))
		return; /* The frame's fragmented over several URBs, we can't handle it for now */

	/* Okay, we've got a full frame in sc->ongoing_frame_data ; parse it */
	p54u_lmac_input_frame(sc, sc->ongoing_frame_data);
}

void p54u_lmac_init(struct p54u_softc *sc)
{
	P54U_DBG_GEN("Setting up SoftMAC protocol layer");
	sc->ongoing_frame = 0;
}

void p54u_lmac_free(struct p54u_softc *sc)
{
	P54U_DBG_GEN("Shutting down SoftMAC protocol layer");
	m_freem(sc->ongoing_frame_data);
}

#endif /* WITH_SOFTMAC */
