/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
  Copyright (C) 2004, 2005, 2006 Jean-Baptiste Note <jean-baptiste.note@m4x.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef __SOFTMAC_H
#define __SOFTMAC_H

#include <sys/types.h>

#include <net80211/ieee80211.h>

#include "eeprom.h"

/* p54u supports physical transports version 0.1 and 1.1 */
#define P54U_LMAC_PHYSICAL_V1            0x0001
#define P54U_LMAC_PHYSICAL_V2            0x0101

/* The firmware must have space before and after the frame to store information about its processing */
#define P54U_LMAC_HEADROOM              0x0070
#define P54U_LMAC_TAILROOM              0x0100

/* Amount of memory to allocate in the device for sending control frames */
#define P54U_LMAC_CONTROL_BUFFER_SIZE   0x0400
/* The device needs some space to store information about the frame it receives */
/* we must allocate it in the driver, and send the address in the RX filter packet */
/* FIXME: this should not be constant, for now we put the maximum observed size */
#define P54U_LMAC_RESERVED_SIZE         0x3500
#define P54U_LMAC_MAXRATES              20


/************************/
/*** Physical headers ***/
/************************/

/* Physical header for all transmitted packets (version 1 devices) */
struct p54u_lmac_86_physical_tx {
	uint32_t address;
	uint16_t length;
#define P54U_LM86_V1_MGMT_FOLLOWUP    0x2 /* no longer used, it seems */

	uint16_t follower;
	uint8_t  padding[8];
} __attribute__ ((packed));

/* Physical header for all received packets (version 1 devices) */
/* With these devices, one frame can be fragmented over several URBs */
/* we must use the "length" field of this header to reassemble the frame */
struct p54u_lmac_86_physical_rx {
	uint16_t     length;
	uint8_t      padding[14];
	char    data[];
} __attribute__ ((packed));

/* With physical transport 0.1 on the 3887 : */
/* One frame per URB : the device uses their length */
/* - TX: Send the address then the frame in two separate URBs */
/* - RX: No RX physical header */

/* Physical header for all transmitted packets (version 2 devices with physical transport v1.1) */
struct p54u_lmac_87_v11_physical_tx {
	uint32_t address;
	/* Frames must include a 32-bit signature on their content */
	/* It's not a CRC, it rather looks like a lame attempt at making reversing harder */
	/* I don't know how to compute it for now - seems based on bit operations on the 32-bit ARM words of the frame */
	/* Anyway, good old 0.1 still works with p54u, and at long term, SoftMAC should be obsolete :D */
	uint32_t signature;
} __attribute__ ((packed));

/* no RX physical header */

/******************/
/*** id1 header ***/
/******************/

/* Header identifying the type of packet, it is the same for TX and RX */
/* Immediately followed by an id2 header in the frame */
struct p54u_lmac_id1 {
	uint8_t flags;
	uint8_t code;

#define P54U_LMAC_ID1_CODE_80211DATARX    0x00
#define P54U_LMAC_ID1_CODE_80211MGMTRX    0x01
#define P54U_LMAC_ID1_CODE_80211TXALIGN   0x40
#define P54U_LMAC_ID1_CODE_80211TX        0x00
#define P54U_LMAC_ID1_CODE_CONTROL        0x80

	/* Flags for P54U_LMAC_CODE_80211_xxxxRX */
#define P54U_LMAC_ID1_FLAGS_RX_FCS_OK   0x01

	/* Flags for P54U_LMAC_CODE_80211_TX */
#define P54U_LMAC_ID1_FLAGS_MGMT   0x00
#define P54U_LMAC_ID1_FLAGS_DATA   0x10
#define P54U_LMAC_ID1_FLAGS_BEACON 0x02

	/* Flags for P54U_LMAC_CODE_CONFIG */
#define P54U_LMAC_ID1_FLAGS_CONTROL_NOREADBACK 0x01

	/* Length should be the length of data *after* the headers */
	uint16_t    length;
} __attribute__ ((packed));

/*******************/
/*** id2 headers ***/
/*******************/

/* id2 header suitable for all packets (TX and RX) with P54U_LMAC_ID1_CODE_CONTROL and P54U_LMAC_ID1_CODE_80211TX */
struct p54u_lmac_id2_control {
	struct p54u_lmac_id1 id1;

	/* This field is a unique number identifying the frame */
	/* Sending two frames with the same req_id will confuse the device */
	/* In p54u, we set req_id to the Prism address since this is convenient for keeping track of allocated addresses */
	/* The official driver sets it to a weird number, maybe an internal pointer */
	uint32_t     req_id;

	uint16_t     type;
	/* Types suitable for packets with P54U_LMAC_ID1_CODE_CONTROL */
#define P54U_LMAC_ID2_CONTROL_TYPE_FILTER            0x0000
#define P54U_LMAC_ID2_CONTROL_TYPE_CHANNEL           0x0001
#define P54U_LMAC_ID2_CONTROL_TYPE_FREQDONE          0x0002
#define P54U_LMAC_ID2_CONTROL_TYPE_TXPOWER           0x0003 /* not for sure */
#define P54U_LMAC_ID2_CONTROL_TYPE_ABORTADDR         0x0007
#define P54U_LMAC_ID2_CONTROL_TYPE_TXDONE            0x0008
#define P54U_LMAC_ID2_CONTROL_TYPE_STATS             0x000A
#define P54U_LMAC_ID2_CONTROL_TYPE_EEPROM            0x000C
#define P54U_LMAC_ID2_CONTROL_TYPE_LEDS              0x000D
	/* Types suitable for packets with P54U_LMAC_ID1_CODE_80211TX */
#define P54U_LMAC_ID2_CONTROL_TYPE_DATANOACK         0x0000
#define P54U_LMAC_ID2_CONTROL_TYPE_DATAACK           0x0001

	uint16_t     flags;
	/* Flags suitable for packets with P54U_LMAC_ID1_CODE_80211TX */
#define P54U_LMAC_ID2_CONTROL_FLAGS_ACK              0x0707
#define P54U_LMAC_ID2_CONTROL_FLAGS_NOACK            0x0101
} __attribute__ ((packed));

/* id2 header suitable for packets with P54U_LMAC_CODE_80211xxxxRX */
/* the 802.11 frame immediately follows */
struct p54u_lmac_id2_80211rx {
	struct p54u_lmac_id1 id1;
	uint16_t freq;      /* frequency, in MHz, le */
	uint8_t unk1;
	uint8_t rate;
	uint8_t antsignal;
	uint8_t padding;
	uint16_t unk2;
	uint32_t timestamp; /* timestamp, in microsec, le */
	uint32_t unk3;      /* Set for some control frames */
	char ieee80211_data[];
} __attribute__ ((packed));

/*************************************/
/*** p54u_lmac_id2_control payload ***/
/*************************************/

/* Abort Prism address packet */
struct p54u_lmac_abortaddress {
	struct p54u_lmac_id2_control header;

	uint32_t aborted_address;
} __attribute__ ((packed));

/* EEPROM packet */
struct p54u_lmac_eeprom {
	struct p54u_lmac_id2_control header;

	uint16_t        offset;
	uint16_t        len;
	char    data[];
} __attribute__ ((packed));

/* LEDs control packet */
struct p54u_lmac_leds {
	struct p54u_lmac_id2_control header;
	uint16_t mode;
#define P54U_LMAC_LED_MODE_SET       0x03
#define P54U_LMAC_LED_MODE_RXBLINK   0x01

	/* the leds will be under control of led_setting_temporary
	   for temporary_setting_duration time, in msec, then they
	   are set according to led_setting_permanent.
	   This is for both leds.
	   For instance, if you want the led A to blink while the
	   led B is still on, be sure to set it to LED_A | LED_B
	   otherwise the led B will be off during the blink */
	uint16_t led_setting_temporary;
	uint16_t led_setting_permanent;
#define P54U_LMAC_LED_A   0x0001
#define P54U_LMAC_LED_B   0x0002

	/* duration of the temporary settings, in ms */
	uint16_t temporary_setting_duration;
} __attribute__ ((packed));

/* Used to build the frequency setting packet */
struct p54u_lmac_iq_autocal_channel {
	uint16_t frequency;
	uint16_t iq_param_1;
	uint16_t iq_param_2;
	uint16_t iq_param_3;
	uint16_t iq_param_4;
} __attribute__ ((packed));

struct p54u_lmac_output_limits_channel {
	uint8_t limit_barker;
	uint8_t limit_bpsk;
	uint8_t limit_qpsk;
	uint8_t limit_16qam;
	uint8_t limit_64qam;
} __attribute__ ((packed));

struct p54u_lmac_pa_curve_data_sample {
	uint8_t rf_power;
	uint8_t pa_detector;
	uint8_t pcv_barker;
	uint8_t pcv_bpsk;
	uint8_t pcv_qpsk;
	uint8_t pcv_16qam;
	uint8_t pcv_64qam;
	uint8_t padding;
} __attribute__ ((packed));

#define P54U_LMAC_CURVE_SAMPLES 8

/* Frequency and TX policy setting packet */
struct p54u_lmac_channel {
	struct p54u_lmac_id2_control header;

	uint16_t magic1; /* This is a bitfield */
	uint16_t magic2; /* this is a timing -- maybe CIFS, beacon interval, dwell time ? */
	/* Setting when passive-scanning */
#define P54U_TX_CONTROL_CHANNEL_MAGIC1_SCAN           0x0002
#define P54U_TX_CONTROL_CHANNEL_MAGIC2_SCAN           0x0078
	// OR with default scan values when reaching channel 14
#define P54U_TX_CONTROL_CHANNEL_MAGIC1_SCAN_WRAP      0x0008

	/* Setting when active-scanning.
	 * Such a frequency change packet triggers
	 * the emission of the packet with flag P54U_FRAME_PROBE
	 */
#define P54U_TX_CONTROL_CHANNEL_MAGIC1_EMIT_TXSLOT    0x0006
#define P54U_TX_CONTROL_CHANNEL_MAGIC2_EMIT_TXSLOT    0x0028

	/* TX mode. When in this mode, subsequent TX packets with mode
	 * P54U_FRAME_DATA, P54U_FRAME_MGMT, P54U_FRAME_BEACON are directly
	 * emitted. A correct tx/rx filter is needed for BEACON type.
	 */
#define P54U_TX_CONTROL_CHANNEL_MAGIC1_TX             0x0001
#define P54U_TX_CONTROL_CHANNEL_MAGIC2_TX             0x0000

	/* This mode is used in ad-hoc mode, when you want to emit a packet
	 * (null data in dabe's log) while still scanning. Is this possible ?
	 * See for instance Dabe's log.
	 */
#define P54U_TX_CONTROL_CHANNEL_MAGIC1_SET2           0x3

	uint8_t  padding1[20];

	struct p54u_lmac_iq_autocal_channel iq_autocal;
	uint8_t n_points_in_pa_curve;
	struct p54u_lmac_output_limits_channel output_limits;
	struct p54u_lmac_pa_curve_data_sample curve_data[];
	/* + 32-bit zero'd padding - really needed ? */
} __attribute__ ((packed));

/* Successful frequency setting packet, RX only */
/* the readback has a different structure than the sent packet */
struct p54u_lmac_frequencydone {
	struct p54u_lmac_id2_control header;

	uint16_t padding;
	uint16_t frequency;
} __attribute__ ((packed));

/* RX/TX filter packet */
struct p54u_lmac_filter {
	struct p54u_lmac_id2_control header;

	uint16_t filter_type;
#define P54U_TX_CONTROL_FILTER_NOTYPE    0
#define P54U_TX_CONTROL_FILTER_STA       (1<<0)
#define P54U_TX_CONTROL_FILTER_ADHOC     (1<<1)
#define P54U_TX_CONTROL_FILTER_HOSTAP    (1<<2)
#define P54U_TX_CONTROL_FILTER_MONITOR   (1<<4)
	// Seen in many snoops : 1<<5. means reset ?
#define P54U_TX_CONTROL_FILTER_RESET     (1<<5)

	uint8_t destination[IEEE80211_ADDR_LEN];
	uint8_t source[IEEE80211_ADDR_LEN];
	uint16_t magic2;
	// FIXME / TODO : understand what those fields mean. If not
	// zeroed, version2 devices are going completely askew.
	// Do cache-coloring in order to understand if their value is
	// meaningfull in version1 devices.
#define P54U_TX_CONTROL_FILTER_MAGIC2_FORM1 0 //xde02
#define P54U_TX_CONTROL_FILTER_MAGIC2_FORM2 0 //xbd02
#define P54U_TX_CONTROL_FILTER_MAGIC2_FORM3 0 //xbd00
#define P54U_TX_CONTROL_FILTER_MAGIC2_STA     0 //xbd02
#define P54U_TX_CONTROL_FILTER_MAGIC2_MONITOR 0x0000

	uint32_t magic3;
#define P54U_TX_CONTROL_FILTER_MAGIC3_FORM1   0
#define P54U_TX_CONTROL_FILTER_MAGIC3_STA     0x15f
#define P54U_TX_CONTROL_FILTER_MAGIC3_MONITOR 0x0

	uint32_t magic4; /* rates */
#define P54U_TX_CONTROL_FILTER_MAGIC4_FORM1 0x01040608
#define P54U_TX_CONTROL_FILTER_MAGIC4_STA     0x0
#define P54U_TX_CONTROL_FILTER_MAGIC4_MONITOR 0x0

	uint32_t magic5; /* rest of the rates */
	/* where the device will store received frames before they are DMA'd */
	uint32_t reserved_address;
	uint16_t magic7;
	uint16_t rxhw;
#define P54U_TX_CONTROL_FILTER_MAGIC7_VER1 0x0610
#define P54U_TX_CONTROL_FILTER_MAGIC7_VER2 0x0620

	uint16_t magic8;
#define P54U_TX_CONTROL_FILTER_MAGIC8_STA   0x01f4

	uint16_t magic9;
#define P54U_TX_CONTROL_FILTER_MAGIC9_FORM1 0xf642
} __attribute__ ((packed));

/* 802.11 TX packet */
/* immediately followed by the 802.11 frame, without FCS (offloaded) */
struct p54u_lmac_wepkey {
	uint8_t      key_present;
#define P54U_WEPKEY_PRESENT 0x1

	uint8_t      key_length;
	uint8_t      key[16];
} __attribute__ ((packed));

struct p54u_lmac_80211tx {
	struct p54u_lmac_id2_control header;

	/* range [0-8] + bitmask 20 / 10 */
	uint8_t      rateset[8];
	/* maybe offload type */
	uint16_t     padding;
	struct p54u_lmac_wepkey wepkey;
	uint32_t     frame_type;
#define P54U_FRAME_DATA                 0x4
#define P54U_FRAME_MGMT                 0x2
#define P54U_FRAME_PROBE                0x1
#define P54U_FRAME_BEACON               0x0

	uint32_t     padding2;
	uint32_t     magic4;
	uint32_t     magic5;
	/* 02 7f 23 with 00 40 magic seems okay for id purposes */
	/* 02 is for some kind of mode ? */
	/* 0x7f000000, other form with 00 40 as magic, and 0x23 below (dabe's device) */
#define P54U_TX_FRAME_MAGIC4_MGMT       0x7f020000
	/* 0x33, other form with 10 40 as magic, and 7f02 above */
#define P54U_TX_FRAME_MAGIC5_MGMT       0x23
#define P54U_TX_FRAME_MAGIC4_PROBE      0x0
#define P54U_TX_FRAME_MAGIC5_PROBE      0x0
	/* 4 is the max padding */
	uint8_t frame_padding;
	uint8_t extra_padding[3]; /* not part of the protocol */
} __attribute__ ((packed));

/* packet received when a data packet is acked by the AP, or when the ack times out */
struct p54u_lmac_txdone {
	struct p54u_lmac_id2_control header;

	uint16_t      status;
	/* i'll have to make sure of this */
#define PKT_STATUS_OK        0x100
#define PKT_STATUS_NOTACKED  0x001

	uint16_t      ack_rssi;
	/* the device takes care of the sequence control field
	 * in the packet. Here is what it filled it with */
	uint16_t      sequence;
	/* this is, i think another field filled in by the device.
	 * duration / ID field, maybe ? Seems to be field 0x22-0x23
	 * repeated */
	uint16_t      unknown;
} __attribute__ ((packed));

/* statistics packet */
/* send one with all fields zeroed, and the device sends the filled frame back */
struct p54u_lmac_stats {
	struct p54u_lmac_id2_control header;

	uint32_t rx_packets;
	uint32_t unk1; /* counter */
	uint32_t unk2; /* counter */
	uint32_t unk3; /* counter */
	uint32_t unk4;
	uint32_t unk5;
	uint32_t time; /* in microseconds */
	uint32_t unk6; /* read straight from 0xa00000bc (D45bc) */
	uint32_t unk7;
} __attribute__ ((packed));

/**********************************/
/*** Macros for computing sizes ***/
/**********************************/

/* For all these macros : */
/* If a payload follows (i.e. EEPROM or 802.11 TX packets), its length must be added */

/* Gives the physical header length field required for a particular structure */
#define P54U_LM8X_V1_PHYSLEN(x) (sizeof(x))

/* Gives the id1 length field required for a particular structure */
#define P54U_LMAC_ID1LEN(x) (sizeof(x)-sizeof(struct p54u_lmac_id2_control))

/******************************/
/*** Size majoration macros ***/
/******************************/

#define p54u_lmac_abortaddress_size()                    (sizeof(struct p54u_lmac_abortaddress))
#define p54u_lmac_eeprom_readreq_size(length)            (sizeof(struct p54u_lmac_eeprom)+length)
#define p54u_lmac_leds_size()                            (sizeof(struct p54u_lmac_leds))
#define p54u_lmac_channel_change_size(size_pacurve)      (sizeof(p54u_lmac_channel)+size_pacurve)
#define p54u_lmac_set_filter_size()                      (sizeof(struct p54u_lmac_filter))
#define p54u_lmac_80211tx_size(size_80211)               (sizeof(struct p54u_lmac_80211tx)+size_80211+4) /* max padding */
#define p54u_lmac_stats_readreq_size()                   (sizeof(struct p54u_lmac_stats))

/**************************/
/*** Protocol functions ***/
/**************************/

struct p54u_softc;

#ifdef WITH_USB
int p54u_lmac_usb_transport(struct p54u_softc *sc, struct mbuf *m, uint32_t address);
#endif

struct mbuf *p54u_lmac_abortaddress(struct p54u_softc *sc, uint32_t req_id, uint32_t aborted_address);
struct mbuf *p54u_lmac_eeprom_readreq(struct p54u_softc *sc, uint32_t req_id, uint16_t offset, uint16_t length);
struct mbuf *p54u_lmac_leds(struct p54u_softc *sc, uint32_t req_id, uint16_t mode, uint16_t perm_setting, uint16_t temp_setting, uint16_t duration);

struct mbuf *p54u_lmac_set_filter(struct p54u_softc *sc, uint32_t req_id, uint16_t filter_type, const uint8_t *mac_source, const uint8_t *mac_destination, uint16_t rxhw, uint16_t magic2, uint32_t magic3, uint32_t magic4, uint32_t reserved_address, uint16_t magic8, uint16_t magic9);
struct mbuf *p54u_lmac_80211tx(struct p54u_softc *sc, uint32_t req_id, struct mbuf *m, uint8_t id1_flags, uint32_t frame_type);
struct mbuf *p54u_lmac_stats_readreq(struct p54u_softc *sc, uint32_t req_id);
struct mbuf *p54u_lmac_channel_change(struct p54u_softc *sc, uint32_t req_id, unsigned int channel, uint16_t magic1, uint16_t magic2);
void p54u_lmac_input(struct p54u_softc *sc, struct mbuf *m);

void p54u_lmac_init(struct p54u_softc *sc);
void p54u_lmac_free(struct p54u_softc *sc);

#endif /* __SOFTMAC_H */
