/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
  Copyright (C) 2004 Feyd
  Copyright (C) 2004, 2005, 2006 Jean-Baptiste Note <jean-baptiste.note@m4x.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "config.h"

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/endian.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/sema.h>
#include <sys/time.h>
#include <sys/interrupt.h>
#include <sys/socket.h>
#include <sys/mbuf.h>
#include <sys/selinfo.h>

#ifdef WITH_USB
#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>
#endif

#include "isl38xx_common.h"
#include "isl38xx_host.h"
#ifdef WITH_SOFTMAC
#include "softmac.h"
#endif
#ifdef WITH_FREEMAC
#include "freemac.h"
#endif
#include "devnode.h"
#include "pcidefs.h"
#include "transport.h"
#include "p54u.h"

/* Low-level functions */

#ifdef WITH_USB
int p54u_alloc_pipe(struct p54u_softc *sc, struct p54u_pipe *pipe, uint8_t address, int size)
{
	usbd_status r;

	pipe->addr = address;
	pipe->size = size;
	if((r = usbd_open_pipe(sc->bus.usb.sc_iface, pipe->addr, 0, &pipe->handle)) != USBD_NORMAL_COMPLETION) {
		P54U_ERR("Unable to open pipe 0x%02x (status = %d)", pipe->addr, r);
		pipe->handle = 0;
		return 0;
	}
	pipe->xfer = usbd_alloc_xfer(sc->bus.usb.sc_udev);
	if(pipe->xfer == NULL) {
		P54U_ERR("xfer allocation failed");
		usbd_close_pipe(pipe->handle);
		pipe->handle = 0;
		return 0;
	}
	pipe->buf = usbd_alloc_buffer(pipe->xfer, pipe->size);
	if(pipe->buf == NULL) {
		P54U_ERR("Buffer allocation failed");
		usbd_free_xfer(pipe->xfer);
		usbd_close_pipe(pipe->handle);
		pipe->handle = 0;
		return 0;
	}

	return 1;
}

int p54u_free_pipe(struct p54u_softc *sc, struct p54u_pipe *pipe)
{
	P54U_DBG_GEN("Freeing pipe 0x%02x", pipe->addr);
	if(pipe->handle == 0) {
		P54U_ERR("pipe->handle = 0 (pipe 0x%02x)", pipe->addr);
		return 0;
	}
	if(usbd_abort_pipe(pipe->handle) != USBD_NORMAL_COMPLETION) {
		P54U_ERR("Unable to abort pipe 0x%02x", pipe->addr);
		return 0;
	}
	if(pipe->buf == NULL) {
		P54U_ERR("pipe->buf = NULL (pipe 0x%02x)", pipe->addr);
		return 0;
	}
	usbd_free_buffer(pipe->xfer);
	if(pipe->xfer == 0) {
		P54U_ERR("pipe->xfer = 0 (pipe 0x%02x)", pipe->addr);
		return 0;
	}
	if(usbd_free_xfer(pipe->xfer) != USBD_NORMAL_COMPLETION) {
		P54U_ERR("Unable to free xfer on pipe 0x%02x", pipe->addr);
		return 0;
	}
	if(usbd_close_pipe(pipe->handle) != USBD_NORMAL_COMPLETION) {
		P54U_ERR("Unable to free pipe 0x%02x", pipe->addr);
		return 0;
	}
	pipe->handle = 0;
	pipe->xfer = 0;
	pipe->buf = NULL;
	return 1;
}

static inline int _p54u_alloc_standard_pipes(struct p54u_softc *sc)
{
	/* Allocate data endpoint 1, it is the same on all devices (v1/2) */
	if(!p54u_alloc_pipe(sc, &sc->bus.usb.data_tx, P54U_PIPE_DATA, P54U_MAX_DATA_FRAME_SIZE))
		return 0;
	if(!p54u_alloc_pipe(sc, &sc->bus.usb.data_rx, P54U_PIPE_DATA|UE_DIR_IN, P54U_MAX_DATA_FRAME_SIZE))
		return 0;

	if(sc->device_type == P54U_DEVTYPE_USB1) {
		/* Also allocate NET2280 endpoints on version 1 devices */
		if(!p54u_alloc_pipe(sc, &sc->bus.usb.dev_tx, P54U_PIPE_DEV, P54U_MAX_DEV_FRAME_SIZE))
			return 0;
		if(!p54u_alloc_pipe(sc, &sc->bus.usb.dev_rx, P54U_PIPE_DEV|UE_DIR_IN, P54U_MAX_DEV_FRAME_SIZE))
			return 0;
		if(!p54u_alloc_pipe(sc, &sc->bus.usb.brg_tx, P54U_PIPE_BRG, P54U_MAX_BRG_FRAME_SIZE))
			return 0;
		if(!p54u_alloc_pipe(sc, &sc->bus.usb.brg_rx, P54U_PIPE_BRG|UE_DIR_IN, P54U_MAX_BRG_FRAME_SIZE))
			return 0;
		if(!p54u_alloc_pipe(sc, &sc->bus.usb.int_rx, P54U_PIPE_INT|UE_DIR_IN, P54U_MAX_INT_FRAME_SIZE))
			return 0;
	}

	return 1;
}

int p54u_alloc_standard_pipes(struct p54u_softc *sc)
{
	P54U_DBG_GEN("Setting up USB structures");
	if(!_p54u_alloc_standard_pipes(sc)) {
		P54U_ERR("Failed to set up USB pipes");
		return 0;
	}
	P54U_DBG_GEN("Setup of the USB structures succeeded");
	return 1;
}

int p54u_free_standard_pipes(struct p54u_softc *sc)
{
	int ok;

	P54U_DBG_GEN("Shutting down USB endpoints");

	ok = 1;
	ok &= p54u_free_pipe(sc, &sc->bus.usb.data_tx);
	ok &= p54u_free_pipe(sc, &sc->bus.usb.data_rx);

	if(sc->device_type == P54U_DEVTYPE_USB1) {
		ok &= p54u_free_pipe(sc, &sc->bus.usb.dev_tx);
		ok &= p54u_free_pipe(sc, &sc->bus.usb.dev_rx);
		ok &= p54u_free_pipe(sc, &sc->bus.usb.brg_tx);
		ok &= p54u_free_pipe(sc, &sc->bus.usb.brg_rx);

		ok &= p54u_free_pipe(sc, &sc->bus.usb.int_rx);
	}

	if(!ok)
		P54U_WARN("Unable to free all endpoint resources, possible memory leak");

	return ok;
}

static void p54u_data_rx_cb(usbd_xfer_handle xfer, usbd_private_handle data, usbd_status status);
static void p54u_int_rx_cb(usbd_xfer_handle xfer, usbd_private_handle data, usbd_status status);

int p54u_data_rx_submit(struct p54u_softc *sc)
{
	usbd_setup_xfer(sc->bus.usb.data_rx.xfer, sc->bus.usb.data_rx.handle, (usbd_private_handle)sc, sc->bus.usb.data_rx.buf, sc->bus.usb.data_rx.size, USBD_NO_COPY|USBD_SHORT_XFER_OK, USBD_NO_TIMEOUT, p54u_data_rx_cb);
	if(usbd_transfer(sc->bus.usb.data_rx.xfer) != USBD_IN_PROGRESS) {
		P54U_ERR("Unable to submit DATA RX xfer");
		sc->err = 1;
		return 0;
	}
	return 1;
}

int p54u_int_rx_submit(struct p54u_softc *sc)
{
	usbd_setup_xfer(sc->bus.usb.int_rx.xfer, sc->bus.usb.int_rx.handle, (usbd_private_handle)sc, sc->bus.usb.int_rx.buf, sc->bus.usb.int_rx.size,  USBD_NO_COPY|USBD_SHORT_XFER_OK, USBD_NO_TIMEOUT, p54u_int_rx_cb);
	if(usbd_transfer(sc->bus.usb.int_rx.xfer) != USBD_IN_PROGRESS) {
		P54U_ERR("Unable to submit INT RX xfer");
		sc->err = 1;
		return 0;
	}
	return 1;
}

usbd_status p54u_bulk_transfer(struct p54u_softc *sc, struct p54u_pipe *pipe, uint32_t *size, uint16_t flags)
{
	usbd_status r;

	r = usbd_bulk_transfer(pipe->xfer, pipe->handle, USBD_NO_COPY|flags, P54U_TRANSFER_TIMEOUT, pipe->buf, size, "p54u_bulk_transfer");
#ifdef WITH_DEBUG_GEN
	if(debug_dumpxfers) {
		if(r == USBD_NORMAL_COMPLETION) {
			p54u_data_debug(sc, pipe->addr, pipe->buf, *size);
		} else {
			P54U_DBG_GEN("Bulk transfer status %d on pipe %02x", r, pipe->addr);
		}
	}
#endif
	return r;
}

#endif /* WITH_USB */

/* Reads and writes registers (ISL PCI space + NET2280 if applicable). Works with USB V1 + PCI */
uint32_t p54u_reg_rw(struct p54u_softc *sc, int write, int ep, int port, uint32_t addr, uint32_t val)
{
#ifdef WITH_USB
	if(sc->device_type == P54U_DEVTYPE_USB1) {
		struct p54u_pipe *pipe_tx, *pipe_rx;
		struct p54u_reg *reg;
		uint32_t len;
		uint32_t retval;
		uint32_t *r;
	
		if(sc->err)
			return 0;
	
		switch(ep) {
			case P54U_PIPE_DEV:
				pipe_tx = &sc->bus.usb.dev_tx;
				pipe_rx = &sc->bus.usb.dev_rx;
				break;
			case P54U_PIPE_BRG:
				pipe_tx = &sc->bus.usb.brg_tx;
				pipe_rx = &sc->bus.usb.brg_rx;
				break;
			default:
				P54U_ERR("Invalid endpoint in p54u_reg_rw");
				sc->err = 1;
				return 0;
		}
	
		reg = pipe_tx->buf;
		reg->port = htole16(port);
		reg->addr = htole32(addr);
		reg->val = ((port & P54U_PORT_U32) == P54U_PORT_U32) ? htole32(val) : htole32(val&0xffff);
		len = write ? P54U_REG_WSIZE : P54U_REG_RSIZE;
	
		if(p54u_bulk_transfer(sc, pipe_tx, &len, 0) != USBD_NORMAL_COMPLETION) {
			P54U_ERR("Could not send data in p54u_reg_rw");
			sc->err = 1;
			return 0;
		}
	
		if(write)
			return 1;
	
		len = sizeof(*r);
		if(p54u_bulk_transfer(sc, pipe_rx, &len, 0) != USBD_NORMAL_COMPLETION) {
			P54U_ERR("Could not receive data in p54u_reg_rw");
			sc->err = 1;
			return 0;
		}
		r = (uint32_t *)pipe_rx->buf;
		retval = ((port & P54U_PORT_U32) == P54U_PORT_U32) ? le32toh(*r) : (le32toh(*r)&0xffff);
		return retval;
	}
#endif
#ifdef WITH_PCI
	if(sc->device_type == P54U_DEVTYPE_PCI) {
		if(ep != P54U_PIPE_DEV) {
			P54U_WARN("Attempted to access bridge on a pure PCI device, ignored.\n");
			return 0;
		}
		addr &= 0xffff;
		if(write) {
			bus_space_write_4(sc->bus.pci.btag, sc->bus.pci.bhandle, addr, val);
			return 1;
		} else {
			return bus_space_read_4(sc->bus.pci.btag, sc->bus.pci.bhandle, addr);
		}
	}
#endif
	return 0;
}

int p54u_initialize_device(struct p54u_softc *sc)
{
	sc->firmware_running = 0;
	switch(sc->device_type) {
#ifdef WITH_USB
		case P54U_DEVTYPE_USB1: {
			/* VERSION 1 DEVICES */
			uint32_t reg;
	
			P54U_DBG_GEN("Intializing device");
	
			/* Reset the usb<->pci bridge */
			P54U_DBG_GEN("Reset bridge");
			reg = p54u_brg_readl(sc, NET2280_GPIOCTL);
			if(reg&P54U_BRG_POWER_UP)
				P54U_DBG_GEN("Bridge already up");
			p54u_brg_writel(sc, NET2280_GPIOCTL, (reg|P54U_BRG_POWER_DOWN) & ~P54U_BRG_POWER_UP);
			if(sc->err)
				return 0;
			p54u_delay(100);
			p54u_brg_writel(sc, NET2280_GPIOCTL, (reg|P54U_BRG_POWER_UP) & ~P54U_BRG_POWER_DOWN);
			if(sc->err)
				return 0;
			P54U_DBG_GEN("Reset bridge OK");
			p54u_delay(100);
	
			/* See 11.5.1 on net2280 doc */
			P54U_DBG_GEN("Setting up the PCI bus");
			p54u_brg_writel(sc, NET2280_DEVINIT, NET2280_CLK_30MHZ|NET2280_PCI_ENABLE|NET2280_PCI_SOFT_RESET);
			if(sc->err)
				return 0;
			P54U_DBG_GEN("DEVINIT OK");
			p54u_delay(20);
	
			/* Enable mmio and busmaster on the bridge */
			p54u_pcicfg_brg_writew(sc, PCI_COMMAND, (PCI_COMMAND_MEMORY|PCI_COMMAND_MASTER));
			if(sc->err)
				return 0;
			P54U_DBG_GEN("MMIO+BUSMASTER OK");
	
			/* Set base address 0 */
			P54U_DBG_GEN("Setting up registers base address");
			p54u_pcicfg_brg_writel(sc, PCI_BASE_ADDRESS_0, NET2280_BASE_REGISTERS);
			if(sc->err)
				return 0;
			P54U_DBG_GEN("Registers base address OK");
	
			/* Set PCI_STATUS_REC_MASTER_ABORT (why ?) */
			reg = p54u_pcicfg_brg_readw(sc, PCI_STATUS);
			p54u_pcicfg_brg_writew(sc, PCI_STATUS, (reg|PCI_STATUS_REC_MASTER_ABORT));
			if(sc->err)
				return 0;
			P54U_DBG_GEN("PCI_STATUS_REC_MASTER_ABORT OK");
	
			if(sc->err)
				return 0;
			P54U_DBG_GEN("PCI bus setup OK");
	
			P54U_DBG_GEN("Setting up bulk endpoints");
			p54u_brg_writel(sc, NET2280_EPA_RSP, NET2280_CLEAR_NAK_OUT_PACKETS_MODE);
			p54u_brg_writel(sc, NET2280_EPC_RSP, NET2280_CLEAR_NAK_OUT_PACKETS_MODE);
			if(sc->err)
				return 0;
			P54U_DBG_GEN("Endpoints setup OK");
	
			/* Set base address 2 for bridge : where the endpoint fifos are */
			P54U_DBG_GEN("Setting up endpoint FIFOs base address");
			p54u_pcicfg_brg_writel(sc, PCI_BASE_ADDRESS_2, NET2280_BASE_ENDPOINTS);
			if(sc->err)
				return 0;
			P54U_DBG_GEN("Endpoint FIFOs OK");
	
			/* Enable mmio and busmaster on the device */
			P54U_DBG_GEN("Setting up device PCI resources");
			p54u_pcicfg_dev_writew(sc, PCI_COMMAND|0x10000, (PCI_COMMAND_MEMORY|PCI_COMMAND_MASTER));
	
			/* Set TRDY_TIMEOUT and RETRY_TIMEOUT to 0 */
			p54u_pcicfg_dev_writew(sc, P54U_TRDY_TIMEOUT|0x10000, 0);
	
			/* Set base address 0 */
			p54u_pcicfg_dev_writel(sc, PCI_BASE_ADDRESS_0|0x10000, P54U_DEV_BASE);
			if(sc->err)
				return 0;
			P54U_DBG_GEN("Device PCI resources OK");
	
			/* See 7.6.5.5 and chapter 9 of net2280 doc */
			P54U_DBG_GEN("Setting up interrupt endpoint");
			p54u_brg_writel(sc, NET2280_USBIRQENB1, 0);
			p54u_brg_writel(sc, NET2280_IRQSTAT1, NET2280_PCI_INTA_INTERRUPT);
			if(sc->err)
				return 0;
			P54U_DBG_GEN("Endpoint setup OK");
	
			/* Assert wakeup */
			P54U_DBG_GEN("Asserting ISL38XX chipset");
			p54u_dev_writel(sc, (ISL38XX_HARDWARE_REG+ISL38XX_DEV_INT_REG), ISL38XX_INT_WAKEUP);
			if(sc->err)
				return 0;
	
			/* (see old driver for INT_ABORT definition) */
			p54u_dev_writel(sc, (ISL38XX_HARDWARE_REG+ISL38XX_DEV_INT_REG), ISL38XX_INT_ABORT);
			if(sc->err)
				return 0;
	
			P54U_DBG_GEN("Asserting device done");
	
			P54U_DBG_GEN("Flushing NET2280 endpoints");
			p54u_brg_writel(sc, NET2280_EPA_STAT, NET2280_FIFO_FLUSH);
			p54u_brg_writel(sc, NET2280_EPB_STAT, NET2280_FIFO_FLUSH);
			p54u_brg_writel(sc, NET2280_EPC_STAT, NET2280_FIFO_FLUSH);
			p54u_brg_writel(sc, NET2280_EPD_STAT, NET2280_FIFO_FLUSH);
			if(sc->err)
				return 0;
			P54U_DBG_GEN("NET2280 endpoints flushed");
			
			/* Reset the device */
			P54U_DBG_GEN("Stopping the ARM core");
			/* clear the RAMBoot, the CLKRUN and the Reset bit */
			reg = p54u_dev_readl(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG);
			reg &= ~(ISL38XX_CTRL_STAT_RESET|ISL38XX_CTRL_STAT_RAMBOOT|ISL38XX_CTRL_STAT_CLKRUN);
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG, reg);

			P54U_DBG_GEN("ARM core stopped");
			P54U_DBG_GEN("Resetting hardware");
			/* set the Reset bit without reading the register */
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG, reg|ISL38XX_CTRL_STAT_RESET);
			/* clear the Reset bit */
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG, reg);
			if(sc->err) {
				P54U_ERR("Unable to write to registers before upload");
				return 0;
			}
			/* wait for the device to reboot */;
			p54u_delay(50);
	
			P54U_DBG_GEN("Device initialization done");
			return 1;
		}
		case P54U_DEVTYPE_USB2: {
			/* VERSION 2 DEVICES */
			static const char init_string[4] = "~~~~";
			uint32_t size;
	
			P54U_DBG_GEN("Initializing device");
			size = sizeof(init_string);
			memcpy(sc->bus.usb.data_tx.buf, &init_string[0], size);
			if(p54u_bulk_transfer(sc, &sc->bus.usb.data_tx, &size, 0) != USBD_NORMAL_COMPLETION) {
				P54U_ERR("Could not send reset command");
				sc->err = 1;
				return 0;
			}
			p54u_delay(100);
			P54U_DBG_GEN("Device initialization done");
			return 1;
		}
#endif
#ifdef WITH_PCI
		case P54U_DEVTYPE_PCI: {
			uint32_t reg;
	
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_EN_REG, 0);DELAY(ISL38XX_WRITEIO_DELAY);
			DELAY(ISL38XX_WRITEIO_DELAY);
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_ACK_REG, 0xffffffff);
			DELAY(ISL38XX_WRITEIO_DELAY);
			
			/* Reset the device */
			P54U_DBG_GEN("Stopping the ARM core");
			/* clear the RAMBoot, and the Reset bit */
			reg = p54u_dev_readl(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG);
			reg &= ~(ISL38XX_CTRL_STAT_RESET|ISL38XX_CTRL_STAT_RAMBOOT);
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG, reg);
			DELAY(ISL38XX_WRITEIO_DELAY);
		
			P54U_DBG_GEN("ARM core stopped");
			P54U_DBG_GEN("Resetting hardware");
			/* set the Reset bit without reading the register */
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG, reg|ISL38XX_CTRL_STAT_RESET);
			DELAY(ISL38XX_WRITEIO_DELAY);
			/* clear the Reset bit */
			p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_CTRL_STAT_REG, reg);
			DELAY(ISL38XX_WRITEIO_DELAY);
			if(sc->err) {
				P54U_ERR("Unable to write to registers before upload");
				return 0;
			}
			/* wait for the device to reboot */
			p54u_delay(50);
			
			return 1;
		}
#endif
	}
	return 0;
}

/* PCI DMA */

#ifdef WITH_PCI

static void p54u_dbg_qindexes(struct p54u_softc *sc)
{
	P54U_DBG_GEN("CB drv Qs: [%d][%d][%d][%d]",
		le32toh(sc->bus.pci.cb->driver_curr_frag[0]),
		le32toh(sc->bus.pci.cb->driver_curr_frag[1]),
		le32toh(sc->bus.pci.cb->driver_curr_frag[2]),
		le32toh(sc->bus.pci.cb->driver_curr_frag[3]));
	P54U_DBG_GEN("CB dev Qs: [%d][%d][%d][%d]",
		le32toh(sc->bus.pci.cb->device_curr_frag[0]),
		le32toh(sc->bus.pci.cb->device_curr_frag[1]),
		le32toh(sc->bus.pci.cb->device_curr_frag[2]),
		le32toh(sc->bus.pci.cb->device_curr_frag[3]));
}

static void p54u_dma_map_addr2(void *arg, bus_dma_segment_t *segs, int nseg, bus_size_t mapsize, int error)
{
	uint32_t *paddr;

	if(error) return;

	paddr = arg;
	*paddr = segs->ds_addr;
}

static int p54u_alloc_one_dma_rx(struct p54u_softc *sc, struct p54u_queue *q, unsigned int index)
{
	uint32_t addr;
	
	q->membufs[index].m = m_getcl(M_DONTWAIT, MT_DATA, M_PKTHDR);
	if(q->membufs[index].m == NULL) return 0;
	
	q->membufs[index].m->m_len = q->membufs[index].m->m_pkthdr.len = MCLBYTES;
	
	if(bus_dmamap_load_mbuf(q->tag, q->membufs[index].map, q->membufs[index].m, p54u_dma_map_addr2, &addr, BUS_DMA_NOWAIT)) {
		m_freem(q->membufs[index].m);
		return 0;
	}
	
	bus_dmamap_sync(q->tag, q->membufs[index].map, BUS_DMASYNC_PREREAD);
	
	q->fragments[index].address = htole32(addr);
	q->fragments[index].arm_address = htole32(0);
	q->fragments[index].size = htole32(MCLBYTES);
	q->fragments[index].flags = htole32(0);
	
	P54U_DBG_GEN("Allocated DMA RX at 0x%08x, size 0x%04x, index %d", addr, MCLBYTES, index);
	
	return 1;
}

static int p54u_alloc_queue(struct p54u_softc *sc, struct p54u_queue *q, int type, unsigned int size, struct isl38xx_fragment *fragments, unsigned int qindex, unsigned int maxframe)
{
	unsigned int i;
	
	if(bus_dma_tag_create(NULL, 4, 0, BUS_SPACE_MAXADDR_32BIT, BUS_SPACE_MAXADDR, NULL, NULL, size*maxframe, 1, size*maxframe, 0, NULL, NULL, &q->tag)) return 0;
	
	q->type = type;
	q->size = size;
	q->fragments = fragments;
	q->free_index = 0;
	q->device_index = &sc->bus.pci.cb->device_curr_frag[qindex];
	q->driver_index = &sc->bus.pci.cb->driver_curr_frag[qindex];
	
	q->membufs = malloc(size*sizeof(struct p54u_dma_membuf), M_P54U, M_NOWAIT);
	if(q->membufs == NULL) {
		bus_dma_tag_destroy(q->tag);
		return 0;
	}
	
	for(i=0;i<size;i++) {
		if(bus_dmamap_create(q->tag, 0, &q->membufs[i].map)) {
			free(q->membufs, M_P54U);
			bus_dma_tag_destroy(q->tag);
			return 0;
		}
		if(type == P54U_QUEUE_TYPE_RX) {
			if(!p54u_alloc_one_dma_rx(sc, q, i)) {
				free(q->membufs, M_P54U);
				bus_dma_tag_destroy(q->tag);
				return 0;
			}
		}
	}
	
	if(type == P54U_QUEUE_TYPE_RX) *(q->driver_index) = htole32(size);
	
	return 1;
}

static void p54u_free_queue(struct p54u_softc *sc, struct p54u_queue *q)
{
	unsigned int i;
	
	for(i=0;i<q->size;i++) {
		if(q->membufs[i].m != NULL) {
			P54U_DBG_GEN("Unloading DMA map");
			bus_dmamap_unload(q->tag, q->membufs[i].map);
			m_freem(q->membufs[i].m);
			q->membufs[i].m = NULL;
		}
		P54U_DBG_GEN("Destroying DMA map");
		bus_dmamap_destroy(q->tag, q->membufs[i].map);
	}
	P54U_DBG_GEN("Destroying DMA tag");
	bus_dma_tag_destroy(q->tag);
	P54U_DBG_GEN("Cleanup");
	bzero(q->fragments, sizeof(struct isl38xx_fragment));
	q->free_index = 0;
	*(q->device_index) = htole32(0);
	*(q->driver_index) = htole32(0);
	free(q->membufs, M_P54U);
}

static void p54u_process_queue(struct p54u_softc *sc, struct p54u_queue *q)
{
	int update;
	unsigned int buf_index;
	struct mbuf *m;
	
	update = 0;
	while(q->free_index < le32toh(*(q->device_index))) {
		buf_index = q->free_index % q->size;
		
		bus_dmamap_sync(q->tag, q->membufs[buf_index].map, q->type == P54U_QUEUE_TYPE_RX ? BUS_DMASYNC_POSTREAD : BUS_DMASYNC_POSTWRITE);
		bus_dmamap_unload(q->tag, q->membufs[buf_index].map);
		m = q->membufs[buf_index].m;
		
		if(q->type == P54U_QUEUE_TYPE_RX) {
			m->m_pkthdr.len = m->m_len = le32toh(q->fragments[buf_index].size);
			
#ifdef WITH_DEBUG_GEN
			if(debug_dumpxfers)
				p54u_data_debug(sc, 2, m->m_data, m->m_len);
#endif
		
#ifdef WITH_SOFTMAC
			if(P54U_IS_SOFTMAC(sc))
				p54u_lmac_input(sc, m);
#endif
#ifdef WITH_FREEMAC
			if(P54U_IS_FREEMAC(sc))
				p54u_fmac_input(sc, m);
#endif
		} else {
			m_freem(m);
		}
		
		q->membufs[buf_index].m = NULL;
		
		q->free_index++;
		
		if(q->type == P54U_QUEUE_TYPE_RX) {
			p54u_alloc_one_dma_rx(sc, q, le32toh(*(q->driver_index)) % q->size);
			*(q->driver_index) = htole32(le32toh(*(q->driver_index))+1);
			update = 1;
		}
		
	}
	
	if(update) {
		bus_dmamap_sync(sc->bus.pci.cb_tag, sc->bus.pci.cb_map, BUS_DMASYNC_PREWRITE);
		p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_DEV_INT_REG, ISL38XX_INT_UPDATE);
	}
}

int p54u_queue_can_submit(struct p54u_softc *sc, struct p54u_queue *q)
{
	return (le32toh(*(q->driver_index)) - q->free_index) < q->size;
}

int p54u_queue_submit(struct p54u_softc *sc, struct p54u_queue *q, struct mbuf *m, uint32_t arm_address)
{
	int index;
	uint32_t pci_address;
	
	if(!p54u_queue_can_submit(sc, q)) {
		P54U_WARN("Cannot sumbit DMA buffer for TX (no space left in queue)");
		return 0;
	}
	P54U_DBG_GEN("TX submit : driver index %d size %d", *(q->driver_index), q->size);
	index = le32toh(*(q->driver_index)) % q->size;
	
#ifdef WITH_DEBUG_GEN
	if(debug_dumpxfers)
		p54u_data_debug(sc, 1, m->m_data, m->m_len);
#endif
	
	P54U_DBG_GEN("DMA TX index %d", index);
	
	if(bus_dmamap_load_mbuf(q->tag, q->membufs[index].map, m, p54u_dma_map_addr2, &pci_address, 0)) return 0;
	bus_dmamap_sync(q->tag, q->membufs[index].map, BUS_DMASYNC_PREWRITE);
	
	q->membufs[index].m = m;
	q->fragments[index].address = htole32(pci_address);
	q->fragments[index].arm_address = htole32(arm_address);
	q->fragments[index].size = htole32(m->m_pkthdr.len);
	q->fragments[index].flags = htole32(0);
	
	*(q->driver_index) = htole32(le32toh(*(q->driver_index))+1);
	P54U_DBG_GEN("TX submit : new driver index %d", le32toh(*(q->driver_index))); 
	
	bus_dmamap_sync(sc->bus.pci.cb_tag, sc->bus.pci.cb_map, BUS_DMASYNC_PREWRITE);
	p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_DEV_INT_REG, ISL38XX_INT_UPDATE);
	
	P54U_DBG_GEN("Device notified of new DMA");
	p54u_dbg_qindexes(sc);
	
	return 1;
}

int p54u_alloc_queues(struct p54u_softc *sc)
{
	if(!p54u_alloc_queue(sc, &sc->bus.pci.data_rx_hipri, P54U_QUEUE_TYPE_RX, ISL38XX_CB_RX_DATAHIPRI_QSIZE, sc->bus.pci.cb->rx_hipri, ISL38XX_CB_RX_DATAHIPRI, MCLBYTES)) goto err_1;
	if(!p54u_alloc_queue(sc, &sc->bus.pci.data_tx_hipri, P54U_QUEUE_TYPE_TX, ISL38XX_CB_TX_DATAHIPRI_QSIZE, sc->bus.pci.cb->tx_hipri, ISL38XX_CB_TX_DATAHIPRI, P54U_MAX_DATA_FRAME_SIZE)) goto err_2;
	if(!p54u_alloc_queue(sc, &sc->bus.pci.data_rx, P54U_QUEUE_TYPE_RX, ISL38XX_CB_RX_DATA_QSIZE, sc->bus.pci.cb->rx, ISL38XX_CB_RX_DATA, MCLBYTES)) goto err_3;
	if(!p54u_alloc_queue(sc, &sc->bus.pci.data_tx, P54U_QUEUE_TYPE_TX, ISL38XX_CB_TX_DATA_QSIZE, sc->bus.pci.cb->tx, ISL38XX_CB_TX_DATA, P54U_MAX_DATA_FRAME_SIZE)) goto err_4;
	
	return 1;

err_4:
	p54u_free_queue(sc, &sc->bus.pci.data_rx);
err_3:
	p54u_free_queue(sc, &sc->bus.pci.data_tx_hipri);
err_2:
	p54u_free_queue(sc, &sc->bus.pci.data_rx_hipri);
err_1:
	return 0;
}

void p54u_free_queues(struct p54u_softc *sc)
{
	p54u_free_queue(sc, &sc->bus.pci.data_rx_hipri);
	p54u_free_queue(sc, &sc->bus.pci.data_tx_hipri);
	p54u_free_queue(sc, &sc->bus.pci.data_rx);
	p54u_free_queue(sc, &sc->bus.pci.data_tx);
}

static void p54u_pci_update(struct p54u_softc *sc)
{
	if(sc->device_type != P54U_DEVTYPE_PCI) {
		P54U_WARN("Received UPDATE interrupt from non-PCI device");
		return;
	}
	P54U_DBG_GEN("IRQ: Update flag");
	bus_dmamap_sync(sc->bus.pci.cb_tag, sc->bus.pci.cb_map, BUS_DMASYNC_POSTREAD);
	
	p54u_dbg_qindexes(sc);
	
	p54u_process_queue(sc, &sc->bus.pci.data_rx_hipri);
	p54u_process_queue(sc, &sc->bus.pci.data_tx_hipri);
	p54u_process_queue(sc, &sc->bus.pci.data_rx);
	p54u_process_queue(sc, &sc->bus.pci.data_tx);
	
	bus_dmamap_sync(sc->bus.pci.cb_tag, sc->bus.pci.cb_map, BUS_DMASYNC_PREREAD);
}

#endif

/* UARTPCI */

/* Called when the firmware is ready to accept data from us (with Giant and TX locks) */
static void p54u_uart_cts(struct p54u_softc *sc)
{
	uint8_t data;

	if(!sc->devnode_initialized) {
		P54U_WARN("Unexpected CTS interrupt");
		return;
	}

	P54U_DBG_GEN("CTS");

	if(p54u_devnode_dequeue(sc, &data, 1) == 1) {
		p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_GEN_PURP_COM_REG_1, data);
		/* Give DR */
		p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_DEV_INT_REG, ISL38XX_INT_PCIUART_DR);
	} else {
		sc->uart_cts = 1;
	}
}

/* Called when the firmware wants to send data to us (with Giant and TX locks) */
static void p54u_uart_dr(struct p54u_softc *sc)
{
	uint8_t data;

	if(!sc->devnode_initialized) {
		P54U_WARN("Unexpected DR interrupt");
		return;
	}

	P54U_DBG_GEN("DR");

	data = p54u_dev_readl(sc, ISL38XX_HARDWARE_REG+ISL38XX_GEN_PURP_COM_REG_2) & 0xff;
	if(p54u_devnode_enqueue(sc, &data, 1) == 1) {
		/* Give CTS */
		p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_DEV_INT_REG, ISL38XX_INT_PCIUART_CTS);
	}
}

int p54u_uart_start(struct p54u_softc *sc)
{
	uint32_t reg;

	P54U_DBG_GEN("Enabling UARTPCI protocol");

	p54u_tx_lock(sc);

	/* Make sure CTS/DR interrupts are disabled */
	reg = p54u_dev_readl(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_EN_REG);
	reg &= ~(ISL38XX_INT_PCIUART_CTS|ISL38XX_INT_PCIUART_DR);
	p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_EN_REG, reg);

	/* ACK any previous CTS/DR interrupt */
	p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_ACK_REG, ISL38XX_INT_PCIUART_CTS|ISL38XX_INT_PCIUART_DR);

	/* Write magic to activate UART via PCI protocol */
	p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_GEN_PURP_COM_REG_1, ISL38XX_UARTPCI_MAGIC);
	p54u_delay(50);

	/* Check if the protocol is initialized */
	reg = p54u_dev_readl(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_IDENT_REG);
	P54U_DBG_GEN("After UARTPCI activation, ISL38XX interruption ident is %08x", reg);
	reg &= ISL38XX_INT_PCIUART_CTS;

	if(sc->err)  {
		P54U_ERR("PCI bus failed during UARTPCI initialization (before firmware response)");
		p54u_tx_unlock(sc);
		return 0;
	}

	if(reg == 0) {
		P54U_WARN("No firmware response to UARTPCI magic number");
		p54u_tx_unlock(sc);
		return 0;
	}

	/* Remove magic, so that the device stops sending CTS (SoftMAC behaves like that) */
	p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_GEN_PURP_COM_REG_1, 0);

	/* ACK the interrupt */
	p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_ACK_REG, ISL38XX_INT_PCIUART_CTS);

	P54U_DBG_GEN("UARTPCI protocol initialized");

	/* Enable CTS/DR interrupts */
	reg = p54u_dev_readl(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_EN_REG);
	reg |= (ISL38XX_INT_PCIUART_CTS|ISL38XX_INT_PCIUART_DR);
	p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_EN_REG, reg);

	/* And give CTS to the device */
	p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_DEV_INT_REG, ISL38XX_INT_PCIUART_CTS);

	sc->uart_cts = 1;

	if(sc->err) {
		P54U_ERR("PCI bus failed during UARTPCI initialization (after firmware response)");
		p54u_tx_unlock(sc);
		return 0;
	}

	p54u_tx_unlock(sc);
	return 1;
}

/* IRQ handler */

static void p54u_irq_task(void *context, int pending)
{
	struct p54u_softc *sc = (struct p54u_softc *)context;
	uint32_t inten;
	uint32_t reg;

	P54U_DBG_GEN("Entering interrupt task");

	p54u_tx_lock(sc);
#ifdef WITH_USB
	if(sc->device_type == P54U_DEVTYPE_USB1) mtx_lock(&Giant);
#endif

	/* Disable interrupts in the Prism chip */
	inten = p54u_dev_readl(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_EN_REG);
	p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_EN_REG, 0);

	/* Ack the interrupts */
	reg = p54u_dev_readl(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_IDENT_REG);
	if(sc->err) {
		P54U_ERR("Failed to read the ISL38XX interruption ident information");
		mtx_unlock(&Giant);
		p54u_tx_unlock(sc);
		return;
	}
	P54U_DBG_GEN("ISL38XX interruption ident is 0x%08x", reg);
	reg &= inten; /* Ignore unwanted interrupts */
	P54U_DBG_GEN("Taking 0x%08x into account", reg);
	p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_ACK_REG, reg);
	if(reg & ISL38XX_INT_INIT)
		sema_post(&sc->init_sema);
#ifdef WITH_PCI
	if(reg & ISL38XX_INT_UPDATE)
		p54u_pci_update(sc);
#endif
	if(reg & ISL38XX_INT_PCIUART_CTS)
		p54u_uart_cts(sc);
	if(reg & ISL38XX_INT_PCIUART_DR)
		p54u_uart_dr(sc);

#ifdef WITH_USB
	/* Clear the interrupt at NET2280 level */
	if(sc->device_type == P54U_DEVTYPE_USB1) {
		p54u_brg_writel(sc, NET2280_IRQSTAT1, NET2280_PCI_INTA_INTERRUPT);
		if(sc->err) {
			if(sc->device_type == P54U_DEVTYPE_USB1) mtx_unlock(&Giant);

			p54u_tx_unlock(sc);
			P54U_ERR("Failed to clear NET2280 interrupt");
			return;
		}
	}
#endif
	/* Re-enable interrupts in the Prism chip */
	p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_INT_EN_REG, inten);
	if(sc->err) {
#ifdef WITH_USB
		if(sc->device_type == P54U_DEVTYPE_USB1) mtx_unlock(&Giant);
#endif
		p54u_tx_unlock(sc);
		P54U_ERR("Failed to re-enable ISL38XX interrupt");
		return;
	}
#ifdef WITH_USB
	if(sc->device_type == P54U_DEVTYPE_USB1) mtx_unlock(&Giant);
#endif
	p54u_tx_unlock(sc);

	P54U_DBG_GEN("Leaving interrupt task");
}

#ifdef WITH_PCI
static void p54u_pci_intr(void *arg);
#endif

int p54u_irq_init(struct p54u_softc *sc)
{
	P54U_DBG_GEN("Setting up IRQ handler");
	TASK_INIT(&sc->irq_task, 0, p54u_irq_task, sc);
#ifdef WITH_USB
	if(sc->device_type == P54U_DEVTYPE_USB1) {
		P54U_DBG_GEN("Submitting initial INT RX xfer");
		if(!p54u_int_rx_submit(sc))
			return 0;
	}
#endif
#ifdef WITH_PCI
	if(sc->device_type == P54U_DEVTYPE_PCI) {
	if(bus_setup_intr(sc->bus.pci.dev, sc->bus.pci.interrupt, INTR_TYPE_NET|INTR_MPSAFE, p54u_pci_intr, sc, &sc->bus.pci.interrupt_handler))
			return 0;
	}
#endif
	return 1;
}

int p54u_irq_free(struct p54u_softc *sc)
{
	P54U_DBG_GEN("Shutting down interrupt handler");
	/* Make sure the taskqueue will not be triggered again */
#ifdef WITH_USB
	if(sc->device_type == P54U_DEVTYPE_USB1) usbd_abort_pipe(sc->bus.usb.int_rx.handle);
#endif
#ifdef WITH_PCI
	if(sc->device_type == P54U_DEVTYPE_PCI) bus_teardown_intr(sc->bus.pci.dev, sc->bus.pci.interrupt, sc->bus.pci.interrupt_handler); 
#endif
	return 1;
}

/* Devnode callbacks */

int p54u_devnode_onread(struct p54u_softc *sc)
{
	return 1;
}

int p54u_devnode_onwrite(struct p54u_softc *sc)
{
	if((sc->device_type == P54U_DEVTYPE_USB1) || (sc->device_type == P54U_DEVTYPE_PCI)) {
		if(sc->uart_cts) {
			sc->uart_cts = 0;
			p54u_tx_lock(sc);
			p54u_uart_cts(sc);
			p54u_tx_unlock(sc);
		}
	}
	return 1;
}

#ifdef WITH_USB

/* Synchronous function that sends a frame contained in mbuf m to the firmware */
/* Will free the mbuf */
/* Must aquire appropriate locks before called */
int p54u_usb_transport(struct p54u_softc *sc, struct mbuf *m)
{
	int len;

	len = roundup(len, sizeof(uint32_t));

	switch(sc->device_type) {
		case P54U_DEVTYPE_USB1:
			/* Send the announcement IRQ to the Prism chip  */
			if(!p54u_dev_writel(sc, ISL38XX_HARDWARE_REG+ISL38XX_DEV_INT_REG, ISL38XX_INT_DATA)) {
				P54U_ERR("Unable to send DATA interrupt to the Prism chipset");
				m_freem(m);
				return 0;
			}
			/* fall through */
		case P54U_DEVTYPE_USB2: {
			usbd_status err;
			
			/* Load data into the DMA buffer and free the mbuf chain */
			m_copydata(m, 0, m->m_pkthdr.len, sc->bus.usb.data_tx.buf);
			len = m->m_pkthdr.len;
			m_freem(m);
			if((err = p54u_bulk_transfer(sc, &sc->bus.usb.data_tx, &len, 0)) != USBD_NORMAL_COMPLETION) {
				P54U_ERR("Error on DATA TX pipe in p54u_announced_message, status = %d", err);
				sc->err = 1;
				return 0;
			}
			break;
		}
	}
	return 1;
}

/* Callbacks */

static void p54u_int_rx_cb(usbd_xfer_handle xfer, usbd_private_handle data, usbd_status status)
{
	struct p54u_softc *sc = (struct p54u_softc *)data;
#ifdef WITH_DEBUG_GEN
	unsigned int len;
#endif

	if(status != USBD_NORMAL_COMPLETION) {
		P54U_DBG_GEN("INT RX abnormal status (%d)", status);
		return;
	}

#ifdef WITH_DEBUG_GEN
	usbd_get_xfer_status(sc->bus.usb.int_rx.xfer, NULL, NULL, &len, NULL);
	if(debug_dumpxfers)
		p54u_data_debug(sc, sc->bus.usb.int_rx.addr, sc->bus.usb.int_rx.buf, len);
#endif

	taskqueue_enqueue(taskqueue_thread, &sc->irq_task);

	p54u_int_rx_submit(sc);
}

static void p54u_data_rx_cb(usbd_xfer_handle xfer, usbd_private_handle data, usbd_status status)
{
	struct p54u_softc *sc = (struct p54u_softc *)data;
	unsigned int actual_size;
	struct mbuf *m;
	char *p;

	if(status != USBD_NORMAL_COMPLETION) {
		P54U_DBG_GEN("DATA RX abnormal status (%d)", status);
		return;
	}

	usbd_get_xfer_status(sc->bus.usb.data_rx.xfer, NULL, NULL, &actual_size, NULL);

#ifdef WITH_DEBUG_GEN
	if(debug_dumpxfers)
		p54u_data_debug(sc, sc->bus.usb.data_rx.addr, sc->bus.usb.data_rx.buf, actual_size);
#endif

	/* Handle crashes involving ROM sending a memory dump - 3887 only */
	if(sc->device_type == P54U_DEVTYPE_USB2) {
		if(sc->crash_data != NULL) {
			P54U_WARN("Received more crash dump data (%d bytes)", actual_size);
			p = realloc(sc->crash_data, sc->crash_data_size+actual_size, M_P54U, M_NOWAIT);
			if(p == NULL) {
				P54U_ERR("Could not realloc crash data");
				free(sc->crash_data, M_P54U);
				sc->crash_data = NULL;
				return;
			}
			sc->crash_data = p;
			bcopy(sc->crash_data+sc->crash_data_size, sc->bus.usb.data_rx.buf, actual_size);
			sc->crash_data_size += actual_size;
			p54u_data_rx_submit(sc);
			return;
		}
		if((actual_size >= sizeof(uint32_t)) && (memcmp(sc->bus.usb.data_rx.buf, ISL3887_ROM_SIGNATURE, ISL3887_ROM_SIGNATURE_SIZE) == 0)) {
			P54U_ERR("Internal device error, saving crash dump (%d bytes so far)", actual_size);
			sc->err = 1;
			sc->crash_data = malloc(actual_size, M_P54U, M_NOWAIT);
			if(sc->crash_data == NULL) {
				P54U_WARN("Unable to allocate memory for crash dump");
				return;
			}
			bcopy(sc->crash_data, sc->bus.usb.data_rx.buf, actual_size);
			sc->crash_data_size = actual_size;
			p54u_data_rx_submit(sc);
			return;
		}
	}

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if(m == NULL) {
		P54U_WARN("First mbuf allocation failed in p54u_data_rx_cb, packet lost");
		p54u_data_rx_submit(sc);
		return;
	}
	m->m_pkthdr.len = m->m_len = 0;
	m_copyback(m, 0, actual_size, sc->bus.usb.data_rx.buf);
#ifdef WITH_SOFTMAC
	if(P54U_IS_SOFTMAC(sc))
		p54u_lmac_input(sc, m);
#endif
#ifdef WITH_FREEMAC
	if(P54U_IS_FREEMAC(sc))
		p54u_fmac_input(sc, m);
#endif

	p54u_data_rx_submit(sc);
}

#endif

#ifdef WITH_PCI

static void p54u_pci_intr(void *arg)
{
	taskqueue_enqueue(taskqueue_thread, &((struct p54u_softc *)arg)->irq_task);
}

#endif
