/*
  Copyright (C) 2005, 2006 Sebastien Bourdeauducq <lekernel@prism54.org>
 
  This file is part of p54u.
 
  p54u is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  p54u is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with p54u ; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/param.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <p54_ioctl.h>

#define STATS_SHOW_UNKNOWN

static int ioctl_err(int fd, unsigned long request, void *addr)
{
	if(ioctl(fd, request, addr) == -1) {
		perror("ioctl");
		if(errno == EAGAIN)
			fprintf(stderr, "Are you sure you have uploaded firmware (p54ctl -u/-1/-2) ?\n");
		return 0;
	}
	return 1;
}

static int quiet = 0;

static int reset_device(int fd)
{
	if(!quiet) {
		printf(">> Resetting device...");
		fflush(stdout);
	}
	if(!ioctl(fd, P54RESET) == -1) {
		printf("\n");
		perror("ioctl");
		return 0;
	}
	if(!quiet)
		printf("done\n");
	return 1;
}

static int upload_firmware(int fd, const char *filename, int version)
{
	FILE *rfd;
	struct p54_firmware firmware;

	if(version != -1) {
		struct p54_hwinfo hwinfo;

		if(!ioctl_err(fd, P54HWINFO, &hwinfo))
			return 0;
		if(hwinfo.device_type != version)
			return 1;
	}
	if(!quiet) {
		printf(">> Uploading firmware: '%s'...", filename);
		fflush(stdout);
	}
	rfd = fopen(filename, "rb");
	if(rfd == NULL) {
		perror("fopen");
		return 0;
	}
	fseek(rfd, 0, SEEK_END);
	firmware.size = ftell(rfd);
	fseek(rfd, 0, SEEK_SET);
	firmware.data = malloc(firmware.size);
	if(firmware.data == NULL) {
		perror("malloc");
		fclose(rfd);
		return 0;
	}
	if(fread(firmware.data, firmware.size, 1, rfd) != 1) {
		perror("read");
		fclose(rfd);
		return 0;
	}
	fclose(rfd);
	if(!ioctl_err(fd, P54FIRMWARE, &firmware)) {
		free(firmware.data);
		return 0;
	}
	free(firmware.data);
	if(!quiet)
		printf("done\n");
	return 1;
}

static const char *device_type_str(int device_type)
{
	switch(device_type) {
		case P54_DEVTYPE_USB1:
			return "USB version 1";
		case P54_DEVTYPE_USB2:
			return "USB version 2";
		case P54_DEVTYPE_PCI:
			return "PCI";
		default:
			return "Please report bug";
	}
}

static int show_hwinfo(int fd)
{
	struct p54_hwinfo hwinfo;

	if(!ioctl_err(fd, P54HWINFO, &hwinfo))
		return 0;
	if(!quiet)
		printf(">> Hardware information:\n");
	printf("Device type:\t\t%s\n", device_type_str(hwinfo.device_type));
	if(!hwinfo.eeprom_available) {
		printf("No more information available (EEPROM data not available)\n");
	} else {
		switch(hwinfo.platform_id) {
		case P54_PLATFORM_ID_3880:
			printf("Platform ID:\t\tISL3880\n");
			break;
		case P54_PLATFORM_ID_3886:
			printf("Platform ID:\t\tISL3886\n");
			break;
		case P54_PLATFORM_ID_3887:
			printf("Platform ID:\t\tGW3887\n");
			break;
		default:
			printf("Platform ID: Unknown (id 0x%08x)\n", hwinfo.platform_id);
			break;
		}
		printf("Manufacturing part nr.:\t%s\n", &hwinfo.manufacturing_part_number[0]);
		printf("NIC serial number:\t%s\n", &hwinfo.nic_serial_number[0]);
		printf("Antenna 1 gain:\t\t%d\n", hwinfo.antenna1_gain);
		printf("Antenna 2 gain:\t\t%d\n", hwinfo.antenna2_gain);
	}
	return 1;
}

static int show_fwinfo(int fd)
{
	struct p54_fwinfo fwinfo;

	if(!ioctl_err(fd, P54FWINFO, &fwinfo))
		return 0;
	if(!quiet)
		printf(">> Firmware information:\n");
	printf("Type:\t\t\t%c%c%c%c\n", fwinfo.type[0], fwinfo.type[1], fwinfo.type[2], fwinfo.type[3]);
	printf("Version:\t\t%s\n", &fwinfo.version[0]);
	printf("Build:\t\t\t%s\n", &fwinfo.build[0]);
	printf("Variant:\t\t%d.%d\n", (fwinfo.variant & 0xff00) >> 8, fwinfo.variant & 0xff);
	printf("Bottom compatibility:\t%d.%d\n", (fwinfo.bottom_compat & 0xff00) >> 8, fwinfo.bottom_compat & 0xff);
	printf("Top compatibility:\t%d.%d\n", (fwinfo.top_compat & 0xff00) >> 8, fwinfo.top_compat & 0xff);
	printf("Variant (phy):\t\t%d.%d\n", (fwinfo.physical_variant & 0xff00) >> 8, fwinfo.physical_variant & 0xff);
	printf("Bottom compat (phy):\t%d.%d\n", (fwinfo.physical_bottom_compat & 0xff00) >> 8, fwinfo.physical_bottom_compat & 0xff);
	printf("Top compat (phy):\t%d.%d\n", (fwinfo.physical_top_compat & 0xff00) >> 8, fwinfo.physical_top_compat & 0xff);
	return 1;
}

static int show_stats(int fd)
{
	struct p54_stats stats;

	if(!ioctl_err(fd, P54STATS, &stats))
		return 0;
	if(!quiet)
		printf(">> Stats:\n");
	printf("RX packets:\t\t%u\n", stats.rx_packets);
	printf("Time:\t\t\t%u\n", stats.time);
#ifdef STATS_SHOW_UNKNOWN
	printf("Unknown 1:\t\t%u\n", stats.unk1);
	printf("Unknown 2:\t\t%u\n", stats.unk2);
	printf("Unknown 3:\t\t%u\n", stats.unk3);
	printf("Unknown 4:\t\t%u\n", stats.unk4);
	printf("Unknown 5:\t\t%u\n", stats.unk5);
	printf("Unknown 6:\t\t%u\n", stats.unk6);
	printf("Unknown 7:\t\t%u\n", stats.unk7);
#endif

	return 1;
}

static const char *pmalloc_purpose_name(int p)
{
	switch(p) {
	case P54_PMALLOC_PURPOSE_CONTROL:
		return "Control";
	case P54_PMALLOC_PURPOSE_EEPROM:
		return "EEPROM";
	case P54_PMALLOC_PURPOSE_80211MGMT:
		return "802.11 management frame";
	case P54_PMALLOC_PURPOSE_80211DATA:
		return "802.11 data frame";
	case P54_PMALLOC_PURPOSE_80211BEACON:
		return "802.11 beacon";
	case P54_PMALLOC_PURPOSE_FWRESERVED:
		return "Reserved for firmware use";
	case P54_PMALLOC_PURPOSE_MISC:
		return "Miscellaneous";
	default:
		return "Unknown - make sure you are using p54ctl shipped with your current driver";
	}
}

static int show_pmalloc(int fd)
{
	struct p54_pmalloc pmalloc;
	unsigned int i;

	if(!ioctl_err(fd, P54PMALLOC, &pmalloc))
		return 0;
	if(!quiet)
		printf(">> Prism memory allocator state:\n");
	printf("Start address:\t\t0x%08x\n", pmalloc.start_address);
	printf("End address:\t\t0x%08x\n", pmalloc.end_address);
	printf("Allocated blocks:\n");
	if(pmalloc.allocated_blocks == 0) {
		printf("(none)\n");
	} else {
		for(i=0;i<pmalloc.allocated_blocks;i++) {
			printf(" [0x%08x] 0x%08x-0x%08x - %s\n", pmalloc.blocks[i].start_address+pmalloc.blocks[i].headroom, pmalloc.blocks[i].start_address, pmalloc.blocks[i].end_address, pmalloc_purpose_name(pmalloc.blocks[i].purpose));
		}
	}
	return 1;
}

static int dump_eeprom(int fd, const char *filename)
{
	FILE *wfd;
	char *eeprom_data;

	wfd = fopen(filename, "wb");
	if(wfd == NULL) {
		perror("fopen");
		return 0;
	}
	eeprom_data = malloc(P54_EEPROM_SIZE);
	if(eeprom_data == NULL) {
		perror("malloc");
		fclose(wfd);
		return 0;
	}
	if(!ioctl_err(fd, P54EEPROMDUMP, &eeprom_data))
		return 0;
	if(fwrite(eeprom_data, P54_EEPROM_SIZE, 1, wfd) != 1) {
		perror("fwrite");
		free(eeprom_data);
		fclose(wfd);
		return 0;
	}
	free(eeprom_data);
	if(fclose(wfd) != 0) {
		perror("fclose");
		return 0;
	}
	if(!quiet)
		printf(">> Successfully dumped EEPROM contents to file \"%s\"\n", filename);
	return 1;
}

#define CRASHDUMP_INITSIZE 65535

static int dump_crash(int fd, const char *filename)
{
	FILE *wfd;
	struct p54_crashdump crashdump;

	crashdump.size = CRASHDUMP_INITSIZE;
	crashdump.storage = malloc(CRASHDUMP_INITSIZE);
	if(crashdump.storage == NULL) {
		perror("malloc");
		return 0;
	}
	if(!ioctl_err(fd, P54CRASHDUMP, &crashdump))
		return 0;
	if(crashdump.size == 0) {
		printf("No crash data available\n");
		free(crashdump.storage);
		return 1;
	}
	if(crashdump.size > CRASHDUMP_INITSIZE) {
		free(crashdump.storage);
		crashdump.storage = malloc(crashdump.size);
		if(crashdump.storage == NULL) {
			perror("malloc");
			return 0;
		}
		if(!ioctl_err(fd, P54CRASHDUMP, &crashdump))
			return 0;
	}
	wfd = fopen(filename, "wb");
	if(wfd == NULL) {
		perror("fopen");
		free(crashdump.storage);
		return 0;
	}
	if(fwrite(crashdump.storage, crashdump.size, 1, wfd) != 1) {
		perror("fwrite");
		free(crashdump.storage);
		fclose(wfd);
		return 0;
	}
	if(fclose(wfd) != 0) {
		perror("fclose");
		free(crashdump.storage);
		return 0;
	}
	free(crashdump.storage);
	if(!quiet)
		printf(">> Successfully saved crash dump to file \"%s\"\n", filename);
	return 1;
}

#undef CRASHDUMP_INITSIZE

static int send_leds_control(int fd, struct p54_leds *leds)
{
	if(!ioctl_err(fd, P54LEDS, leds))
		return 0;
	if(!quiet)
		printf(">> LEDs control OK\n");
	return 1;
}

static int parse_led_cmd(char *cmd, struct p54_leds *leds)
{
	char *current;
	char *c;

	if(strcmp(optarg, "auto") == 0) 	{
		leds->drv_auto = 1;
	} else {
		leds->drv_auto = 0;
		current = cmd;

		c = strchr(current, ' ');
		if(c == NULL)
			return 0;
		*c = 0;
		if(strcmp(current, "set") == 0)
			leds->mode = P54_LED_MODE_SET;
		else {
			if(strcmp(current, "rxblink") == 0)
				leds->mode = P54_LED_MODE_ACTBLINK;
			else
				return 0;
		}
		current = c+1;

		c = strchr(current, ' ');
		if(c == NULL)
			return 0;
		*c = 0;
		leds->perm_setting = 0;
		if(strcmp(current, "1") == 0)
			leds->perm_setting |= P54_LED_A;
		else {
			if(strcmp(current, "0") != 0)
				return 0;
		}
		current = c+1;

		c = strchr(current, ' ');
		if(c == NULL)
			return 0;
		*c = 0;
		if(strcmp(current, "1") == 0)
			leds->perm_setting |= P54_LED_A;
		else {
			if(strcmp(current, "0") != 0)
				return 0;
		}
		current = c+1;

		c = strchr(current, ' ');
		if(c == NULL)
			return 0;
		*c = 0;
		leds->temp_setting = 0;
		if(strcmp(current, "1") == 0)
			leds->temp_setting |= P54_LED_A;
		else {
			if(strcmp(current, "0") != 0)
				return 0;
		}
		current = c+1;

		c = strchr(current, ' ');
		if(c == NULL)
			return 0;
		*c = 0;
		if(strcmp(current, "1") == 0)
			leds->temp_setting |= P54_LED_B;
		else {
			if(strcmp(current, "0") != 0)
				return 0;
		}
		current = c+1;

		leds->duration = strtoul(current, &c, 0);
		if(*c != 0)
			return 0;
	}
	return 1;
}

static int send_firmware_message(int fd, uint32_t address)
{
	char *linebuffer;
	struct p54_frame frame;
	int i;
	char *b, *c, *d;

	linebuffer = malloc(P54_MAX_FRAME_SIZE*3+2);
	if(linebuffer == NULL) {
		perror("malloc");
		return 0;
	}
	if(!quiet)
		printf(">> Sending firmware frame\n");
	if(!quiet)
		printf("Input firmware frame:\n");
	fgets(linebuffer, P54_MAX_FRAME_SIZE*3+2, stdin);
	*(strchr(linebuffer, '\n')) = 0;

	i = 0;
	c = linebuffer;
	while(1) {
		d = strchr(c, ' ');
		if(d == NULL) {
			frame.data[i] = strtoul(c, &b, 16);
			if(*b != 0) {
				fprintf(stderr, "Incorrect character in frame input near %s\n", c);
				free(linebuffer);
				return 0;
			}
			i++;
			break;
		}
		*d = 0;
		frame.data[i] = strtoul(c, &b, 16);
		if(*b != 0) {
			fprintf(stderr, "Incorrect character in frame input near %s\n", c);
			free(linebuffer);
			return 0;
		}
		i++;
		c = d+1;
	}
	free(linebuffer);
	frame.size = i;
	frame.address = address;
	if(!ioctl_err(fd, P54FRAME, &frame))
		return 0;
	if(!quiet)
		printf("Firmware frame successfully sent\n");
	return 1;
}

static int send_abortaddr(int fd, uint32_t address)
{
	if(!ioctl_err(fd, P54ABORTADDR, &address))
		return 0;
	if(!quiet)
		printf(">> Address 0x%08x aborted\n", address);
	return 1;
}

static int dump_memory(int fd, const char *filename, unsigned int addr, unsigned int size)
{
	FILE *wfd;
	struct p54_memread memread;


	wfd = fopen(filename, "wb");
	if(wfd == NULL) {
		perror("fopen");
		return 0;
	}
	memread.startaddr = addr;
	memread.length = size;
	memread.storage = malloc(size);
	if(memread.storage == NULL) {
		perror("malloc");
		fclose(wfd);
		return 0;
	}
	if(!ioctl_err(fd, P54MEMREAD, &memread))
		return 0;
	if(fwrite(memread.storage, size, 1, wfd) != 1) {
		perror("fwrite");
		free(memread.storage);
		fclose(wfd);
		return 0;
	}
	free(memread.storage);
	if(fclose(wfd) != 0) {
		perror("fclose");
		return 0;
	}
	if(!quiet)
		printf(">> Successfully saved memory data to file %s\n", filename);
	return 1;
}

static char *device_name = NULL;
static int mainfd;

static int reset_dev = 0;

static char *firmware_name = NULL;
static char *firmware_usb1_name = NULL;
static char *firmware_usb2_name = NULL;
static char *firmware_pci_name = NULL;

static int s_hwinfo = 0;
static int s_fwinfo = 0;
static int s_stats = 0;
static int s_pmalloc = 0;

static char *edump_name = NULL;
static char *cdump_name = NULL;

static int control_leds;
static struct p54_leds leds;

static int send_frame = 0;
static uint32_t prism_address;

static int do_abortaddr = 0;
static uint32_t abortaddr;

static unsigned int mdump_size = 0x1000;
static unsigned int mdump_addr = 0x00020000;
static char *mdump_name = NULL;

static int process()
{
	if(reset_dev) {
		if(!reset_device(mainfd))
			return 0;
	}
	if(firmware_name) {
		if(!upload_firmware(mainfd, firmware_name, -1))
			return 0;
	}
	if(firmware_usb1_name) {
		if(!upload_firmware(mainfd, firmware_usb1_name, P54_DEVTYPE_USB1))
			return 0;
	}
	if(firmware_usb2_name) {
		if(!upload_firmware(mainfd, firmware_usb2_name, P54_DEVTYPE_USB2))
			return 0;
	}
	if(firmware_pci_name) {
		if(!upload_firmware(mainfd, firmware_pci_name, P54_DEVTYPE_PCI))
			return 0;
	}
	if(s_hwinfo) {
		if(!show_hwinfo(mainfd))
			return 0;
	}
	if(s_fwinfo) {
		if(!show_fwinfo(mainfd))
			return 0;
	}
	if(s_stats) {
		if(!show_stats(mainfd))
			return 0;
	}
	if(s_pmalloc) {
		if(!show_pmalloc(mainfd))
			return 0;
	}
	if(edump_name != NULL) {
		if(!dump_eeprom(mainfd, edump_name))
			return 0;
	}
	if(cdump_name != NULL) {
		if(!dump_crash(mainfd, cdump_name))
			return 0;
	}
	if(control_leds) {
		if(!send_leds_control(mainfd, &leds))
			return 0;
	}
	if(send_frame) {
		if(!send_firmware_message(mainfd, prism_address))
			return 0;
	}
	if(do_abortaddr) {
		if(!send_abortaddr(mainfd, abortaddr))
			return 0;
	}
	if(mdump_name != NULL) {
		if(!dump_memory(mainfd, mdump_name, mdump_addr, mdump_size))
			return 0;
	}
	return 1;
}

static void cleanup()
{
	free(device_name);
	free(firmware_name);
	free(firmware_usb1_name);
	free(firmware_usb2_name);
	free(firmware_pci_name);
	free(edump_name);
	free(cdump_name);
	free(mdump_name);
}

int main(int argc, char *argv[])
{
	int c;
	char *b;

	while((c = getopt(argc, argv, "Qi:ru:1:2:p:hwcqe:d:s:l:f:m:n:o:")) != -1) {
		switch(c) {
		case 'Q':
			quiet = 1;
			break;
		case 'r':
			reset_dev = 1;
			break;
		case 'u':
			free(firmware_name);
			firmware_name = strdup(optarg);
			break;
		case '1':
			free(firmware_usb1_name);
			firmware_usb1_name = strdup(optarg);
			break;
		case '2':
			free(firmware_usb2_name);
			firmware_usb2_name = strdup(optarg);
			break;
		case 'p':
			free(firmware_pci_name);
			firmware_pci_name = strdup(optarg);
			break;
		case 'i':
			free(device_name);
			device_name = strdup(optarg);
			break;
		case 'h':
			s_hwinfo = 1;
			break;
		case 'w':
			s_fwinfo = 1;
			break;
		case 'c':
			s_stats = 1;
			break;
		case 'q':
			s_pmalloc = 1;
			break;
		case 'e':
			edump_name = strdup(optarg);
			if(edump_name == NULL) {
				perror("strdup");
				cleanup();
				return 1;
			}
			break;
		case 'd':
			cdump_name = strdup(optarg);
			if(cdump_name == NULL) {
				perror("strdup");
				cleanup();
				return 1;
			}
			break;
		case 'l':
			control_leds = 1;
			b = strdup(optarg);
			if(b == NULL) {
				perror("strdup");
				cleanup();
				return 1;
			}
			if(!parse_led_cmd(b, &leds)) {
				fprintf(stderr, "Failed to parse LED parameters\n");
				free(b);
				cleanup();
				return 1;
			}
			free(b);
			break;
		case 's':
			prism_address = strtoul(optarg, &b, 0);
			if(*b != 0) {
				fprintf(stderr, "Incorrect Prism address\n");
				cleanup();
				return 1;
			}
			send_frame = 1;
			break;
		case 'f':
			abortaddr = strtoul(optarg, &b, 0);
			if(*b != 0) {
				fprintf(stderr, "Incorrect address\n");
				cleanup();
				return 1;
			}
			do_abortaddr = 1;
			break;
		case 'm':
			mdump_name = strdup(optarg);
			if(mdump_name == NULL) {
				perror("strdup");
				cleanup();
				return 1;
			}
			break;
		case 'n':
			mdump_addr = strtoul(optarg, &b, 0);
			if(*b != 0) {
				fprintf(stderr, "Incorrect address for memory dump\n");
				cleanup();
				return 1;
			}
			break;
		case 'o':
			mdump_size = strtoul(optarg, &b, 0);
			if(*b != 0) {
				fprintf(stderr, "Incorrect size for memory dump\n");
				cleanup();
				return 1;
			}
			break;
		default:
			cleanup();
			return 1;
		}
	}
	if(device_name == NULL) {
		fprintf(stderr, "Device not specified\n");
		return 1;
	}

	mainfd = open(device_name, O_RDWR);
	if(mainfd == -1) {
		perror("open");
		cleanup();
		return 1;
	}

	if(!quiet)
		printf(">> Device:\t\t%s\n", device_name);

	if(!process()) {
		cleanup();
		close(mainfd);
		return 1;
	}

	cleanup();
	close(mainfd);
	return 0;
}
